# Functions for managing the administrative web server
#
# $Id: Admserv.pm,v 1.34 1999/10/02 16:44:26 duncan Exp $
#
# Written by Brion Moss (brion@cobaltnet.com, brion@nda.com)
#
# Copyright 1998 Cobalt Networks http://www.cobaltnet.com
#

package Cobalt::Admserv;

use vars qw(@ISA @EXPORT @EXPORT_OK);

require Exporter;
require SelfLoader;
@ISA    = qw(Exporter SelfLoader);
@EXPORT = qw(admserv_get_access admserv_set_access
	     admserv_set_user_cgis admserv_get_user_cgis
		admserv_restart);

use Cobalt::Product;
use Cobalt::Util;

1;
__DATA__

sub admserv_set_access
# admserv_set_access: set the access rights to a directory
# Arguments: path, mode, type, and name
#   mode may be "any", "admin", or "valid":
#    any -- open access
#    admin -- require user admin
#    site-admin -- require a site administrator
#    valid -- require valid-user
#    require -- require the given user
#    none -- no access restrictions (remove the file)
#   The type and name arguments is optional, unless the require mode is used
#   The admin user or group are always allowed access
# Return value: status string of the usual sort
# Side effects: creates the htaccess files
# Example: Cobalt::Admserv::admserv_set_access("/home/httpd/html/joe","require","user","joe");
{
    my ($dir,$mode,$type,$user) = @_;

    return MSG_error("admserv_invalid_dir", $dir) 
      if (($dir !~ /^$Htusers_dir/) && ($dir !~ /^$Htgroup_dir/) && ($dir !~ /^$Top_dir/o));
    if ($mode eq "none")
    {
	if (-e "$dir/$Access_file")
	{
	    unlink("$dir/$Access_file") ||
		return MSG_error("admserv_cant_remove_accessfile", $dir);
	}
	return MSG_ok("admserv_ok_opened", $dir);
    }
    open(HTA,">$dir/$Access_file") ||
	return MSG_error("admserv_cant_write_accessfile", $dir);
    # No explicit <Limit>, because we want to restrict all methods
    print HTA "# Access file for $dir ($mode $type $user)\n";
    print HTA "order allow,deny\n";
    print HTA "allow from all\n";
    if    ($mode eq "valid")   { print HTA "require valid-user\n";   }
    elsif ($mode eq "admin")   { print HTA "require user $Adminuser\n";   }
    elsif ($mode eq "require") { print HTA "require $type $user $Adminuser\n";}
    elsif ($mode eq "site-admin") {
      print HTA "require group $MainGroup $Sites_AdminGroup $Adminuser\n";
    }
    print HTA "Authname $Usr_Authname\nAuthtype $Authtype\n";
    unless ($Features{'pam'}) {
      print HTA "AuthUserFile $Htpasswd\n";
      print HTA "AuthGroupFile $Htgroup\n";
    }
    close HTA;

    # web server needs to read the file
    chmod 0644, "$dir/$Access_file";

    return MSG_ok("admserv_ok_set_rights", $dir);
}

sub admserv_get_access
# admserv_get_access: get the access rights for a directory
# Arguments: a directory name
# Return value: a mode, as above
# Example: $protected = Cobalt::Admserv::admserv_get_access("/home/admserv/.cobalt/place");
{
    my ($dir) = @_;
    return "none" if (! -e "$dir/$Access_file");
    open(HTA,"$dir/$Access_file") ||
	return MSG_error("admserv_cant_read_accessfile", $dir);
    while (<HTA>)
    {
	next unless (/^require/o);
	if (/^require valid/o)
	{
	    close HTA;
	    return "valid";
	}
	elsif (/^require (\w+) (.+) $Adminuser/o)
	{
	    close HTA;
	    return "require $1 $2";
	}
    }
    close HTA;
    return "any";
}

sub admserv_set_user_cgis
# Turn CGI access for users on or off
# Arguments: 1 for on, 0 for off
# Side effects: creates the file which tells cgiwrap wether or not
#   to allow user CGIs.  cgiwrap just looks to see if the file
#   exists, and if it does it permits users to run CGIs in their
#   own area.
# Return value: the usual status string
# Example: Cobalt::Admserv::admserv_set_user_cgis(0);
{
    my ($turnon) = @_;

    return MSG_error("admserv_unsupported_cgis")
      unless ($Features{'www'});

    if ($turnon && -e $Allow_cgis_file)
    {
	return MSG_ok("admserv_ok_cgis_already_on");
    }
    elsif ($turnon)
    {
	open(ALLOW,">$Allow_cgis_file")
	    || return MSG_error("admserv_cant_cgis_on", $!);
	close ALLOW;
	return MSG_ok("admserv_ok_cgis_on");
    }
    elsif (-e $Allow_cgis_file)
    {
	unlink($Allow_cgis_file)
	    || return MSG_error("admserv_cant_cgis_off", $!);
	return MSG_ok("admserv_ok_cgis_off");

    }
    else
    {
	return MSG_ok("admserv_ok_cgis_already_off");
    }
}

sub admserv_get_user_cgis
# Report on wether user CGIs are allowed
# Arguments: none
# Return value: 1 if yes, 0 if no
# Side effects: none
# Example: $cgis_are_on = Cobalt::Admserv::admserv_get_user_cgis();
{
    (-e $Allow_cgis_file) ? (return 1) : (return 0);
}

sub restart_httpd
# Restart the web server
# Arguments: none
# Side effects: sends a HUP to the server
{
    my ($pid,$l);
    open(PID,$Htpid) 
	|| return MSG_error("admserv_cant_read_wwwpid", $Htpid);
    $l = <PID>;
    if ($l =~ /^(\d+)$/o)
    {
	$pid = $1;
    }
    close PID;
    kill('HUP',$pid) if ($pid);

    return MSG_ok("admserv_ok_www_restarted", $pid);
}

sub admserv_restart
# Restart the web server
# Arguments: none
# Side effects: sends a HUP to the server
{
    my ($pid,$l);
    open(PID,$AdminPid) 
	|| return MSG_error("admserv_cant_read_wwwpid", $Htpid);
    $l = <PID>;
    if ($l =~ /^(\d+)$/o)
    {
	$pid = $1;
    }
    close PID;
    kill('USR1',$pid) if ($pid);

    return MSG_ok("admserv_ok_www_restarted", $pid);
}

1;
