# $Id: Atalk.pm,v 1.14 1999/10/02 16:44:26 duncan Exp $
#
# Atalk.pm
# Netatalk related routines
#
# John D. Blair <jdblair@cobaltnet.com>
# Copyright 1998 Cobalt Networks, Inc., http://www.cobaltnet.com
#

package Cobalt::Atalk;
use vars qw(@ISA @EXPORT @EXPORT_OK);

require Exporter;
require SelfLoader;
@ISA    = qw(Exporter SelfLoader);
@EXPORT	= qw(atalk_lock_passwd
	     atalk_unlock_passwd
	     atalk_commit_passwd
	     atalk_set_passwd
	     atalk_del_user
	     atalk_set_volume
	     atalk_del_volume);

use strict;
use Cobalt::Util;
use Cobalt::Product;

1;
__DATA__

# atalk_lock_passwd
#
# locks the netatalk passwd file by creating $Netatalk_ptmp
#
# no args, no return unless error
#
# this *must* be called before atalk_set_passwd or atalk_del_user
#
sub atalk_lock_passwd 
{
  print STDERR "atalk: called Cobalt::Atalk::atalk_lock_passwd()\n" if ($Debug);

  open(PTMP, ">$Netatalk_ptmp")
    or return MSG_error("atalk_cant_open_ptmp", $Netatalk_ptmp);
  close(PTMP);
  return MSG_ok("atalk_ok_passwd_locked");
}

# atalk_unlock_passwd
#
# unlocks the netatalk passwd file by removing $Netatalk_ptmp
#
# no args, no return unless error
#
sub atalk_unlock_passwd 
{
  print STDERR "atalk: called Cobalt::Atalk::atalk_unlock_passwd()\n" if ($Debug);

  unlink($Netatalk_ptmp)
    or return MSG_error("atalk_cant_unlink_ptmp", $Netatalk_ptmp);
  return MSG_ok("atalk_ok_passwd_unlocked");
}

# atalk_passwd_commit
#
# commits changes to the netatalk passwd file
#
# no args, no return unless error
#
sub atalk_commit_passwd 
{
  print STDERR "atalk: called Cobalt::Atalk::atalk_commit_passwd()\n" if ($Debug);
  print STDERR "atalk: renaming $Netatalk_ptmp to $Netatalk_passwd\n" if ($Debug);
  rename($Netatalk_ptmp, $Netatalk_passwd)
    or return MSG_error("atalk_cant_commit_passwd", $Netatalk_ptmp);
  return MSG_ok("atalk_ok_passwd_committed");
}

# atalk_set_passwd
# sets the plaintext password in the netatalk password file
# args:
#       $user = username
#       $passwd = password
#
# Note: The Samba equivelent of this function is in User::user_add.
#       In retrospect, I think it was a mistake not to seperate it out
#
sub atalk_set_passwd 
{
  my ($user, $passwd) = @_;
  my $line;
  my $found = 0;

  print STDERR "atalk: called Cobalt::Atalk::atalk_set_passwd($user, $passwd)\n" if ($Debug);

  # make sure files can only be read by root
  umask(0177);

  # does $Netatalk_passwd exist?
  if (! -e $Netatalk_passwd) {
    open(ATALKPASS, ">$Netatalk_passwd")
      or return MSG_error("atalk_cant_write_passwd", $Netatalk_passwd);
    close(ATALKPASS);
  }

  # security check
  # this is *very* important... this file contains *plaintext* passwords
  chown(0,0,$Netatalk_passwd);
  chmod(0600,$Netatalk_passwd);

  # parse & modify passwd file
  open(PTMP, ">>$Netatalk_ptmp")
    or return MSG_error("atalk_cant_append_ptmp", $Netatalk_ptmp);
  # protect cleartext password
  chmod(0600,$Netatalk_ptmp);
  open(ATALKPASS, "$Netatalk_passwd")
    or return MSG_error("atalk_cant_open_passwd", $Netatalk_passwd);
  my $inconfig = 0;
  while (<ATALKPASS>) {
    print STDERR "atalk: $_" if ($Debug);
    # comment?
    if (/^\s*\#/) {
      print PTMP;
      next;
    }

    # valid password line?
    if (/^([^:]+)+:.+\n$/o) {
      # write it out
      if ($user eq $1) {
	print PTMP "$user:$passwd\n";
	print STDERR "atalk: replacing $user:$passwd\n" if ($Debug);
	$found = 1;
      } else {
	print PTMP;
      }
    }
  }

  # if we've made it here and $found = 0 then the user didn't exist in the
  # the file
  if (! $found) {
    print PTMP "$user:$passwd\n";
    print STDERR "atalk: adding $user:$passwd\n" if ($Debug);
  }

  close(PTMP);
  close(ATALKPASS);
  return MSG_ok("atalk_ok_set_passwd", $user);
}


# atalk_del_user
# remove a user's record from the netatalk passwd file
# args:
#       $user = username
#
# No check is done to make sure $user is a valid user.
#
# Note: The Samba equivelent of this function is in User::user_del.
#       In retrospect, I think it was a mistake not to seperate it out
#
sub atalk_del_user 
{
  my ($user) = @_;
  my $line;

  print STDERR "atalk: called Cobalt::Atalk::atalk_del_user($user)\n" if ($Debug);

  # make sure files can only be read by root
  umask(0177);
  # security check
  # this is *very* important... this file contains *plaintext* passwords
  chown(0,0,$Netatalk_passwd);
  chmod(0600,$Netatalk_passwd);

  # parse & modify passwd file
  open(PTMP, ">>$Netatalk_ptmp")
    or return MSG_error("atalk_cant_append_ptmp", $Netatalk_ptmp);
  open(ATALKPASS, "$Netatalk_passwd")
    or return MSG_error("atalk_cant_open_passwd", $Netatalk_passwd);
  while (<ATALKPASS>) {
    # comment?
    if (/^\s*\#/) {
      print PTMP;
      next;
    }

    # valid password line?
    if (/^([^:]+)+:.+\n$/o) {
      # write out the line if it doesn't belong to $user
      if ($user ne $1) {
	print PTMP;
      }
    }
  }
  close(PTMP);
  close(ATALKPASS);
  return MSG_ok("atalk_ok_user_removed", $user);
}

#
# atalk_set_volume
#
# modifies an existing volume or adds a volume to the volume list
#
# arg: $volume = path to volume
#
sub atalk_set_volume 
{
  my ($volume, $name) = @_;

  print STDERR "atalk_set_volume: modifying/adding volume $volume, $name\n" if ($Debug);

  # remove " from $name
  $name =~ s/\"//go;

  return MSG_error("atalk_cant_lock_applevolumes", $AppleVolumes_lock) if (-e $AppleVolumes_lock);
  open (VOLLOCK, ">$AppleVolumes_lock")
    or return MSG_error("atalk_cant_open_lock", $AppleVolumes_lock);
  open (VOL, "$AppleVolumes")
    or return MSG_error("atalk_cant_open_applevolumes", $AppleVolumes);

  my $inconfig = 0;
  while (<VOL>) {
    if ($inconfig) {
      if (/^\# Cobalt config stop/) {
	print VOLLOCK "$volume \"$name\"\n";
	$inconfig = 0;
      }
      if (/^$volume\s+\"/) {
	# skip this line
	next;
      }
    }
    if (/^\# Cobalt config start/) {
      $inconfig = 1;
    }
    print VOLLOCK;
  }
  close(VOLLOCK);
  close(VOL);

  rename($AppleVolumes_lock, $AppleVolumes)
    or return MSG_error("atalk_cant_rename_applevolumes", $AppleVolumes_lock);
  
  # success
  return MSG_ok("atalk_ok_volume_modified", $name);
}


#
# atalk_del_volume
#
# adds a volume to the volume list
#
# arg: $volume = path to volume
#
sub atalk_del_volume 
{
  my $volume = shift;

  my $found = 0;

  return MSG_error("atalk_cant_lock_applevolumes", $AppleVolumes_lock) if (-e $AppleVolumes_lock);
  open (VOLLOCK, ">$AppleVolumes_lock")
    or return MSG_error("atalk_cant_open_lock", $AppleVolumes_lock);
  open (VOL, "$AppleVolumes")
    or return MSG_error("atalk_cant_open_applevolumes", $AppleVolumes);
  
  my $inconfig = 0;
  while (<VOL>) {
    if ($inconfig) {
      if (/^\# Cobalt config stop/) {
	$inconfig = 0;
      }
      if (/^$volume\s+\"/) {
	# skip this line
	next;
      }
    }
    if (/^\# Cobalt config start/) {
      $inconfig = 1;
    }
    print VOLLOCK;
  }
  close(VOLLOCK);
  close(VOL);

  rename($AppleVolumes_lock, $AppleVolumes)
    or return MSG_error("atalk_cant_rename_applevolumes", $AppleVolumes_lock);
  return MSG_ok("atalk_ok_volume_removed", $volume);
}
