#
# $Id: 
#
# Written by
#   Duncan Laurie  <duncan@cobaltnet.com>
#
# Copyright 1999 Cobalt Networks http://www.cobaltnet.com/
#

package Cobalt::BandWidth;

use strict;
use vars qw(@ISA @EXPORT @EXPORT_OK);

require Exporter;
require SelfLoader;
@ISA    = qw(Exporter SelfLoader);
@EXPORT = qw(bw_setip bw_getip bw_delip bw_syncip bw_statip bw_shutdown);

use Cobalt::Util;
use Cobalt::Meta;
use Cobalt::Product;

use integer;

1;

__DATA__

sub bw_setip
# set a bandwidth limit on an IP#
# 1. store the bandwidth limit in the database
# 2. calls bw_sync to sync database with kernel
# 
# arguments: 
#            IP: ip address to limit
#            Limit: the limit in bytes/second
#            Protocol: protocol (tcp/udp)
# [opt]      Port: optional port to limit
# [nyi,opt]  Direction: optional direction ( r | rw | w )
#
{
    my %opt = @_;

    my $limit = (($opt{"bwlimit"} * 1000) / 8);
    $opt{"direction"} = 'w' unless (defined ($opt{"direction"}));
    $opt{"protocol"} = 'tcp' unless (defined ($opt{"protocol"}));
    my $obj = Cobalt::Meta->create("type"      => "bw",
				   "method"    => "ip",
				   "name"      => $opt{"ip"},
				   "bwlimit"   => $opt{"bwlimit"},
				   "protocol"  => $opt{"protocol"},
				   "port"      => $opt{"port"},
				   "direction" => $opt{"direction"});
    $obj->save;
    my $ip = $opt{"ip"};
    my $protocol = $opt{"protocol"};

    my $cmdline = "$bw_setip_prog -i $ip -b $limit -p $protocol";
    $cmdline .= "/$port" if (defined ($port));
    $cmdline .= " -t $direction" if (defined ($direction));
    system("$cmdline > /dev/null 2>&1");

    return MSG_ok("bw_setip_ok");
}

sub bw_delip
# delete bandwidth setting for the specified IP
# arguments: "ip" => $ip
{
    my $ip = shift;

    my $obj = Cobalt::Meta->create("type" => "bw", "method" => "ip");
    $obj->retrieve($ip);
    my $prot = $obj->get("protocol");
    my $cmdline = "$bw_setip_prog -i $ip -p $prot -d";
    system("$cmdline > /dev/null 2>&1");
    $obj->delete;

    return MSG_ok("bw_delip_ok");
}

sub bw_getip
# get bandwidth limiting info for a particular ip
# arguments: "ip" => $ip
# returns: bytes/second bandwidth limit for that ip
{
    my $ip = shift;

    my $obj = Cobalt::Meta->create("type" => "bw",
				   "method" => "ip");
    $obj->retrieve($ip);

    return $obj->get("bwlimit");
}

sub bw_statip
# get bandwidth limiting stats for a particular ip
# arguments: "ip" => $ip
# returns: array of [5min 10min 15min] limiting statistics
{
    my $ip = shift;
    my $statfile = "/proc/bw_limit/ip/".$ip;
    my @stats;

    FILE_lock_r($statfile, *STAT);
    while (<STAT>) {
	if (/([\d\.]+) ([\d\.]+) ([\d\.]+)/) {
	    push @stats, $1, $2, $3;
	}
    }
    FILE_unlock(*STAT);

    return @stats;
}

sub bw_syncip
# synchronizes the bandwidth table with the kernel
{
    my $obj = Cobalt::Meta->create("type" => "bw");
    my @limits = $obj->getall;
    my ($ip,$bwlimit,$protocol,$port,$direction);

    foreach (@limits) {
	$obj->retrieve($_);
	$ip = $obj->get("name");
	next if ($ip eq "default");
	$bwlimit = ($obj->get("bwlimit") > 0) ? (($obj->get("bwlimit") * 1000) / 8) : (1250);
	$protocol = $obj->get("protocol");
	$port = $obj->get("port");
	$direction = $obj->get("direction");

	$cmdline = "$bw_setip_prog -i $ip -b $bwlimit -p $protocol";
	$cmdline .= "/$port" if (defined ($port));
	$cmdline .= " -t $direction" if (defined ($direction));
	system("$cmdline > /dev/null 2>&1");
    }

    return MSG_ok("bw_syncip_ok");
}

sub bw_shutdown
# removes all bandwidth limits
{
    my $obj = Cobalt::Meta->create("type" => "bw");
    my @limits = $obj->getall;
    my ($ip,$protocol);

    foreach (@limits) {
	$obj->retrieve($_);
	$ip = $obj->get("name");
	next if ($ip eq "default");
	$protocol = $obj->get("protocol");

	$cmdline = "$bw_setip_prog -i $ip -p $protocol -d";
	system("$cmdline > /dev/null 2>&1");
    }

    return MSG_ok("bw_shutdown_ok");
}

