#
# Written by
#   Will DeHaan  will@cobaltnet.com
#
# Copyright 1998 Cobalt Networks http://www.cobaltnet.com/
#

package Cobalt::Dhcpd;
use vars qw(@ISA @EXPORT @EXPORT_OK);

require Exporter;
require SelfLoader;

@ISA    = qw(Exporter SelfLoader);
@EXPORT = qw(dhcpd_get_parameters dhcpd_set_parameters dhcpd_set_name
 	     dhcpd_get_macs dhcpd_set_macs dhcpd_set_range dhcpd_get_range
	     dhcpd_add_macs dhcpd_del_macs dhcpd_create_conf
	     dhcpd_clean_ranges);

require Cobalt::Network;
import Cobalt::Network qw(:DEFAULT);
require Cobalt::Netutil;
import Cobalt::Netutil qw(:DEFAULT);
use Cobalt::Util;
use Cobalt::Product;

use strict;

1;
__DATA__

sub dhcpd_del_macs
# Wrapper to dchpd_set_macs to delete mac address and ip pairs for
# static dhcpd naming
# Arguments: hash of the form mac->name
# Return value: return code and descriptive string of the normal sort
{
	my @delmax = @_;
	my(%max) = Cobalt::Dhcpd::dhcpd_get_macs();

	my($mac);
	foreach $mac (@delmax)
	{
		delete $max{$mac};
	}

	return Cobalt::Dhcpd::dhcpd_set_macs(%max);
}

sub dhcpd_add_macs
# Wrapper to dchpd_set_macs to add mac address and ip pairs for
# static dhcpd naming
# Arguments: hash of the form mac->name
# Return value: return code and descriptive string of the normal sort
{
	my(%newmax) = @_;
	my(%max) = Cobalt::Dhcpd::dhcpd_get_macs();

	return Cobalt::Dhcpd::dhcpd_set_macs( (%newmax,%max) );
}

sub dhcpd_get_macs
# Obtains a listing of mac address associated names/ips
# Arguments: none
# Return value: hash of the form mac->name
#   mac address is of the form 55:66:77:AA:BB:CC
#   name may be a computer name or an ip address
{
	my($mac,$host,%max);
	# it's not an error to have no config file; it just means
	# we haven't made one yet.
	return if (! -e $Dhcpd_conf);
	open(DHCPcnf,"$Dhcpd_conf") || return MSG_error("dhcpd_cant_noconfig", $!);
	while(<DHCPcnf>)
	{
		$mac = $1 if (/^\s*hardware ethernet ([^\;]+)\;/o);
		$host = $1 if (/^\s*fixed-address ([^\;]+)\;/o);
		if ($mac && $host)
		{
			$max{$mac} = $host;
			$mac = $host = '';
		}
	}
	close(DHCPcnf);

	return(%max);
}

sub dhcpd_set_macs
# Saves a list of fixed ip addresses by mac address
# Arguments: hash of mac->ip/name
# Return value: Assuring string of the form MSG_ok("2121")
{
	my(%max) = @_;
	sleep 2 if (-e "$Lockdir/dhcpd.conf"); # Avoid race
	return MSG_error("dhcpd_cant_locked") if (-e "$Lockdir/dhcpd.conf");

	my($ignore); # Hats off for all the lost code out there..
	open(DHCPcnf,"$Dhcpd_conf") 
          || return MSG_error("dhcpd_cant_noconfig", $!);
	open(DHCPnew,"> $Lockdir/dhcpd.conf") 
	  || return MSG_error("dhcpd_cant_update", $!);
	while(<DHCPcnf>)
	{
		print DHCPnew unless (/^host /o ... /^}\s*$/o);
	}
	close(DHCPcnf);

	my($mac);
	foreach $mac (keys %max)
	{
	  print STDERR "Adding $mac and ", $max{$mac}, "\n" if ($Debug);
	  # Check our arguments
	  if ($mac !~ /^(([0-9A-Fa-f]{1,2}):){5}[0-9A-Fa-f]{2}$/o) {
	    unlink("$Lockdir/dhcpd.conf");
	    return MSG_error("dhcpd_invalid_macaddr", $mac);
	  }
	  if (!Cobalt::Netutil::netutil_test_ipaddr($max{$mac}) && ($max{$mac} !~ /[a-z]/io)) {
	    unlink("$Lockdir/dhcpd.conf");
	    return MSG_error("dhcpd_invalid_ipaddr", $max);
	  }

	  my $hname; # Must respect both IP and name arguments
	  if ($max{$mac} =~ /[a-z]/io)
	  {
	    $hname = $max{$mac};
	  } else # Try to resolve the name of this host
	  {
	    $hname = gethostbyaddr(pack('C4',split('\.',$max{$mac})),2);
	    $hname = $max{$mac} unless ($hname);  # default to IP address
	  }
	  
	  print DHCPnew <<EOF if ($max{$mac} && $mac);
host $hname {
  hardware ethernet $mac;
  fixed-address $max{$mac};
}
EOF
	}

	close(DHCPnew);

	rename("$Lockdir/dhcpd.conf","$Dhcpd_conf") 
            || return MSG_error("dhcpd_cant_update", $!);
	
	Cobalt::Dhcpd::dhcpd_hup(); # Kick it
	return MSG_ok("dhcpd_ok_updated");
}

sub dhcpd_get_range
# Obtains a list of dhcp service ranges
# Arguments: none
# Return value: hash of the form (low ip -> high ip)
{
	my(%range);
	
	# it's not an error to have no config file; it just means
	# we haven't made one yet.
	return if (! -e $Dhcpd_conf);
	open(DHCPcnf,"$Dhcpd_conf") || return MSG_error("dhcpd_cant_noconfig", $!);
	while(<DHCPcnf>)
	{
		$range{$1}=$2 if (/^\s*range ([\d\.]+) ([\d\.]+)\;/o);
	}
	close(DHCPcnf);

	return(%range);
}

sub dhcpd_set_range
# Allocate a range of addresses to give out dynamically
# Arguments: a hash table which lists a set of ranges, in the format
#          %hash = ($low1 => $high1, $low2 => $high2, ...);
# Return value: usual status string of form: MSG_error("4819")
# Side effects: modifies the DHCPD config file, HUPs dhcpd
{
	my(%range) = @_;
	my($setRange,$low);
	sleep 2 if (-e "$Lockdir/dhcpd.conf"); # Avoid race
	return MSG_error("dhcpd_cant_locked") if (-e "$Lockdir/dhcpd.conf");

	# Get the network parameters for the primary interface
	my($ipaddr,$nm) = Cobalt::Network::net_get_ipaddr();

	foreach $low (keys %range) {
	  unless (Cobalt::Network::net_network_ismember($low,$ipaddr,$nm)) {
	    return MSG_error("dhcpd_invalid_range_mismatch");
	  }
	  unless (Cobalt::Network::net_network_ismember($range{$low},$ipaddr,$nm)) {
	    return MSG_error("dhcpd_invalid_range_mismatch");
	  }
	  my @low = split(/\./o,$low);
	  my @hi  = split(/\./o,$range{$low});
	  unless (($low[3] <= $hi[3]) &&
		  ($low[2] <= $hi[2]) &&
		  ($low[1] <= $hi[1]) &&
		  ($low[0] <= $hi[0])) {
	    return MSG_error("dhcpd_invalid_wrongorder");
	  }
	}

	open(DHCPcnf,"$Dhcpd_conf") 
	  || return MSG_error("dhcpd_cant_noconfig", $!);
	open(DHCPnew,"> $Lockdir/dhcpd.conf") 
	  || return MSG_error("dhcpd_cant_update", $!);
	while(<DHCPcnf>)
	{
		print DHCPnew  unless (/^\s*subnet /o ... /^\s*\}\s*$/o);
	}
	close(DHCPcnf);

	# Create a subnet block if necessary
	my ($subnet) = Cobalt::Network::net_get_network($ipaddr,$nm);
	print DHCPnew "subnet $subnet netmask $nm {\n";
	foreach $low (keys %range)
	{
		print DHCPnew "  range $low ".$range{$low}.";\n";
	}
	print DHCPnew "}\n";
	close(DHCPnew);

	rename("$Lockdir/dhcpd.conf","$Dhcpd_conf")
	  || return MSG_error("dhcpd_cant_update", $!);

	Cobalt::Dhcpd::dhcpd_hup(); # Kick it
	return MSG_ok("dhcpd_ok_updated");
}

sub dhcpd_get_parameters
# Returns dhcp server settings
# Arguments: none
# Return value: domain name, dns server 1, dns server 2, 
#   gateway, netmask, maximum lease time, 
# Reads $Dhcpd_conf; returns null if no conf file
#  
{
	my($domain,$dns1,$dns2,$gw,$nm,$maxlease);

	# if no config file, read in our defaults from system settings
	return if (! -e "$Dhcpd_conf");
	#if (! -e "$Dhcpd_conf") {
	#  ($domain,$dns1,$dns2) = Cobalt::Network::net_get_dnsclient();
	#  $gw = (Cobalt::Network::net_get_gateway())[0];
	#  $nm = (Cobalt::Network::net_get_ipaddr())[1];
	#  $maxlease = $Dhcpd_default_lease;
	#}
	#else {
	  # read from the config file
	  open(DHCPcnf,"$Dhcpd_conf") || return MSG_error("dhcpd_cant_noconfig", $!);
	  while(<DHCPcnf>)
	  {
	    if (/^option domain-name \"([^\"]+)\"\;/o) { 
	      $domain = $1; 
	    }
	    elsif (/^option domain-name-servers ([\d\.]+)(, ([\d\.]+))?\;/o) {
	      ($dns1,$dns2) = ($1,$3);
	    }
	    elsif (/option routers ([\d\.]+)\;/o) {
	      $gw = $1;
	    }
	    elsif (/option subnet-mask ([\d\.]+)\;/o) {
	      $nm = $1;
	    }
	    elsif (/max-lease-time (\d+)\;/o) {
	      $maxlease = $1;
	    }
	  }
	  close(DHCPcnf);
	#}

	return ($domain,$dns1,$dns2,$gw,$nm,$maxlease);
}

sub dhcpd_set_name
# Set the server identifier for the DHCP server
# Arguments: hostname
# Side effects: modifies dhcpd.conf, restarts dhcpd
# Return value: status string like MSG_ok("2867")
{
  my ($name) = @_;
  sleep 2 if (-e "$Lockdir/dhcpd.conf"); # Avoid race
  return MSG_error("dhcpd_cant_locked") if (-e "$Lockdir/dhcpd.conf");
  open(DHCPcnf,$Dhcpd_conf) 
    || return MSG_error("dhcpd_cant_noconfig", $!);
  open(DHCPnew,"> $Lockdir/dhcpd.conf") 
    || return MSG_error("dhcpd_cant_update", $!);
  while (<DHCPcnf>) {
    if (/^server-identifier /o) {
      print DHCPnew "server-identifier $name;\n";
    } else {
      print DHCPnew;
    }
  }
  close DHCPcnf;
  close DHCPnew;
  rename("$Lockdir/dhcpd.conf","$Dhcpd_conf")
    || return MSG_error("dhcpd_cant_update", $!);
  Cobalt::Dhcpd::dhcpd_hup();
  return MSG_ok("dhcpd_ok_updated");
}

sub dhcpd_set_parameters
# Set the dhcpd server paramters
# Arguments: domain name, dns server 1, dns server 2,
#   gateway, netmask, maximum lease time
#   Any argument except the netmask can be empty.
# Side Effects: writes $Dhcpd_conf; creates a new one if none exists
#   This function does *not* restart dhcpd; it assumes that
#   dhcpd_clean_ranges() will be called next, and that function
#   takes care of restarting the daemon.
{
	my($domain,$dns1,$dns2,$gw,$nm,$maxlease) = @_;
	return MSG_error("dhcpd_invalid_netmask", $nm)
	  if (($nm eq "0.0.0.0") ||
	      (Cobalt::Network::net_convert_netmask($nm) !~ /^\d+$/o));

	my($ipaddr) = (Cobalt::Network::net_get_ipaddr())[0];
	my($broadcast) = Cobalt::Network::net_get_broadcast($ipaddr,$nm);

	sleep 2 if (-e "$Lockdir/dhcpd.conf"); # Avoid race
	return MSG_error("dhcpd_cant_locked") if (-e "$Lockdir/dhcpd.conf");
	return MSG_error("dhcpd_invalid_domain", $domain) 
	  if ($domain && (! Cobalt::Netutil::netutil_test_domainname($domain)));
	return MSG_error("dhcpd_invalid_dns") 
	  if ($dns1 && (! Cobalt::Netutil::netutil_test_ipaddr($dns1)));
	return MSG_error("dhcpd_invalid_dns") 
	  if ($dns2 && (! Cobalt::Netutil::netutil_test_ipaddr($dns2)));
	return MSG_error("dhcpd_invalid_gw", $gw) 
	  if ($gw && (! Cobalt::Netutil::netutil_test_ipaddr($gw)));
	return MSG_error("dhcpd_invalid_maxlease", $maxlease) 
	  if ($maxlease && ($maxlease !~ /^\d+$/o));

	open(DHCPnew,"> $Lockdir/dhcpd.conf") 
	  || return MSG_error("dhcpd_cant_update", $!);
	if (-r $Dhcpd_conf) {
	  open(DHCPcnf, "$Dhcpd_conf") || return MSG_error("dhcpd_cant_update", $!);
	  while(<DHCPcnf>) {
	        # All arguments are considered optional...if aren't given an
	        # argument, then we comment out the corresponding line.
	        # We leave the line around as a placeholder so it's easy
	        # to add back in later.
	        if (/^\#?option domain-name\s+/o)
		{
		  print DHCPnew "#" unless ($domain); # comment it out
		  print DHCPnew "option domain-name \"$domain\";\n";
		} elsif (/^\#?option domain-name-servers/o)
		{
		  # if no nameservers are defined, leave it commented out
		  print DHCPnew "#" unless ($dns1);
		  my $dns2str;
		  $dns2str = ", $dns2" if ($dns2);
		  print DHCPnew "option domain-name-servers $dns1$dns2str;\n";
		} elsif (/^\#?option routers /o)
		{
		        print DHCPnew "#" unless ($gw); # comment out
			print DHCPnew "option routers $gw;\n";
		} elsif (/^\#?option subnet-mask /o)
		{
		        print DHCPnew "#" unless ($nm);  # comment out
			print DHCPnew "option subnet-mask $nm;\n";
		} elsif (/^\#?default-lease-time /o)
		{
		        print DHCPnew "#" unless ($maxlease);  # comment out
			print DHCPnew "default-lease-time $maxlease;\n";
		} elsif (/^\#?max-lease-time /o)
		{
		        print DHCPnew "#" unless ($maxlease);  # comment out
			print DHCPnew "max-lease-time $maxlease;\n";
		} elsif (/^\#?option broadcast-address /o)
		{
		        print DHCPnew "#" unless ($broadcast);  # comment out
			print DHCPnew "option broadcast-address $broadcast;\n";
		} elsif (/^subnet ([\d\.]+) netmask ([\d\.]+)/o)
		{
		  # Set subnet parameters based on the new netmask and gateway
		  my $subnet = Cobalt::Network::net_get_network($ipaddr,$nm);
		  print STDERR "Changing dynamic range from subnet $1/$2 to $subnet/$nm\n" if ($Debug);
		  print DHCPnew "subnet $subnet netmask $nm {\n";
		} else
		{
			print DHCPnew;
		}
	      }
	  close(DHCPcnf);
	} else {
	  # Creating a fresh new config file
	  $dns1 .= ", $dns2" if ($dns2);
	  print DHCPnew "# Configuration file for ISC dhcpd\n";
	  print DHCPnew "# Created by the Cobalt user interface\n\n";
	  print DHCPnew "server-identifier ", Cobalt::Network::net_get_hostname(), ";\n\n";
	  print DHCPnew "# common option definitions\n";
	  if ($domain) {
	    print DHCPnew "option domain-name \"$domain\";\n";
	  } else {
	    print DHCPnew "#option domain-name \n";
	  }
	  if ($dns1) {
	    print DHCPnew "option domain-name-servers $dns1;\n"
	  } else {
	    print DHCPnew "#option domain-name-servers\n"
	  }
	  if ($gw) {
	    print DHCPnew "option routers $gw;\n";
	  } else {
	    print DHCPnew "#option routers \n";
	  }
	  if ($nm) {
	    print DHCPnew "option subnet-mask $nm;\n";
	  } else {
	    print DHCPnew "#option subnet-mask \n";
	  }
	  if ($broadcast) {
	    print DHCPnew "option broadcast-address $broadcast;\n";
	  } else {
	    print DHCPnew "#option broadcast-address \n";
	  }
	  if ($maxlease) {
	    print DHCPnew "default-lease-time $maxlease;\n";
	    print DHCPnew "max-lease-time $maxlease;\n";
	  } else {
	    print DHCPnew "#default-lease-time \n";
	    print DHCPnew "#max-lease-time \n";
	  }
	  print DHCPnew "use-host-decl-names on;\n\n";
	}
	close(DHCPnew);

	rename("$Lockdir/dhcpd.conf","$Dhcpd_conf")
	  || return MSG_error("dhcpd_cant_update", $!);
	
	Cobalt::Dhcpd::dhcpd_hup();
	return MSG_ok("dhcpd_ok_updated");
}

sub dhcpd_create_conf
# Create the configuration file for dhcpd server with defaults
#   which are largely from current network parameters
# Arguments: none
# Side Effects: writes $Dhcpd_conf; creates a new one if none exists
# Return value: status string like '4000 S.O.S.'
{
    if( -e $Dhcpd_conf ) { return MSG_error("dhcpd_conf_exist"); }

    my ( $domainName ,$dnsAddr1, $dnsAddr2 )=Cobalt::Network::net_get_dnsclient();
    my ($addr,$mask) = Cobalt::Network::net_get_ipaddr();
    my $gateway=( Cobalt::Network::net_get_gateway() )[ 0 ];

    # Check that the gateway is on the same subnet as the DHCP server
    if ($gateway && $gateway ne '0.0.0.0' &&
	!Cobalt::Netutil::netutil_test_network_in_network($gateway,32,
							  $addr,$mask)) {
      # The gateway is not on eth0's network, use the eth0's address
      # as the gateway.
      $gateway = $addr;
    }

    my $netmask=( Cobalt::Network::net_get_ipaddr() )[ 1 ];

    my $ret1=Cobalt::Dhcpd::dhcpd_set_parameters( $domainName, $dnsAddr1, $dnsAddr2, $gateway, $netmask, $Dhcpd_default_lease );
    if( $ret1!~/^2/ ) { return $ret1; }

    # create null subnet declaration
    my $ret2=Cobalt::Dhcpd::dhcpd_set_range();
    if( $ret2!~/^2/ ) { return $ret2; }

    return $ret1;
}

sub dhcpd_clean_ranges
# Clean out the list of addresses to give out dynamically; should be called
#   after dhcpd_set_parameters().
# Arguments: none
# Return value: a hash table of the ranges which were deleted in the form:
#    %hash = ($low1 => $high1, $low2 => $high2, ...);
# Side effects: calls dhcpd_get_params() to figure out what subnet
#    eth0 is on, and removes all ranges that are not on that subnet;
#    then restarts dhcpd.
{
  # Get the current settings
  my ($gateway,$netmask) = (dhcpd_get_parameters())[3,4];
  my %wiped; # our hash of what ranges we wiped

  return MSG_error("dhcpd_cant_locked") if (-e "$Lockdir/dhcpd.conf");
  open(LOCK,"> $Lockdir/dhcpd.conf") 
    || return MSG_error("dhcpd_cant_update", $!);
  open(CONF, "$Dhcpd_conf") || return MSG_error("dhcpd_cant_update", $!);
  while(<CONF>) {
    if (/range ([\d\.]+) ([\d\.]+);/o) {
      # if we just changed the netmask and/or gateway, we need
      # to clear any dynamic ranges that will no longer work.
      if (Cobalt::Network::net_network_ismember($1,$gateway,$netmask) &&
	  Cobalt::Network::net_network_ismember($2,$gateway,$netmask)) {
	# This range is out of...range.
	$wiped{$1} = $2;
      }
      else {
	# This range still on the same subnet as eth0
	print LOCK;
      }
    } else {
      # Not a range statement, leave it alone
      print LOCK;
    }
  }
  close LOCK;
  close CONF;
  rename("$Lockdir/dhcpd.conf","$Dhcpd_conf") || return MSG_error("dhcpd_cant_update", $!);

  Cobalt::Dhcpd::dhcpd_hup(); # Kick it
  return MSG_ok("dhcpd_ok_updated");
}

# = = Local subroutines = = = = = = = = = = = = = = = = = = = = = = = =

sub dhcpd_hup
# Restart the DHCP daemon
# Arguments: none
# Return value: none
# Side effect: calls the DHCP init script to stop and start the daemon,
#   if the daemon is currently running.
{
  return if (! -e $Dhcpd_pidfile); # Let sleeping dogs lie
  # Oooh, bad...we're not checking our return code.  Thing is,
  # none of the functions that call us look at our return value,
  # so not much we can do without lots of code changes.
  system("$Dhcpdscript restart >>$Sauce_log 2>&1");
}

1;
