#!/usr/bin/perl
#
# Adrian Sun <asun@cobaltnet.com>
# Copyright 1999 Cobalt Networks http://www.cobaltnet.com/
#
# Description:
# this module provides an interface to the partitioning/formatting/mount 
# code. it restricts itself to using "/mnt/disks" for generic partitions 
# and "/vol" for labeled partitions. here's an outline of automated
# disk mangling:
#   foreach $disk (disk_getunpartitioned)
#     disk_partition($disk)
#   endfor
#   foreach $partition (disk_unmounted)
#     if not disk_mount($mount) and disk_mount($partition, "nolabel") 
#         disk_format($partition);
#         disk_mount($partition);
#     endif
#   endfor
#
#   result:
#     the above will auto- partition/format/mount disks in /vol.
#     if there's a conflict but the disk is still useable, it keeps
#     it mounted in /mnt/disks. more advanced options can specify stuff
#     in greater detail.
#
# todo: needs to return different error codes.
#   1) format/partitioning failed
#   2) mountpoint is busy 
#   3) another filesystem is mounted at that mount point.
#
# right now, it returns the same error for all error conditions.
#
package Cobalt::DiskUtils;

use vars qw(@ISA @EXPORT @EXPORT_OK);

require Exporter;
require SelfLoader;
@ISA    = qw(Exporter SelfLoader);
@EXPORT = qw(disk_getpartitions   disk_partition   disk_getunpartitioned
             disk_format          disk_checkpartition  
             disk_getumounts      disk_getmounts   
             disk_mount           disk_umount
             disk_getsignature    disk_addsignature
	     disk_lock            disk_unlock
	     disk_checklock       disk_getlabel
	     disk_mountok
	     );

use Cobalt::Product;
use Cobalt::Util;

1;

__DATA__

## local routines

# these two routines know how mount-2.9o deals with special characters.
sub local_mangle
{
    my $name = shift;

    return $name unless $name =~ /[\\\s]/;
    $name =~ s/([\\\s])/sprintf("\\%0.3o", ord($1))/ge;
    return $name;
}

sub local_unmangle
{
    my $name = shift;
    return $name unless $name =~ /\\/;
    
    $name =~ s/\\(\d\d\d)/pack("c", oct($1))/ge;
    return $name;
}


sub local_gethostid
{
   my ($line,$id);

   open(IFCONFIG, "$Ifconfig_cmd eth0 |") or return undef;
   $id = $1 if ($line = <IFCONFIG> and $line =~ /HWaddr (\S+)/);
   close(IFCONFIG);
   return $id;
}


sub local_df
# description: returns a parsed form of df
#
# inputs: optional partition
#
# output: a hash of either just the partition given or all partitions. this will return undef
#         if there's a problem.
{
    my $partition = shift;

    open (DF, " $Df_cmd -k $partition |") or return undef;
    my %hash = ();
    while (<DF>) {
	if (m"^/dev/") {
	    my @fields = split;
	    my $key = $fields[0];

	    $hash{$key}{Total} = $fields[1];
	    $hash{$key}{Used} = $fields[2];
	    $hash{$key}{Available} = $fields[3];
	    $hash{$key}{PercentUsed} = $fields[4];
	    $hash{$key}{MountPoint} = $fields[5];
	}
    }
    close(DF);
    return %hash;
}

sub local_fixfstab
# description: make sure fstab has an entry for quota usage.
# 
# input: $part, $mountpoint
#
# output: 0 on success. -1 on bad error. 1 on transient error.
{
    my ($part, $mountpoint) = @_;
    my $format = "ext2\tnoauto,usrquota,grpquota,grpid\t0 0";
    my $fstab = "/etc/fstab";
    my $line, $exist;
 
    return -1 unless defined($part) and defined($mountpoint);

    # we need to mangle what the mount entry looks like;
    $mountpoint = local_mangle($mountpoint);

    open(IN, "<$fstab") or return 1;
    while (<IN>) {
	# re-write same entry as before
	if (/^(\S+)\s/ and ($1 eq $part)) {
	    $line .= "$part\t$mountpoint\t$format\n";
	    $exist = 1;
	} elsif (not /\s$mountpoint\s/) {
	    $line .= $_;
	}
    }
    close(IN);
    
    $line .= "$part\t$mountpoint\t$format\n" unless $exist;

    open(OUT, ">$fstab");
    print OUT $line;
    close(OUT);
    return 0;
}

sub local_quotaon
# description: turn quota suppport on. this will fiddle with
# /etc/fstab if necessary.
# 
# input: $part, $mountpoint
#
# output: 0 on success. -1 on error.
{
    my ($part, $mountpoint) = @_;
    my $foundit;

    return -1 unless defined($part) and defined($mountpoint);

    # just see if it works. we might not have to do anything.
    return 0 unless `$Quotaon_cmd $part 2>&1`;
    return -1 if local_fixfstab($part, $mountpoint);
    return `$Quotaon_cmd $part 2>&1` ? -1 : 0;
}


## exported routines

sub disk_checklock
{
    return -e "$Lockdir/diskutils" ? 1 : 0;
}

sub disk_lock
{
    if (not disk_checklock) {
	open(LOCK, ">$Lockdir/diskutils");
	close(LOCK);
	return 0;
    }
    return -1;
}


sub disk_unlock
{
    my $lockfile = "$Lockdir/diskutils";
    unlink($lockfile) if -e $lockfile;
}


sub disk_addsignature
# description: add a ($label, $hostid, $enabled) signature to the disk
# at $mountpoint so that disks can tell the os where they want to be
# mounted.
#
# input: ($mountpoint, $label, $hostid, $enabled)
#
# output $label on success.
{
    my ($mount, $label, $hostid, $enabled, $quotacheck) = @_;

    open(SIGFILE, ">$mount/.cobaltsig") or return 0;
    $label = local_mangle($label); # mangle to handle spaces
    print SIGFILE "LABEL=$label\n";
    print SIGFILE "HOSTID=$hostid\n";
    print SIGFILE "ENABLED=$enabled\n";
    print SIGFILE "QUOTACHECK=$quotacheck\n";
    close(SIGFILE);
    return $label;
}

sub disk_getsignature
# description: returns the signature for a disk.
# 
# input: $mountpoint
#
# output: ($label, $hostid, $enabled, $quotacheck) if successful. undef otherwise.
{
    my $mountpoint = shift;
    my $label;
    my $id;
    my $enabled;
    my $quotacheck;

    return undef unless defined($mountpoint) and 
	open(COBALTSIG, "$mountpoint/.cobaltsig");
    while (<COBALTSIG>) {
      SWITCH: {
	  $id = $1, last SWITCH if m/^HOSTID=(\S+)/;
	  $label = $1, last SWITCH if m/^LABEL=(\S+)/;
	  $enabled = $1, last SWITCH if m/^ENABLED=(\w+)/;
	  $quotacheck = $1, last SWITCH if m/^QUOTACHECK=(\d+)/;
      }
    }
    close(COBALTSIG);
    $label = local_unmangle($label);
    return ($label, $id, $enabled, $quotacheck);
}

sub disk_format
# description: format a partition and add a disk signature.
#
# input: ($partition, $label, optional $flags)
# currently understood flags: check  check for bad blocks    
#
# output: 0 on success, -1 on error.
{
    my ($part, $label, $flags) = @_;
    my $hostid = local_gethostid;
    my $format_cmd = "/sbin/mke2fs -q -m 1";
    my ($mountpoint, $volmount, $opts);
    my $volescaped;

    return -1 unless $part =~ /^\/dev\/(\S+)/;
    $mountpoint = "$Vol_diskdir/$1";
    $label = $1 unless $label;
    $volmount = "$Vol_rootdir/$label";

    $volescaped = $volmount;
    $volescaped =~ s/\\/\\\\/g;

    mkdir($Vol_diskdir, 0755) unless -d $Vol_diskdir;
    mkdir($mountpoint, 0755) unless -d $mountpoint;
    
    # format and touch various files
    $opts = "-c" if $flags =~ /\bcheck\b/;
    `$format_cmd $opts $part >& /dev/null`;
    return -1 if `$Mount_cmd $part $mountpoint 2>&1`;
    mkdir("$mountpoint/sites", 0755) unless -d "$mountpoint/sites";
    mkdir("$mountpoint/spool", 0755) unless -d "$mountpoint/spool";
    unless (-d "$mountpoint/spool/mail") {
	mkdir("$mountpoint/spool/mail", 0775);
	chmod(0775, "$mountpoint/spool/mail"); 
	chown(-1, (getgrnam($MailGroup))[2], "$mountpoint/spool/mail");
    }
    mkdir($volmount, 0755) unless -d $volmount;

    `$Umount_cmd $mountpoint`;
    # now add in quota support
    local_fixfstab($part, $volmount);
    `$Mount_cmd $volescaped`;
    `$Quotacheck_cmd $part >& /dev/null`; 

    # finally, add disk signature
    disk_addsignature($volmount, $label, $hostid, "yes", 0);

    `$Umount_cmd $volescaped`;
    return 0;
}

sub disk_partition
# description: write the partition table. this will over-write any
# existing partition table. in addition, it will only set up the
# primary partitions.
#
# input: $disk (/dev/[sh]d[a-z])
#
# output: 0 on success, -1 on failure.
{
    my $disk = shift;
    my $count; 
    my $string;

    return -1 unless $disk =~ m"^/dev/\w+[\D]";
      
    open(SFDISK, "| $Sfdisk_cmd -uM $disk >& /dev/null") or return -1;
    while ($count++ < 4) { # we only handle primary partitions
	my $size = shift;
	print SFDISK defined($size) ? ",$size\n" : ";\n";
    }
    print SFDISK "y\n"; # write the partition table
    close(SFDISK);
    return $?;
}

sub disk_checkpartition
# description: checks to make sure everything with the disk partition
#              is all right. right now, this just does an fsck. quotacheck
#              really only works on mounted filesystems.
# 
# input: $partition + optional args.
#
# output: return non-zero exit status if something went wrong.
{
    my ($partition, $args) = @_;
    my $fsck_cmd = "/sbin/e2fsck -a";

    return -1 unless defined($partition);
    
    # return fsck's error codes. error codes 0, 1, and 2 are okay.
    `$fsck_cmd $args $partition >& /dev/null`;
    return ($? and (($? >> 8) & 0xFC)) ? ($? >> 8) : 0;
}


sub disk_umount
# description: unmount a disk.
# 
# input: $mountpoint, optional $flags
# flags: disable will disable the disk
#
# output: a string if the umount failed. nothing otherwise.
{
    my $mountpoint = shift;
    my $flags = shift;

    return "no mountpoint" unless $mountpoint;

    my ($label, $hostid, $enable) = disk_getsignature($mountpoint);
    if ($flags =~ /\bdisable\b/) {
	$enable = "no";
    }
    disk_addsignature($mountpoint, $label, $hostid, $enable, 0);
    $mountpoint =~ s/\\/\\\\/g;
    `$Umount_cmd $mountpoint 2>&1`;
}

sub disk_mount
# description: mount a disk or nfs filesystem.
#
# input: ($partition, $flags, $label)
# available flags: noenable, disable, nocheck, ro, enable, override
#
#
# output: $mountpoint
{
    my ($part, $flags, $label) = @_;
    my $hostid = local_gethostid;
    my ($mountpoint, $mountescaped);
    my $nfs;
    my $enabled;
    my $id;
    
    return undef unless defined($part);
    return undef unless ($part =~ m=^/dev/(\S+)$= or 
			 ($nfs = ($part =~ m=\S+?:/[\S+/](\S+)$=)));
    $mountpoint = "$Vol_diskdir/$1";
    $label = $1 unless $label;

    # make sure the partition is happy
    return undef unless ($nfs or ($flags =~ /\bnocheck\b/) or
			 (not disk_checkpartition($part)));

    # make sure some directories exist.
    mkdir ($Vol_diskdir, 0755) unless -d $Vol_diskdir;
    mkdir ($Vol_rootdir, 0755) unless -d $Vol_rootdir;
    mkdir ($mountpoint, 0755) unless -d $mountpoint;

    my $mountopts = "-o ro" if $flags =~ /\bro\b/; 
    return undef if `$Mount_cmd $mountopts $part $mountpoint 2>&1`;

    if ($nfs) { # handle nfs
	`$Umount_cmd $mountpoint`;
	$mountpoint = "$Vol_rootdir/$label";
	mkdir($mountpoint, 0755) unless -d $mountpoint;
	$mountescaped = $mountpoint;
	$mountescaped =~ s/\\/\\\\/g;
	return undef if `$Mount_cmd $mountopts $part $mountescaped 2>&1`;
	return $mountpoint;
    }

    # handle noenable option
    return $mountpoint if $flags =~ /\bnoenable\b/;

    # the override option. it's useful if we decide to remount disks
    # in a different location.
    if ($flags =~ /\boverride\b/) { 
	$enabled = ($flags =~ /\bdisable\b/) ? "no" : "yes";
	`$Umount_cmd $mountpoint`;
	$mountpoint = "$Vol_rootdir/$label";
	mkdir ($mountpoint, 0755) unless -d $mountpoint;
	$mountescaped = $mountpoint;
	$mountescaped =~ s/\\/\\\\/g;
	return undef if `$Mount_cmd $part $mountescaped`;
	disk_addsignature($mountpoint, $label, $hostid, $enabled, 1);
	return $mountpoint unless local_quotaon($part, $mountpoint);
	disk_addsignature($mountpoint, $label, $hostid, $enabled, 0);
	`$Umount_cmd $mountescaped`;
	return undef;
    }

    ($label, $id, $enabled, $quotacheck) = disk_getsignature($mountpoint);

    if ($enabled eq "no" and $flags =~ /\benable\b/) {
	$enabled = "yes";
    }

    if (not defined($label)) {
	`$Umount_cmd $mountpoint`;
	return undef;
    }

    if ($id ne $hostid) {
	`$Umount_cmd $mountpoint`;
	return undef;
    }

    if ($enabled eq "yes") {
	`$Umount_cmd $mountpoint`;
	$mountpoint = "$Vol_rootdir/$label";
	$mountescaped = $mountpoint;
	$mountescaped =~ s/\\/\\\\/g;
	mkdir ($mountpoint, 0755) unless -d $mountpoint;
	return undef if `$Mount_cmd $mountopts $part $mountescaped 2>&1`;
	if ($flags !~ /\bro\b/) {
	    `$Quotacheck_cmd $part >& /dev/null` unless 
	    ($flags =~ /\bnocheck\b/ or (not $quotacheck));
	    local_quotaon($part, $mountpoint);
	}
    }

    disk_addsignature($mountpoint, $label, $id, $enabled, 1);
    return $mountpoint;
}

sub disk_getpartitions
# description: get all the disk partitions
#
# input: none
#
# output: hash with partitions as keys and size (in bytes) as values
{
    my $partition = shift;
    my %hash;

    open(SFDISK, "$Sfdisk_cmd -uB -l $partition |") or return undef;

    # parse sfdisk
    while (<SFDISK>) {
	if (m"^Disk (\S+?): (\d+) heads, (\d+) sectors, (\d+) cylinders") {
	    $hash{$1} = $2 * $3 * $4 / 2;
	} elsif (m"^(/dev/\S+)\s+[ *]\s+\d+[ +-]\s+\d+[ +-]\s+(\d+)[ +-]\s+83") {
            $hash{$1} = $2;
	}
    }
    close (SFDISK);

    return %hash;
}

sub disk_getmounts
# description: get mounted volumes
#
# optional input: ($label) 
# $label: return anything that matches $label
#
# output: hash with partition as keys and {Size, MountPoint, Used, Free} 
#         as values.
{
    my $label = shift;
    my %df = local_df;
    my $size = $BytesPerBlock;
    my %hash = ();
    my $key;

    foreach $key (keys %df) {
	next if ($label and $df{$key}{MountPoint} !~ $label);
	$hash{$key} = {
	    Size => $df{$key}{Total}, 
	    MountPoint => $df{$key}{MountPoint},
	    Used => $df{$key}{Used},
	    Free => $df{$key}{Available}
	}
    }
    return %hash;
}

sub disk_getumounts 
# description: get unmounted volumes
#
# input: none
#
# output: hash of unmounted volumes with partition as key and size (in
#         bytes) as value.
{
    my %partitions = disk_getpartitions;
    my %mounts = disk_getmounts;
    my $key;
    my $value;

    my %hash = ();
    while (($key, $value) = each %partitions) {
	next if $key =~ m/\D$/;
	$hash{$key} = $value unless $mounts{$key};
    }
    return %hash;
}

sub disk_getunpartitioned
# description: unpartitioned disks
#
# input: none
#
# output: hash of all the unpartitioned disks (key: disk partition,
#         value: size in bytes)
{
    my %partitions = disk_getpartitions;
    my @keys = keys %partitions;
    my $list = join(",", @keys);
    my %hash = ();
    my $key;
    
    foreach $key (@keys) {
	$hash{$key} = $partitions{$key} if $key =~ m/\D$/ and 
	    $list !~ m/$key\d+/;
    }
    return %hash;
}


sub disk_getlabel
# description: convert a mountpoint to a disk label
#              this needs to be able to deal with spaces.
#
# input: a path
# output: the label for the disk. 
#
{
    my $dir = shift; 

    $dir = readlink($dir) if -l $dir;
    $dir .= "/";  # stick on a trailing /
    if ($dir =~ /^\/$Vol_home\//o) {
	return $Vol_home;
    } elsif ($dir =~ /^$Vol_rootdir\/(\S+?)\//o) {
	return $1 if (-e "$Vol_rootdir/$1/.cobaltsig");
    }

    return undef;
}


sub disk_mountok
# description: return whether or not a mount point is okay to use.
#              this just makes sure that the mountpoint is 
#              writeable and not an nfs mount.    
# input: a path
#
{
    my $path = shift;
    my %df = local_df($path);
    my @keys = keys %df;

    return 0 unless @keys;
    mkdir($path, 0755) unless -d $path;
    return 0 unless -w $path;
    return 1;
}
