# Routines for managing files and directories
#
# Written by Brion Moss (brion@cobaltnet.com, brion@nda.com)
#
# Copyright 1998 Cobalt Networks http://www.cobaltnet.com
#

package Cobalt::Fileman;
use vars qw(@ISA @EXPORT @EXPORT_OK);

require Exporter;
require SelfLoader;

@ISA    = qw(Exporter SelfLoader);
@EXPORT = qw(file_delete file_move file_copy file_mkdir file_list
	     file_rights);

require Cobalt::Utility;
import Cobalt::Utility qw(:DEFAULT);
use Cobalt::Util;
use Cobalt::Product;

1;
__DATA__

sub file_mkdir
# Make a directory, after checking that the user is allowed to do so
# Arguments: new directory name, name of user making the directory
# Side effects: creates the directory and any parents
# Example: Cobalt::Fileman::file_mkdir("/home/httpd/html/users/joe/new_folder","joe");
{
    my ($dir,$username) = @_;
    $dir =~ /^(.+)\/[^\/]+/o;  # find the parent
    my $parent = $1;

    return MSG_error("fileman_cant_dir_exists", $dir) if (-e $dir);
    # Check that I'm cool to touch this directory
    return MSG_error("fileman_cant_user_nowrite", $username) 
	unless (Cobalt::Fileman::file_rights($username,$parent) == 1);

    Cobalt::Fileman::file_mkdir($parent,$username) unless (-e $parent);
    my ($uid,$gid) = (stat(_))[4,5];
    if ($dir =~ /^$Htgroup_dir/o)
    {
	# it's in the group area
	mkdir($dir,02575)
	    || return MSG_error("fileman_cant_make_group_subdir", $dir);
	chown($uid,$gid,$dir)
	    || return MSG_error("fileman_cant_set_id", $dir);
	chmod(02575,$dir)
	    || return MSG_error("fileman_cant_set_mode", $dir);
    }
    else
    {
	mkdir($dir,0755)
	    || return MSG_error("fileman_cant_mkdir", $dir);
	chown($uid,$gid,$dir)
	    || return MSG_error("fileman_cant_set_id", $dir);
	chmod(0755,$dir)
	    || return MSG_error("fileman_cant_set_mode", $dir);
    }
    return MSG_ok("fileman_ok_mkdir");
}

sub file_copy
# Copy a file, after checking that the user has rights to read the source
# and write the dest
# Arguments: source, dest, user name
# Side effects: copies the file
# Example: Cobalt::Fileman::file_copy("/home/httpd/html/users/lola/oldfile",
#                    "/home/httpd/html/users/lola/newfile","lola");
{
    my ($src,$dest,$username) = @_;

    # Check that I'm cool to touch these files
    return MSG_error("fileman_cant_user_noread", $username) 
	unless (Cobalt::Fileman::file_rights($username,$src));
    if (-e $dest)
    {
	return MSG_error("fileman_cant_user_noclobber", $username) 
	    unless (Cobalt::Fileman::file_rights($username,$dest) == 1);
    }
    else
    {
	# new file, check the directory
	$dest =~ /^(.+)\/[^\/]+/o;  # find the parent
	return MSG_error("fileman_cant_user_nowrite_dir", $username) 
	    unless (Cobalt::Fileman::file_rights($username,$1) == 1);
    }

    if (-d $src)
    {
	Cobalt::Fileman::file_copydir($src,$dest);
    }
    elsif (-f $src)
    {
	Cobalt::Fileman::careful_cp($src,$dest);
    }
    return MSG_ok("fileman_ok_file_copied");
}


sub file_move
# Move or rename a file
# First verifies that the user has permission to do so
# Arguments: source, dest, username
# Side effects: moves or copies the file
# Example: Cobalt::Fileman::file_move("/home/httpd/html/users/lola/oldfile",
#                    "/home/httpd/html/users/lola/newfile","lola");
{
    my ($src,$dest,$username) = @_;

    return MSG_error("fileman_cant_move_src", $username) 
	unless (Cobalt::Fileman::file_rights($username,$src) == 1);
    if (-e $dest)
    {
	return MSG_error("fileman_cant_user_noclobber", $username) 
	    unless (Cobalt::Fileman::file_rights($username,$dest) == 1);
    }
    else
    {
	# new file, check the directory
	$dest =~ /^(.+)\/[^\/]+/o;  # find the parent
	return MSG_error("fileman_cant_user_nowrite_dir", $username) 
	    unless (Cobalt::Fileman::file_rights($username,$1) == 1);
    }

    if (-d $src)
    {
	Cobalt::Fileman::file_copydir($src,$dest);
	Cobalt::Utility::sauce_deltree($src);
    }
    elsif ( (stat($src))[0] == (stat($dest))[0] )
    {
	# On the same devince
	rename($src,$dest);
    }
    else
    {
	# crossing a filesystem
	Cobalt::Fileman::careful_cp($src,$dest);
	unlink($src);
    }
    return MSG_ok("fileman_ok_file_moved");
}

sub file_delete
# Deletes a file or directory, after checking that the user has
# the right to delete it
# Arguments: name of file or directory to delete
# Side Effects: removes the file or directory tree
# Example: Cobalt::Fileman::file_delete("/home/httpd/html/groups/queens/somefile","liz");
{
    my ($tgt,$username) = @_;

    return MSG_error("fileman_cant_user_nodelete", $username) 
	unless (Cobalt::Fileman::file_rights($username,$tgt) == 1);

    if (-d $tgt)
    {
      Cobalt::Utility::sauce_deltree($tgt);
    }
    else
    {
	unlink($tgt);
    }
    return MSG_ok("fileman_ok_file_deleted");
}

sub file_list
# Get a list of files in a directory
# Checks to make sure the user has the right to look at those files
# Arguments: the directory name, the name of the user looking around
# Return value: a list of files
# Side effects: none
# Example: @files = Cobalt::Fileman::file_list("/home/httpd/html/users","kelly");
{
    my ($dir,$username) = @_;

    # Check that I'm cool to see these files
    return MSG_error("fileman_cant_user_no_read_dir", $username) 
	unless (Cobalt::Fileman::file_rights($username,$dir));

    my (@files);
    opendir(DIR,$dir) 
	|| return Cobalt::List::abort(MSG_error("fileman_cant_read_dir", $dir));
    @files = grep(!/^\./o,readdir(DIR));  # don't show dotfiles
    closedir DIR;
    return @files;
}

sub file_rights
# Does a user have rights to a file or directory?
# Arguments: user name, file or directory name (full path)
# Return value: 0 for no permissions, -1 for read only, 1 for read/write
#   For our purposes, there will never be write perms w/o read perms
#   An error condition (bad username, etc) means no permissions
#   No accesses outside the web publishing area are allowed.
#   The admin user is always allowed (within the web area)
# Side effects: none
# Example: &panic unless (Cobalt::Fileman::file_rights("ken","/home/httpd/html/users/ken"));
{
    my ($user,$dir) = @_;
    my $uid = getpwnam($user);
    return 0 unless ($uid >= $Firstuid);  # no uid, no permee
    return 0 unless ($dir);               # no permission to nothing

    if ((! -e $dir) && (! -l $dir))
    {
	# We're looking to create something where it doesn't exist
	$dir =~ /^(.+)\/([^\/]+)$/o;
	$dir = $1;
	print STDERR "Cobalt::Fileman::file_rights() is recursing into parent $dir\n" 
	    if ($Debug);
	return Cobalt::Fileman::file_rights($user,$dir);
    }
    my ($dir_dev,$dir_mode,$dir_uid,$dir_gid) = (stat($dir))[0,2,4,5];

    # They shouldn't be playing with anything outside the web server area
    # We check the device number because they can try to mislead us with ..'s.
    if ((($dir !~ /^$Htusers_dir/o) &&
	($dir !~ /^$Htgroup_dir/o)) ||
	($dir_dev !~ (stat($Htusers_dir))[0]))
    {
	return 0;
    }

    # The admin gets whatever the admin wants
    return 1 if ($user eq $Adminuser);

    # first check against ownership (UID match has top precedence)
    if ($dir_uid == $uid)
    {
	if    ($dir_mode & 0200) { return 1;  }  # write perms, allow anything
	elsif ($dir_mode & 0400) { return -1; }  # read perms
	else                     { return 0;  }  # no perms
    }
    
    # Now check against groupship
    my $members = (getgrgid($dir_gid))[3];
    if ($members =~ /\b$user\b/)
    {
	if    ($dir_mode & 020) { return 1;  }  # write perms, allow anything
	elsif ($dir_mode & 040) { return -1; }  # read perms
	else                    { return 0;  }  # no perms
    }

    # Finally, check against worldship
    if    ($dir_mode & 02) { return 1;  }  # write perms, allow anything
    elsif ($dir_mode & 04) { return -1; }  # read perms
    else                   { return 0;  }  # no perms
    
}

#################################


sub file_copydir
# Copies files from the source to the destination.  If the files
# are the same, don't copy.
# Arguments: a source directory and a destination directory
# Side effects: copies any files which appear to have changed
#   or which don't exist in the destination area.  Does not delete
#   anything.
{
    my ($src,$dst) = @_;
    my ($file,$atime,$mtime,$uid,$gid,$mode);
    unless (-d $src)
    {
	return;
    }
    mkdir($dst,(stat($src))[2])	unless (-d $dst);
    unless (opendir(DIR,$src))
    {
	return;
    }
    my @files = sort(grep(!/^\.\.?$/o,readdir(DIR)));
    closedir DIR;
    foreach $file (@files)
    {
	if ((-d "$src/$file") && (! -l "$src/$file"))
	{
	    # recurse
	    Cobalt::Fileman::file_copydir("$src/$file","$dst/$file");
	    ($atime,$mtime,$mode,$uid,$gid) = (stat("$src/$file"))[8,9,2,4,5];
	    utime($atime,$mtime,"$dst/$file");
	    chmod($mode,"$dst/$file");
	    chown($uid,$gid,"$dst/$file");
	}
	elsif ((! -f "$src/$file") && (! -l "$src/$file"))
	{
	    next;
	}
	else
	{
	    # replace with a fresh copy
	    unlink ("$dst/$file") if (-e "$src/$file");
	    Cobalt::Fileman::careful_cp("$src/$file","$dst/$file");
	}
    }    
}

sub careful_cp
# Copy a file.  Only handles plain files and symlinks
# Arguments: source file, destination file
# Side effects: makes the destination file a duplicate of the source
#   file, and sets the mode, ownership, and timestamp to be the same.
{
    my ($from,$to) = @_;
    my ($offset,$buf,$blksize,$len,$written,$atime,$mtime,$uid,$gid,$mode);
    if (-l $from)
    {
	# it's a symlink
	my $linksrc = readlink($from);
	unless (symlink($linksrc,$to))
	{
	    # Can't link, try to copy
	    Cobalt::Fileman::careful_cp($linksrc,$to);  # The OS will save us from loops
	}
    }
    elsif (! -f $from)
    {
	return;
    }
    else
    {
	# it's a file
	open(FROM,$from) || return MSG_error("fileman_cant_read_source", $from);
	open(TO,">$to") 
	    || return MSG_error("fileman_cant_write_target", $to);
	# this block of code is ripped right out of the perl book
	$blksize = (stat FROM)[11] || 16384;
	$buf = "";
	while ($len = sysread FROM, $buf, $blksize)
	{
	    if (! defined $len)
	    {
		next if ($! =~ /^Interrupted/o);
		return MSG_error("fileman_cant_syserr_read", $from);
	    }
	    $offset = 0;
	    while ($len)  # handle partial writes
	    {
		$written = syswrite TO, $buf, $len, $offset;
		return MSG_error("fileman_cant_syserr_write", $to)
		    unless (defined $written);
		$len -= $written;
		$offset += $written;
	    }
	}
	# end rip
	close FROM;
	close TO;

	# fix timestamp and perms
	($atime,$mtime,$mode,$uid,$gid) = (stat($from))[8,9,2,4,5];
	utime($atime,$mtime,$to);
	chmod($mode,$to);
	chown($uid,$gid,$to);
    }
}

1;
