# $Id: Ftp.pm,v 1.71 1999/10/02 16:44:26 duncan Exp $
#
# ProFTPd backend controls "MajorFTP"
#
# Written by Will DeHaan 
#   will@cobaltnet.com
#
# Copyright 1998 Cobalt Networks http://www.cobaltnet.com/
#

package Cobalt::Ftp;
use vars qw(@ISA @EXPORT @EXPORT_OK);

require Exporter;
require SelfLoader;

@ISA    = qw(Exporter SelfLoader);
@EXPORT = qw(ftp_set_vinterface ftp_del_vinterface ftp_list_config
             ftp_config ftp_get_anonymous);

require Cobalt::Netutil;
import Cobalt::Netutil qw(:DEFAULT);
require Cobalt::Network;
import Cobalt::Network qw(:DEFAULT);
require Cobalt::Group;
import Cobalt::Group qw(:DEFAULT);
require File::Copy;
import File::Copy qw(:DEFAULT);

use Cobalt::Util;
use Cobalt::User;
use Cobalt::Product;

use strict;

1;
__DATA__

sub ftp_set_vinterface
# Adds IP-bound virtual network interface support to proftpd
# Arguments: IP address
# Return value: String of the form MSG_error("IIS fouled my couch")
# Site effects: Edits proftpd.conf
# Example: $return=ftp_set_vinterface($newip);
{
    my($ipaddr)=@_;
    return ftp_config($ipaddr,1,0,'',0,0);
}

sub ftp_del_vinterface
# Removes virtual network interface configuration from proftpd
# Arguments: IP address
# Return value: String of the form MSG_ok("Now that's some good tequila")
# Side effects: Edits proftpd.conf
# Example: $return=ftp_del_vinterface($deletedip);
{
    my($ipaddr)=@_;
    return ftp_config($ipaddr,0,0,'',0,0);
}

sub ftp_config
# Edit ftp configuration
# Arguments: IP, update/delete, anonymous on/off, site group name, 
#                  maximum number of users, incoming quota   
# Return value: string of the form MSG_error("Ouch, stop it!")
# Example: Bind to new address: $return=ftp_config("10.9.25.79",1); 
#   Enable anon ftp: $return=ftp_config("10.9.25.79",1,1,'home',10,50); 
{
    my($ip,$status,$anon,$group,$maxusers,$quota)=@_;
    my($ftpuser)= uc($group); # Anonymous ftp username
    $quota ||= 0; # No incoming dir if none specified
    
    # Basic data verification 
    return MSG_error("ftp_cant_disable_main") if (($group eq $MainGroup) && !$status);
    return MSG_error("ftp_invalid_maxusers", $maxusers) if ($maxusers =~ /\D/o);
    $maxusers = $FTPUsersMax unless ($maxusers =~ /^\d+$/o);
    return MSG_error("ftp_invalid_ipaddr", $ip) if (!netutil_test_ipaddr($ip));

    # Group name validity is only necessary for anonymous ftp
    return MSG_error("ftp_invalid_groupname", $group) if ($anon && !group_exist($group));

    # Build default config data
    my($anonConfig,$siteConfig);
    if($anon)
    {
        $anonConfig=<<END;
	<Anonymous $Htgroup_dir/$group/ftp>
	  User                          $ftpuser
	  Group                         ftp
	  UserAlias                     anonymous $ftpuser
	  UserAlias                     guest $ftpuser
	  MaxClients                    $maxusers
	  <Directory *>
	    <Limit WRITE>
	      DenyAll
	    </Limit>
	  </Directory>
	  <Directory $Htgroup_dir/$group/ftp/incoming>
            Umask			002
            <Limit STOR>
              AllowAll
            </Limit>
            <Limit READ>
              DenyAll
            </Limit>
	  </Directory>
	</Anonymous>
END
    }
    $siteConfig=<<END;
<VirtualHost $ip>
	DefaultRoot		/ admin
	DefaultRoot		~/../.. $Sites_AdminGroup
	DefaultRoot		~ !$Sites_AdminGroup
	AllowOverwrite		on
	DisplayLogin		.ftphelp
$anonConfig</VirtualHost>
END
    $siteConfig=$anonConfig if ($group eq $MainGroup); # Minor special treatment

    # Read existing ftp configuration
    my(%ftpstatus)=ftp_list_config();

    # Disallow anon ftp if a different virtual site
    #   is using this IP for anonymous ftp
    my($siteIP,$modify);
    foreach $siteIP (keys %ftpstatus)
    {
        return MSG_error("ftp_cant_ipalreadyinuse") if ($ftpstatus{$ip} && $anon && ($ftpstatus{$ip} ne $group));
	$modify=1 if ($siteIP eq $ip);
    } 

    # Bail if we're deleting an undefined site
    return MSG_ok("ftp_ok_vsite_undefined") if (!$modify && !$status);

    # Test/add anonymous ftp user
    if (!$anon) {
	my($ret)=Cobalt::User::user_del($ftpuser) if (Cobalt::User::user_exist($ftpuser));
    } else {
	my($ret)=Cobalt::User::user_add($ftpuser,"$group anonymous FTP user","*",$quota,$Badshell,$group,"$Htgroup_dir/$group/ftp") unless (Cobalt::User::user_exist($ftpuser));
    }

    # Verify/create deletable ftp directories
    if($group && $anon) # group is not optional if anonymous ftp is not being enabled
    {
        mkdir("$Htgroup_dir/$group/ftp",02775) unless (-d "$Htgroup_dir/$group/ftp");
        chown((getpwnam($WWWuser))[2],(getgrnam($group))[2],"$Htgroup_dir/$group/ftp");
        chmod(02775,"$Htgroup_dir/$group/ftp"); # Always reset perms
        mkdir("$Htgroup_dir/$group/ftp/incoming",02770) unless (-d "$Htgroup_dir/$group/ftp/incoming");
	chown((getpwnam($WWWuser))[2],(getgrnam($group))[2],"$Htgroup_dir/$group/ftp/incoming");
    }

    # Set incoming folder permissions
    if ($quota && $anon)
    {
        chmod(02773,"$Htgroup_dir/$group/ftp/incoming");
        my($ret)=Cobalt::User::user_mod($ftpuser,'','',$quota);
        return $ret unless ($ret =~ /^2/);
    } else 
    {
        chmod(02770,"$Htgroup_dir/$group/ftp/incoming");
    }

    # Strict locks
    if (-e "$Lockdir/proftpd.conf")
    { 
       sleep 1; # Avoid race
       return MSG_error("ftp_cant_server_locked") if (-e "$Lockdir/proftpd.conf");
    }
    open(FTPD,"$Ftpd_conf") || return MSG_error("ftp_cant_read_config", $Ftpd_conf);
    open(NEWCNF,"> $Lockdir/proftpd.conf") || return MSG_error("ftp_cant_write_ftpconf", $!);
    while(<FTPD>)
    {
        if (/<VirtualHost $ip>/ ... /<\/VirtualHost>/)
        { # Ignore data in $siteconfig...we'll be re-adding this soon enough
        } elsif (($group eq $MainGroup) && (/<Anonymous $Htgroup_dir\/$MainGroup\// ... /<\/Anonymous>/))
        { # $MainGroup does not appear in VirtualHost block
        } else 
        { # Worth keeping
            print NEWCNF;
        }
    }
    print NEWCNF $siteConfig if ($status);
    close(NEWCNF);
    close(FTPD);

    # Clobber config
    move("$Lockdir/proftpd.conf","$Ftpd_conf") || return MSG_error("ftp_cant_write_ftpconf", $!);
        
    return MSG_ok("ftp_ok_confupdated");
}

sub ftp_get_anonymous
# Returns 1 if anonymous ftp is enabled or 0 if it isn't,
#   maximum number of anonymous users, incoming quota (blocks)
# Arguments: site group name
# Return value: array: (boolean of whether anonymous FTP is on, maximum 
#   number of anonymous users, quota for the anonymous FTP user)
{
    my($group) = $_[0];

    my $on=0;
    my $maxUsers=0;
    if (-r "$Ftpd_conf")
    {
      open(FTPD,"$Ftpd_conf")
	|| return MSG_error("ftp_cant_read_config", $Ftpd_conf);
      while(<FTPD>)
      {
        if(/<Anonymous $Htgroup_dir\/$group\/ftp/) { $on=1; }
        if($on)
        {
	  if(/MaxClients\s+(\S+)/)
          {
            $maxUsers=$1;
            last;
	  }
	  if(/<\/Anonymous>/) { last; }
        }
      }
      close(FTPD);
    }

    my($ftpuser)=uc($group);
    return ($on,$maxUsers,(Cobalt::User::user_show($ftpuser))[2]);
}

sub ftp_list_config
# Abstracts a list of supported IP addresses and associated anonymous ftp 
# sites.  
# Arguments: none
# Return value: hash of the form IP->site group anonymous ftp or 0
# Example %ftpstatus=ftp_list_config();
#         foreach $ip (keys %ftpstatus) { $anongroup=$ftpstatus{$ip}; }
{
    my($group,%ftp);
    my($ipaddr)=(Cobalt::Network::net_get_ipaddr())[0]; # Home site first
    $ftp{$ipaddr}=0;
    my($ip)=$ipaddr; # Init var

    if (-r "$Ftpd_conf")
    {
        open(FTPD,"$Ftpd_conf")
          || return MSG_error("ftp_cant_read_config", $Ftpd_conf);
        while(<FTPD>)
        {
	    if (/^<VirtualHost [\d\.]+>$/ ... /^<\/VirtualHost>$/)
	    {
		if (/^<VirtualHost ([\d\.]+)>$/) {
		    $ip=$1;
		    $ftp{$ip}=0;
		}
		if (/^\s*<Anonymous\s+$Htgroup_dir\/([^\/]+)/o) {
		    $ftp{$ip}=$1 if ($1 ne $MainGroup);
		}
	    } elsif (/^\s*<Anonymous\s+$Htgroup_dir\/$MainGroup/) {
		$ftp{$ipaddr}=$MainGroup;
	    }
	}
        close(FTPD);
    }
    return(%ftp);
}

1;
