 # Package for managing virtual mailing lists
#
# $Id: List.pm,v 1.67 1999/10/06 10:06:27 duncan Exp $
#
# Written by Brion Moss (brion@cobaltnet.com, brion@nda.com)
# Unwritten and adapted for virtual majordomo MLM
#   by Sridhar Gopal (gsri@cobaltnet.com)
# fixed by Duncan Laurie <duncan@cobaltnet.com>
#
# Copyright 1998-99 Cobalt Networks,Inc.
# http://www.cobaltnet.com
#

package Cobalt::List;

require Exporter;
require SelfLoader;
@ISA    = qw(Exporter SelfLoader);
@EXPORT = qw(alias_get_bytarget
	     alias_exist alias_get alias_get_vacationless
             alias_set alias_add alias_delete
	     list_virtualize
	     list_exist list_create list_delete list_modify
	     list_is_mem list_get_mem list_set_mem
	     list_add_mem list_del_mem
	     list_lists list_lists_mem list_lists_del_mem
	     list_config_open list_config_closed
	     list_config_create list_config_password list_config_getpassword
	     list_config_getrestrict list_config_restrict list_config_norestrict
	     list_create_virtual list_delete_virtual list_modify_virtual
	     );

use Cobalt::Product;
require Cobalt::Utility;
import Cobalt::Utility qw(:DEFAULT);
require Cobalt::Email;
import Cobalt::Email qw(:DEFAULT);
require Cobalt::Netutil;
import Cobalt::Netutil qw(netutil_get_fulldomain);
require Cobalt::Vsite;
import Cobalt::Vsite qw(vsite_get_bygroup);

use Cobalt::Util;
use File::Copy;
use integer;
use strict;

1;

__DATA__

sub alias_get_bytarget
# Returns all the aliases which go to a given target
#   In other words, if the aliases file looks like:
#     ron.decline: ron
#     rdecline: ron
#   then alias_get_bytarget("ron") would return ("ron.decline","rdecline");
# Arguments: recipient address
# Return value: an array of (alias1, alias2, ...)
# Note: group aliases all go into the majordomo aliases file, so we don't
#   need to filter those out.  The only alias we could collide with would
#   be forwarding aliases.
{
    my ($target) = @_;
    my @list;
    "joe: $target" =~ /^(.+):\s*$target\s*$/;  # precompile the regexp
    open (ALIAS, "$Aliasfile") or return (MSG_error("list_cant_read_aliases", $Aliasfile));
    while ( <ALIAS> ) {
      if (//) {
	next if (Cobalt::User::user_exist($1));  # skip forwarding aliases
	push(@list,$1);
      }
    }
    close ALIAS;
    return @list;
}

sub alias_exist
# Does the given alias exist?
# Arguments: alias name
# Return value: true if it does, false if it doesn't
{
    my ($alias) = @_;
    return 1 if (alias_exist_in_file($alias, $Aliasfile));
    return alias_exist_in_file($alias, $Majordomo_aliases);
}

sub alias_exist_in_file
# Does the given alias exist in this file?
# Arguments: alias name, file name
# Return value: true if it does, false if it doesn't
{
    my ($alias, $file) = @_;
    return 0 if (! -e $file);
    open(ALIAS, $file) or return(MSG_error("list_cant_read_alias_file", $file));
    "$alias:" =~ /^$alias:/; # compile once per function call
    while (<ALIAS>) {
	(close ALIAS, return 1) if (//);
    }
    close ALIAS;
    return 0;
}

sub alias_get
# Returns all the aliases for an address
# Arguments: address
# Return value: an array of (member1, member2...)
{
    my ($user) = @_;
    my @list = alias_get_from_file($user, $Aliasfile);
    return $list[0] if ($list[0] =~ /^[134]\S*\s/);
    return @list unless ($#list < 0);
    return alias_get_from_file($user, $Majordomo_aliases);
}

sub alias_get_vacationless
# Returns all the aliases for an user but for vacation and \$user
# Arguments: alias name
# Return value: an array of (member1, member2...)
{
    my ($user) = @_;
    my @aliases = alias_get_from_file($user, $Aliasfile);
    return $aliases[0] if ($aliases[0] =~ /^[134]\S*\s/);
    return @aliases unless grep /$Vaca_prog/o, @aliases;
    @aliases = grep !/$Vaca_prog/o, @aliases;
    return grep !/^\\$user$/, @aliases;
}

sub alias_get_from_file
# Returns all the aliases for an address in a file
# Arguments: address, file name
# Return value: an array of (member1, member2...)
{
    my ($user, $file) = @_;
    my @list;
    open (ALIAS, "< $file") or return (MSG_error("list_cant_read_alias_file", $file));
    while ( <ALIAS> ) {
	@list = split(/,\s*/,$1) if (/^$user:\s*(.+)$/);
    }
    close ALIAS;
    return @list;
}

sub alias_set
# Update aliases of the given user with the provided list
# Arguments: user name, new array of aliases
# Return value: sauce return code/string
{
    my ($user, @aliases) = @_;
    return MSG_error("list_invalid_adding_self") if (grep(/^$user$/,@aliases));
    my $lock;
    return $lock if (($lock = lock_aliases()) !~ /^2/);
    open(ALIAS, $Aliasfile) or abort(MSG_error("list_cant_read_aliases", $Aliasfile));
    open (ALIAS_NEW, "> $Lockdir/aliases") or
	abort(MSG_error("list_cant_write_aliases", $Lockdir));
    "$user:" =~ /^$user:/; # compile once per function call
    while (<ALIAS>) {
	print ALIAS_NEW unless (//);
    }
    close ALIAS;
    print ALIAS_NEW "$user:\t" . join(", ", uniq(sort @aliases)) . "\n"
	if ($#aliases >= 0);
    close ALIAS_NEW;
    rename ("$Lockdir/aliases", $Altmp) or
	abort(MSG_error("list_cant_rename_aliases", $Lockdir));
    return abort($lock) if (($lock = commit_aliases()) !~ /^2/);
    return(MSG_ok("list_alias_set_ok"));
}

sub alias_add
# Add new entries to an existing alias; if no such alias exists, add a new one
# Arguments: user name, new array of aliases
# Return value: sauce return code/string
{
    my ($user, @aliases) = @_;
    my @oldaliases;
    if (scalar (@oldaliases = alias_get($user))) {
	return alias_set($user, @oldaliases, @aliases);
    }
    else {
	return alias_set($user, @aliases);
    }
}

sub alias_delete
# Remove the specified aliases from the aliasfile
# Arguments: array of aliases to be removed
# Return value: sauce return code/string
{
    my (@aliases) = @_;
    my ($lock, $line, $count);
    return(MSG_ok("list_alias_delete_ok")) if ($#aliases < 0);
    return $lock if (($lock = lock_aliases()) !~ /^2/);
    open(ALIAS, $Aliasfile) or abort(MSG_error("list_cant_read_aliases", $Aliasfile));
    open (ALIAS_NEW, "> $Lockdir/aliases") or
	abort(MSG_error("list_cant_write_aliases", $Lockdir));
    my %hash = map { $_ => 1; } @aliases;
    while (<ALIAS>) {
	/^(.+?):/;
	print ALIAS_NEW unless (defined $hash{$1});
	++$count;
    }
    close ALIAS;
    close ALIAS_NEW;
    rename ("$Lockdir/aliases", $Altmp) or
	abort(MSG_error("list_cant_rename_aliases", $Lockdir));
    return abort($lock) if (($lock = commit_aliases()) !~ /^2/);
    return(MSG_error("list_alias_invalid")) unless $count;
    return(MSG_ok("list_alias_delete_ok"));
}

######################################################
# for all list_ functions to use virtual majordomo, set $group to
# the virtual site group and $fqdn to the virtual site fqdn

sub list_virtualize
# Does the given list exist?
# Arguments: virtual site group, virtual site fqdn
# Return value: virtualized majordomo list directory, list tag,
#               list owner, list address and wrapper arguments
# Warning: if list owner is set to a list and majordomo is misconfigured, you
# will generate a never ending continuum of undeliverable mails and probably have
# to deal with a highly loaded machine.
{
    my ($group, $fqdn) = @_;

    # remember to edit majordomo.cf if you wanna use $group instead of $fqdn to
    # determine $dir
    # my $dir = $group ? "$Majordomo_dir/$group" : "$Majordomo_dir";
    my $dir = $fqdn ? "$Majordomo_dir/$fqdn" : "$Majordomo_dir";
    my $tag = $group ? "_${group}" : "";
    my $owner = $Majordomo_owner;
    my $addr = $fqdn ? "\@$fqdn" : "";
    my $args = "";
    $args .= "-v $group " if ($group);
    $args .= "-f $fqdn " if ($fqdn);
    return $dir, $tag, $owner, $addr, $args;
}

sub list_exist
# Does the given list exist?
# Arguments: virtual site group, virtual site fqdn, list name
# Return value: true if it does, false if it doesn't
{
    my ($group, $fqdn, $list) = @_;
    my ($dir) = list_virtualize ($group, $fqdn);
    $list =~ tr[A-Z][a-z];	# lowercase the characters
    return (-e "$dir/lists/$list");
}

sub list_create
# Create a list with the specified membership
# Arguments: virtual site group, virtual site fqdn, array of list name, members
# Return value: sauce return code/string
# Side effects: add/udpate file(s) belonging to the list
{
    my ($group, $fqdn, $list, @members) = @_;
    my ($dir, $tag, $owner, $addr, $args) = list_virtualize ($group, $fqdn);
    my $lock;
    $list =~ tr[A-Z][a-z];	# lowercase the characters
    return (MSG_error("list_invalid_already_exists")) if (-e "$dir/lists/$list");
    return MSG_error("list_invalid_adding_self") if (grep(/^$list$/,@members));
    return $lock if (($lock = lock_aliases()) !~ /^2/);
    open (LIST, "> $dir/lists/$list" ) or abort(MSG_error("list_cant_write_list", $Majordomo_dir));
    close (LIST);
    open (ALIAS, ">> $Majordomo_aliases") or
	abort(MSG_error("list_cant_append_maj_aliases", $Majordomo_aliases));
    print ALIAS <<END;
${list}${tag}:\t"|$Majordomo_dir/wrapper $args resend -l $list ${list}${tag}-list"
${list}${tag}-list:\t:include:$dir/lists/$list
${list}${tag}-request:\t"|$Majordomo_dir/wrapper $args request-answer $list"
${list}${tag}-approval:\t$owner
owner-${list}${tag}:\t$owner
${list}${tag}-owner:\towner-${list}${tag}
END

    close ALIAS;

    # add entries to virtusertable so $list$addr works
    if ($tag && $addr) {
	my $ret;
	my %alias = ("${list}${addr}" => "${list}${tag}",
		     "${list}-list${addr}" => "${list}${tag}-list",
		     "${list}-approval${addr}" => "${list}${tag}-approval",
		     "${list}-request${addr}" => "${list}${tag}-request",
		     "owner-${list}${addr}" => "owner-${list}${tag}",
		     "${list}-owner${addr}" => "${list}${tag}-owner");
	if (($ret = mail_virtuser_add_alias(%alias)) !~ /^2/) {	
	    unlock_aliases();
	    return $ret; 
	}
    }
    unlock_aliases();

    # set the members of the list
    my $ret;
    return $ret
	if (($ret = list_set_mem($group, $fqdn, $list, @members)) !~ /^2/);

    # cannot send mail during installation or sysreset
    # kludge..never do it -- gsri
#    if (-e "$Majordomo_aliases") {
#       open (MAIL, "| $Mail_cmd majordomo${addr}") ||
#           return (MSG_error("list_cant_send_mail"));
#       print MAIL "\nconfig $list ${list}.gooey\n";
#       close MAIL;
#    }


    # create the default config file
    return($ret) if (($ret = list_config_create($group, $fqdn, $list)) !~ /^2/);

    return(MSG_ok("list_create_ok", $list));
}

sub list_delete
# Delete the specified lists
# Arguments: virtual site group, virtual site fqdn, lists to be deleted
# Return value: sauce return code/string
# Side effects: remove files belonging to the lists
{
    my ($group, $fqdn, @lists) = @_;
    my ($dir, $tag, $owner, $addr, $args) = list_virtualize ($group, $fqdn);
    my $lock;
    @lists = grep {
        # cannot send mail during installation or sysreset
        # kludge..never do it -- gsri
	if (1) # if (-f "$Majordomo_dir/lists/$_.config")
	{
	    tr[A-Z][a-z];		# lowercase the characters
	    unlink "$dir/lists/$_.intro";
	    unlink "$dir/lists/$_.info";
	    unlink "$dir/lists/$_.auto";
	    unlink "$dir/lists/$_.open";
	    unlink "$dir/lists/$_.closed";
	    unlink "$dir/lists/$_.config";
	    1;
	}
	else { 0; }
    } @lists;

    return $lock if (($lock = lock_aliases()) !~ /^2/);
    open (ALIAS, "$Majordomo_aliases") or
	abort(MSG_error("list_cant_read_maj_aliases", $Majordomo_aliases));
    open (ALIAS_NEW, "> $Lockdir/aliases.majordomo") or
	abort(MSG_error("list_cant_write_maj_aliases", $Lockdir));
    my %hash = map { $_ => 1; } @lists;
    while (<ALIAS>) {
	if (/^(\S+?)${tag}-list:\s/ || /^(\S+?)${tag}-request:\s/ ||
	    /^(\S+?)${tag}-approval:\s/ || /^owner-(\S+?)${tag}:\s/ ||
            /^(\S+?)${tag}-owner:\s/ || /^(\S+?)${tag}:\s/) {
	    print ALIAS_NEW unless (defined $hash{$1});
	}
        else {
	    print ALIAS_NEW;
	}
    }
    close ALIAS_NEW;
    close ALIAS;
    unlink map "$dir/lists/$_", @lists;
    move("$Lockdir/aliases.majordomo", $Majordomo_aliases) or
	abort(MSG_error("list_cant_move_maj_aliases"));

    if ($tag && $addr) {
        my $list;
        foreach $list (@lists) {
	    my $ret;
	    my %alias = ("${list}${addr}" => "${list}${tag}",
			 "${list}-list${addr}" => "${list}${tag}-list",
			 "${list}-approval${addr}" => "${list}${tag}-approval",
			 "${list}-request${addr}" => "${list}${tag}-request",
			 "owner-${list}${addr}" => "owner-${list}${tag}",
			 "${list}-owner${addr}" => "${list}${tag}-owner");
	    if (($ret = mail_virtuser_del_alias(%alias)) !~ /^2/) {
		unlock_aliases();
		return $ret;
	    }
	}
    }

    unlock_aliases();
    return(MSG_ok("list_delete_ok"));
}

sub list_modify
# Modify the specified list
# Arguments: virtual site group, virtual site fqdn, old listname,
#            new listname, new members 
# Return value: usual error string
# Side effects: if old listname != new listname, we could either nix the
#     old list and create a new one or update only the config file (not
#     the info, intro and other paraphrenalia). the rationale: one
#     could delete and recreate the list if they want to do the
#     former. we will do the latter though it is more work.
#     -- gsri 1999/06/18 09:47:01 EDT
{
    my ($group, $fqdn, $oldlist, $list, @members) = @_;
    my ($dir, $tag, $owner, $addr, $args) = list_virtualize ($group, $fqdn);
    $list =~ tr[A-Z][a-z];	# lowercase the characters
    $oldlist =~ tr[A-Z][a-z];	# lowercase the characters

    if ($oldlist ne $list) {
	# return error if new listname exists already
	return (MSG_error("list_invalid_already_exists")) if (-e "$dir/lists/$list");
	
	# rename oldlist files into list files.
	if (-d "$dir/lists") {
	    opendir DIR, "$dir/lists" or
		return(MSG_error("list_cant_open_listdir", "$dir/lists"));
	    "${oldlist}.intro" =~ /^${oldlist}(\.(intro|info|auto|open|closed|config))?$/;
            map { move("$dir/lists/$_", "$dir/lists/$list$1") if (//); } readdir DIR;
            closedir DIR;
        }
    
	# obtain the lock
        my $lock;
	return $lock if (($lock = lock_aliases()) !~ /^2/);

	# update config file, if any
	if (open CONFIG, "$dir/lists/$list.config") {
	    open TMPCONF, "> $Lockdir/$list.config";
	    while (<CONFIG>) {
		s/\b$oldlist\b/$list/g;
		print TMPCONF;
	    }
	    close TMPCONF;
	    close CONFIG;
	    move("$Lockdir/$list.config", "$dir/lists/$list.config") or
		abort(MSG_error("list_cant_move_config"));
	    chmod 0660, "$dir/lists/$list.config";
	    chown (scalar getpwnam($Majordomo_user),
		   scalar getgrnam($group ? $group : $Majordomo_group),
		   "$dir/lists/$list.config");
	}
	
	# update aliases file
	open (ALIAS, "$Majordomo_aliases") or
	  abort(MSG_abort("list_cant_read_maj_aliases", $Majordomo_aliases));
	open (ALIAS_NEW, "> $Lockdir/aliases.majordomo") or
	  abort(MSG_abort("list_cant_write_maj_aliases", $Lockdir));

	"${oldlist}${tag}:" =~ /^(${oldlist}${tag}|${oldlist}${tag}-list|${oldlist}${tag}-request|${oldlist}${tag}-approval|owner-${oldlist}${tag}|${oldlist}${tag}-owner):/;
	while (<ALIAS>) {
	    print ALIAS_NEW unless //;
	}
	print ALIAS_NEW <<EOF;
${list}${tag}:\t"|$Majordomo_dir/wrapper $args resend -l $list ${list}${tag}-list"
${list}${tag}-list:\t:include:$dir/lists/$list
${list}${tag}-request:\t"|$Majordomo_dir/wrapper $args request-answer $list"
${list}${tag}-approval:\t$owner
owner-${list}${tag}:\t$owner
${list}${tag}-owner:\towner-${list}${tag}
EOF
        close ALIAS_NEW;
	close ALIAS;
	move("$Lockdir/aliases.majordomo", $Majordomo_aliases) or
	    abort(MSG_abort("list_cant_move_maj_aliases"));

        # fix virtusertable if necessary
        if ($tag && $addr) {
	    my(%aliases) = mail_virtuser_list_alias();
	    my(%routes) = mail_virtuser_list_route();
    	
	    delete $aliases{"${oldlist}${addr}"};
	    delete $aliases{"${oldlist}-list${addr}"};
	    delete $aliases{"${oldlist}-approval${addr}"};
	    delete $aliases{"${oldlist}-request${addr}"};
	    delete $aliases{"owner-${oldlist}${addr}"};
	    delete $aliases{"${oldlist}-owner${addr}"};
	    $aliases{"${list}${addr}"} = "${list}${tag}";
	    $aliases{"${list}-list${addr}"} = "${list}${tag}-list";
	    $aliases{"${list}-approval${addr}"} = "${list}${tag}-approval";
	    $aliases{"${list}-request${addr}"} = "${list}${tag}-request";
	    $aliases{"owner-${list}${addr}"} = "owner-${list}${tag}";
	    $aliases{"${list}-owner${addr}"} = "${list}${tag}-owner";
	    unless (mail_build_virtusertable(\%aliases, \%routes)) {
		unlock_aliases();
                return MSG_error("list_cant_modify_aliases");
	    }
	}

	# unlock
        unlock_aliases();
    }

    # update members of the list
    my $ret;
    return $ret
	if (($ret = list_set_mem($group, $fqdn, $list, @members)) !~ /^2/);
    return(MSG_ok("list_modify_ok"));
}

######################################################

sub list_is_mem
# Checks for membership in a list
# Arguments: virtual site group, virtual site fqdn, list name, member name
# Return value: true if member is in list, false otherwise
{
    my ($group, $fqdn, $list, $member) = @_;
    my ($dir) = list_virtualize ($group, $fqdn);
    return 0 if ($member eq "nobody");
    $list =~ tr[A-Z][a-z];	# lowercase the characters
    if (-f "$dir/lists/$list") {
	open (LIST, "$dir/lists/$list") or
	    return (MSG_error("list_cant_read_list", $Majordomo_dir));
	while ( <LIST> ) {
	    chop;
	    (close LIST, return 1) if $member eq $_;
	}
	close LIST;
    }
    return 0;
}

sub list_get_mem
# Returns the membership of a list
# Arguments: virtual site group, virtual site fqdn, list name
# Return value: an array of (member1, member2... )
{
    my ($group, $fqdn, $list) = @_;
    my ($dir) = list_virtualize ($group, $fqdn);
    my @members;
    $list =~ tr[A-Z][a-z];	# lowercase the characters
    if (-f "$dir/lists/$list") {
	open (LIST, "$dir/lists/$list") or
	    return (MSG_error("list_cant_read_list", "$dir/lists/$list"));
	while ( <LIST> ) {
	    chop;
	    next if (/^nobody$/);
	    push @members, $_;
	}
	close LIST;
	return @members;
    }
    return;
}

sub list_set_mem
# (Re)sets the membership of a list
# Arguments: virtual site group, virtual site fqdn, list name
# Return value: sauce return code/string
# Side effects: modify list membership file
{
    my ($group, $fqdn, $list, @members) = @_;
    my ($dir) = list_virtualize ($group, $fqdn);
    $list =~ tr[A-Z][a-z];	# lowercase the characters
    return MSG_error("list_invalid_adding_self") if (grep(/^$list$/,@members));
    my $lock;
    return $lock if (($lock = lock_aliases()) !~ /^2/);
    open (LIST, "> $Lockdir/$list.majordomo" ) or
	abort(MSG_error("list_cant_write_newlist", $Lockdir));
    print LIST join "\n", uniq(sort @members);
    if ($#members < 0) {
	print LIST "nobody\n";
    }
    else {
	print LIST "\n";
    }
    close LIST;
    move("$Lockdir/$list.majordomo", "$dir/lists/$list") or
	abort(MSG_error("list_cant_move_list"));
    chmod 0660, "$dir/lists/$list";
    chown (scalar getpwnam($Majordomo_user),
	   scalar getgrnam($group ? $group : $Majordomo_group),
	   "$dir/lists/$list");
    unlock_aliases();
    return(MSG_ok("list_set_membership_ok", $list));
}

######################################################

sub list_add_mem
# Adds members to a list
# Arguments: virtual site group, virtual site fqdn, list name, new members
# Return value: sauce return code/string
{
    my ($group, $fqdn, $list, @new_mem) = @_;
    my (@mem, $ret);
    return $mem[0]
	if ((@mem = list_get_mem($group, $fqdn, $list))[0] =~ /^[134]\S*\s/);
    return $ret
	if (($ret = list_set_mem($group, $fqdn, $list, @mem, @new_mem)) !~ /^2/);
    return(MSG_ok("list_add_membership_ok", $list));
}

sub list_del_mem
# Removes members from a list
# Arguments: virtual site group, virtual site fqdn, list name, old members
# Return value: sauce return code/string
{
    my ($group, $fqdn, $list, @old_members) = @_;
    my (@members, @new_members, $member, $ret);
    return $members[0]
	if ((@members = list_get_mem($group, $fqdn, $list))[0] =~ /^[134]\S*\s/);
    my %hash = map { $_ => 1; } @old_members;
    while ($member = shift @members) {
	 push @new_members, $member unless (defined $hash{$member});
    }
    return $ret
	if (($ret = list_set_mem($group, $fqdn, $list, @new_members)) !~ /^2/);
    return(MSG_ok("list_delete_members_ok", $list));
}

######################################################

sub list_lists
# Returns a list of all the mailing lists
# Arguments: virtual site group, virtual site fqdn
# Return value: an array of all the mailing lists
{
    my ($group, $fqdn) = @_;
    my ($dir) = list_virtualize ($group, $fqdn);
    if (-d "$dir/lists") {
	opendir (DIR, "$dir/lists") or return (MSG_error("list_cant_open_listdir", "$dir/lists"));

	my (@lists, $list);
	foreach $list (sort readdir (DIR)) {
	    next if ($list =~ /[^-_0-9a-z]/);       # skip non-list files
	    next if ($list =~ /^(RCS|CVS|core)$/);  # files and directories to ignore
	    next if (-d "$dir/lists/$list");              # skip directories
	    push @lists, $list;
	}
	
	closedir (DIR);
	return @lists;
    }
    return;
}

sub list_lists_mem
# Returns a list of all the lists and their members in a 2D array
# Arguments: virtual site group, virtual site fqdn
# Return value: an array of (list name, member1, member2, ...)
{
    my ($group, $fqdn) = @_;
    my @lists;
    return $lists[0]
	if ((@lists = list_lists($group, $fqdn))[0] =~ /^[134]\S*\s/);
    return map {
	my $list=$_;
	my @mem;
	return $mem[0]
	    if ((@mem = list_get_mem($group, $fqdn, $list))[0] =~ /^[134]\S*\s/);
        [$list, @mem];
    } sort @lists;
}

sub list_lists_del_mem
# Delete the specified users from all lists
# Arguments: virutal site group, virtual site fqdn, members to be removed
# Return value: sauce return code/string
{
    my ($group, $fqdn, @members) = @_;
    my @lists;

    # i bow to thee oh, Crom! gimme the power to destroy these evil @members!
    return $lists[0]
	if ((@lists = list_lists($group, $fqdn))[0] =~ /^[134]\S*\s/);

    # drop the results to the floor and move on!
    map list_del_mem($group, $fqdn, $_, @members), @lists;
    return(MSG_ok("list_delete_all_ok"));
}

######################################################

sub list_config_create
# generate a list config file
# calls the majordomo writeconfig sub
{
    my ($group, $fqdn, $list) = @_;
    my $ret;
    unshift (@INC, $Majordomo_dir);

    # get listdir
    my ($dir) = list_virtualize($group, $fqdn);

    # check to see if the "config_parse.pl" file exists
    if (! -r "$Majordomo_dir/$Majordomo_config_parse") {
	return(MSG_error("list_cant_config_create"));
    }

    # pull in majordomo config_parse file
    require "$Majordomo_config_parse";

    # call majordomo routine that creates default config file
    &main::get_config("$dir/lists", $list, 0);
    
    # set ownership and permissions
    chmod 0660, "$dir/lists/$list.config";
    chown (scalar getpwnam($Majordomo_user), 
	   scalar getgrnam($group ? $group : $Majordomo_group),
	   "$dir/lists/$list.config");

    return(MSG_ok("list_config_create_ok"));
}

sub list_config_getstate
# find out if a mailing list has open subscription policy
#
# 1 = open
# 0 = closed
#
{
    my ($group, $fqdn, $list) = @_;
    my ($dir) = list_virtualize($group, $fqdn);
    my $state = 0;
    
    FILE_lock_r("$dir/lists/$list.config", *CONFIG) ||
	return(0);
    while (<CONFIG>) {
	if (/^\s*subscribe_policy[^\=]*=\s*open$/) {
	    $state = 1;
	    last;
	}
    }
    FILE_unlock(*CONFIG);
    
    return $state;
}

sub list_config_open
# Open up lists for subscription
# Arguments: virtual site group, virtual site fqdn, lists to be opened
# Return value: number of lists actually opened
{
    my ($group, $fqdn, @lists) = @_;
    map {
	tr[A-Z][a-z];		# lowercase the characters
	my $ret =
	    list_config_update($group, $fqdn, $_,
			       "(un)?subscribe_policy", "open");
	return $ret if ($ret !~ /^2/);
    } @lists;
    return (MSG_ok("list_config_open_ok"));
}

sub list_config_closed
# Close lists for subscription
# Arguments: virtual site group, virtual site fqdn, lists to be closed
# Return value: number of lists actually closed
{
    my ($group, $fqdn, @lists) = @_;
    map {
	tr[A-Z][a-z];		# lowercase the characters
	my $ret = 
	    list_config_update($group, $fqdn, $_,
			       "(un)?subscribe_policy", "closed");
	return $ret if ($ret !~ /^2/);
    } @lists;
    return (MSG_ok("list_config_closed_ok"));
}

sub list_config_password
# set a password for a list
{
    my ($group, $fqdn, $list, $passwd) = @_;
    my $ret = list_config_update($group, $fqdn, $list,
				 "(approve|admin)_passwd", $passwd);
    return $ret if ($ret !~ /^2/);
    return (MSG_ok("list_config_password_ok"));
}

sub list_config_getpassword
# find out a mailing list password
{
    my ($group, $fqdn, $list) = @_;
    my ($dir) = list_virtualize($group, $fqdn);
    my $password = '';

    FILE_lock_r("$dir/lists/$list.config", *CONFIG) ||
	return(0);
    while (<CONFIG>) {
	($password = $1) if (/^\s*admin_passwd\s*=\s*(.*)$/);
    }
    FILE_unlock(*CONFIG);

    return $password;
}

sub list_config_getrestrict
# find out if a mailing list allows spam
{
    my ($group, $fqdn, $list) = @_;
    my ($dir) = list_virtualize($group, $fqdn);
    my $restrict = 1;

    FILE_lock_r("$dir/lists/$list.config", *CONFIG) ||
	return(0);
    while (<CONFIG>) {
	if (/^\s*restrict_post\s*=\s*$/) {
	    $restrict = 0;
	    last;
	}
    }
    FILE_unlock(*CONFIG);

    return $restrict;
}

sub list_config_restrict
# disallow spam to the mailing list
{
    my ($group, $fqdn, $list) = @_;
    my $ret = list_config_update($group, $fqdn, $list,
				 "restrict_post", $list);
    return $ret if ($ret !~ /^2/);
    return (MSG_ok("list_config_restrict"));
}

sub list_config_norestrict
# allow spam to the mailing list
{
    my ($group, $fqdn, $list) = @_;
    my $ret = list_config_update($group, $fqdn, $list,
				 "restrict_post", "");
    return $ret if ($ret !~ /^2/);
    return (MSG_ok("list_config_norestrict"));
}

sub list_config_update
# Update the config info for a list
# Arguments: virtual site group, virtual site fqdn, list name, 
#            config variable regex, config value
# Return value: 1 on success, 0 means failure
{
    my ($group, $fqdn, $list, $regex, $val) = @_;
    my ($dir) = list_virtualize ($group, $fqdn);
    my $lock;
    $list =~ tr[A-Z][a-z];	# lowercase the characters
    return $lock if (($lock = lock_aliases()) !~ /^2/);
    open (LIST, "< $dir/lists/$list.config") or
	abort(MSG_error("list_cant_read_config", $list));
    open (LIST_NEW, "> $Lockdir/$list.config") ||
	abort(MSG_error("list_cant_write_config", $Lockdir));
    while (<LIST>) {
	if (/^$regex/) {
	    print LIST_NEW "$& = $val\n";
	}
	else {
	    print LIST_NEW;
	}
    }
    close LIST;
    close LIST_NEW;
    move("$Lockdir/$list.config", "$dir/lists/$list.config") or
	abort(MSG_error("list_cant_move_config"));
    chmod 0660, "$dir/lists/$list.config";
    chown (scalar getpwnam($Majordomo_user), 
	   scalar getgrnam($group ? $group : $Majordomo_group),
	   "$dir/lists/$list.config");
    unlock_aliases();
    return (MSG_ok("list_config_update_ok"));
}

######################################################

sub list_create_virtual
# Create a virtual list server
# Arguments: virtual site group, virtual site fqdn
# Return value: usual error string
{
    my ($group, $fqdn) = @_;
    return (MSG_error("list_invalid_virtual")) unless ($group && $fqdn);
	
    my ($dir, $tag, $owner, $addr, $args) = list_virtualize ($group, $fqdn);
    my $lock;

    # add new directories
    mkdir ($dir, 02750);
    mkdir ("$dir/lists", 02750);
    mkdir ("$dir/digests", 02750);
    mkdir ("$dir/files", 02750);
    open (LOG, "> $dir/log");
    close (LOG);
    chmod (0640, "$dir/log");
    chown (scalar getpwnam($Majordomo_user), scalar getgrnam($group),
	   $dir, "$dir/lists", "$dir/digests", "$dir/files", "$dir/log");

    # obtain the lock
    return $lock if (($lock = lock_aliases()) !~ /^2/);

    # add new majordomo aliases
    copy ("$Majordomo_aliases", "$Lockdir/aliases.majordomo") or
	abort(MSG_abort("list_cant_copy_maj_aliases", $Lockdir));
    open (ALIAS, ">> $Lockdir/aliases.majordomo") or
	abort(MSG_abort("list_cant_write_maj_aliases", $Lockdir));

    print ALIAS <<EOF;
${Majordomo_server}${tag}:\t"|$Majordomo_dir/wrapper $args majordomo"
owner-${Majordomo_server}${tag}:\t$owner
${Majordomo_server}${tag}-owner:\towner-${Majordomo_server}${tag}
EOF
    close ALIAS;

    move("$Lockdir/aliases.majordomo", $Majordomo_aliases) or
        abort(MSG_abort("list_cant_move_maj_aliases"));

    # add entries to virtusertable so $Majordomo_server$addr works
    if ($tag && $addr) {
	my $ret;
	my %alias = ("${Majordomo_server}${addr}" => "${Majordomo_server}${tag}",
		     "owner-${Majordomo_server}${addr}" =>
		     "owner-${Majordomo_server}${tag}",
		     "${Majordomo_server}-owner${addr}" =>
		     "${Majordomo_server}${tag}-owner");
	if (($ret = mail_virtuser_add_alias(%alias)) !~ /^2/) {
	    unlock_aliases();
	    return $ret;
	}
    }

    # unlock
    unlock_aliases();
    return (MSG_ok("list_crevirt_ok"));
}

sub list_delete_virtual
# Remove a virtual list server
# Arguments: virtual site group, virtual site fqdn
# Return value: usual error string
{
    my ($group, $fqdn) = @_;
    return (MSG_error("list_invalid_virtual")) unless ($group && $fqdn);
	
    my ($dir, $tag, $owner, $addr) = list_virtualize ($group, $fqdn);
    my $lock;

    # delete the files and directories
    system ("/bin/rm -rf $dir") and return (MSG_error("list_cant_delvirt_dir"));

    # obtain the lock
    return $lock if (($lock = lock_aliases()) !~ /^2/);

    # add new majordomo aliases
    open (ALIAS, "$Majordomo_aliases") or
	abort(MSG_abort("list_cant_read_maj_aliases", $Majordomo_aliases));
    open (ALIAS_NEW, "> $Lockdir/aliases.majordomo") or
	abort(MSG_abort("list_cant_write_maj_aliases", $Lockdir));

    "${Majordomo_server}${tag}:" =~ /^(${Majordomo_server}${tag}|owner-${Majordomo_server}${tag}|${Majordomo_server}${tag}-owner):/;
    while (<ALIAS>) {
	print ALIAS_NEW unless //;
    }
    close ALIAS;
    close ALIAS_NEW;

    move("$Lockdir/aliases.majordomo", $Majordomo_aliases) or
        abort(MSG_abort("list_cant_move_maj_aliases"));

    if ($tag && $addr) {
	my $ret;
	my %alias = ("${Majordomo_server}${addr}" => "${Majordomo_server}${tag}",
		     "owner-${Majordomo_server}${addr}" =>
		     "owner-${Majordomo_server}${tag}",
		     "${Majordomo_server}-owner${addr}" =>
		     "${Majordomo_server}${tag}-owner");
	if (($ret = mail_virtuser_del_alias(%alias)) !~ /^2/) {
	    unlock_aliases();
	    return $ret;
	} 
    }

    unlock_aliases();
    return (MSG_ok("list_delvirt_ok"));
}

sub list_modify_virtual
# Rename a virtual list server
# Arguments: virtual site group, old virtual site fqdn, new virtual site fqdn
# Return value: usual error string
{
    my ($group, $oldfqdn, $fqdn) = @_;
    my ($ret,$list);
    return (MSG_error("list_invalid_virtual")) unless ($group && $oldfqdn && $fqdn);
    return (MSG_ok("list_modvirt_ok")) if ($oldfqdn eq $fqdn);
    
    my ($dir1, $tag1, $owner1, $addr1, $args1) =
	list_virtualize ($group, $oldfqdn);
    my ($dir2, $tag2, $owner2, $addr2, $args2) =
	list_virtualize ($group, $fqdn);
    
    my ($ref, @lists);
    return $lists[0]
	if ((@lists = list_lists_mem($group, $oldfqdn))[0] =~ /^[134]\S*\s/);
    
    # rename the directory
    if ($dir1 ne $dir2) {
	move($dir1, $dir2) or abort(MSG_abort("list_cant_modvirt_dir"));
    }
    
    # obtain the lock
    my $lock;
    return $lock if (($lock = lock_aliases()) !~ /^2/);
    
    # modify majordomo aliases
    open (ALIAS, "$Majordomo_aliases") or
	abort(MSG_abort("list_cant_read_maj_aliases", $Majordomo_aliases));
    open (ALIAS_NEW, "> $Lockdir/aliases.majordomo") or
	abort(MSG_abort("list_cant_write_maj_aliases", $Lockdir));
    
    # perl-bug boo boo: does not allow the first expr in the list to
    # be a quoted string either using " or qw or whatever -- gsri 6/26/99
    my %hash = map {
	1 => undef,
	"$$_[0]${tag1}" => 1,
	"$$_[0]${tag1}-list" => 1,
	"$$_[0]${tag1}-approval" => 1,
	"$$_[0]${tag1}-request" => 1,
	"owner-$$_[0]${tag1}" => 1,
	"$$_[0]${tag1}-owner" => 1;
    } @lists;
    # the unless makes sure the above hack will work even with a list named "1"
    delete $hash{1} unless ($hash{1});
    map { $hash{$_} = 1; } "${Majordomo_server}${tag1}", 
    "owner-${Majordomo_server}${tag1}", "${Majordomo_server}${tag1}-owner";
    
    while (<ALIAS>) {
	/^(\S+?):/;
	print ALIAS_NEW unless (defined $hash{$1});
    }
    
    foreach $ref (@lists) {
	$list = $$ref[0];
	print ALIAS_NEW <<END;
${list}${tag2}:\t"|$Majordomo_dir/wrapper $args2 resend -l $list ${list}${tag2}-list"
${list}${tag2}-list:\t:include:$dir2/lists/$list
${list}${tag2}-request:\t"|$Majordomo_dir/wrapper $args2 request-answer $list"
${list}${tag2}-approval:\t$owner2
owner-${list}${tag2}:\t$owner2
${list}${tag2}-owner:\towner-${list}${tag2}
END
    }

    print ALIAS_NEW <<END;
${Majordomo_server}${tag2}:\t"|$Majordomo_dir/wrapper $args2 majordomo"
owner-${Majordomo_server}${tag2}:\t$owner2
${Majordomo_server}${tag2}-owner:\towner-${Majordomo_server}${tag2}
END
    close ALIAS;
    close ALIAS_NEW;

    move("$Lockdir/aliases.majordomo", $Majordomo_aliases) or
        abort(MSG_abort("list_cant_move_maj_aliases"));

    # fix virtusertable if necessary
    if ($tag1 && $addr1 && $tag2 && $addr2) {
	my(%aliases) = mail_virtuser_list_alias();
	my(%routes) = mail_virtuser_list_route();
	
	foreach $ref (@lists) {
	    $list = $$ref[0];

	    delete $aliases{"${list}${addr1}"};
	    delete $aliases{"${list}-list${addr1}"};
	    delete $aliases{"${list}-approval${addr1}"};
	    delete $aliases{"${list}-request${addr1}"};
	    delete $aliases{"owner-${list}${addr1}"};
	    delete $aliases{"${list}-owner${addr1}"};
	    $aliases{"${list}${addr2}"} = "${list}${tag2}";
	    $aliases{"${list}-list${addr2}"} = "${list}${tag2}-list";
	    $aliases{"${list}-approval${addr2}"} = "${list}${tag2}-approval";
	    $aliases{"${list}-request${addr2}"} = "${list}${tag2}-request";
	    $aliases{"owner-${list}${addr2}"} = "owner-${list}${tag2}";
	    $aliases{"${list}-owner${addr2}"} = "${list}${tag2}-owner";
	}

	delete $aliases{"${Majordomo_server}${addr1}"};
	delete $aliases{"owner-${Majordomo_server}${addr1}"};
	delete $aliases{"${Majordomo_server}-owner${addr1}"};
	$aliases{"${Majordomo_server}${addr2}"} = "${Majordomo_server}${tag2}";
	$aliases{"owner-${Majordomo_server}${addr2}"} =
	    "owner-${Majordomo_server}${tag2}";
	$aliases{"${Majordomo_server}-owner${addr2}"} =
	    "${Majordomo_server}${tag2}-owner";
	unless (mail_build_virtusertable(\%aliases, \%routes)) {
	    unlock_aliases();
	    return MSG_error("list_cant_modify_aliases");
	}
    }

    # unlock
    unlock_aliases();

    # update member addresses in individual list files..big filesys suck!!
    # modify the member addr of each list and set the membership
    foreach $ref (@lists) {
	$list = shift @$ref;
	map { $_ =~ s/$addr1/$addr2/; } @$ref;
	return $ret
	    if (($ret = list_set_mem($group, $fqdn, $list, @$ref)) !~ /^2/);
    }

    # update the addr of the admins in the site-adm list
    my @mem = list_get_mem("", "", $Sites_AdminGroup);
    map { $_ =~ s/$addr1/$addr2/; } @mem;
    $ret = list_set_mem("", "", $Sites_AdminGroup, @mem);

    # update the addr of the admins in the group list
    @mem = list_get_mem("", "", $group);
    map { $_ =~ s/$addr1/$addr2/; } @mem;
    $ret = list_set_mem("", "", $group, @mem);

    return (MSG_ok("list_modvirt_ok"));
}

######################################################

sub lock_aliases
# Lock the aliases file so that nobody can collide with us.
# Arguments: none
# Side effects: copies /etc/aliases to /etc/altmp
{
    if (-e $Altmp)
    {
	return (MSG_error("4261"));
    }
    else
    {
	# Lock the aliases file by creating the altmp file as a
	# copy of the aliases file.
	print STDERR "Copying aliases file $Aliasfile to $Altmp\n" if ($Debug);
	copy ($Aliasfile, $Altmp);
    }
    print STDERR "Locked $Aliasfile with $Altmp\n" if ($Debug);

    open (LOCK,">$Lockdir/list-$$") or
      abort (MSG_error("4266"));
    print LOCK "$$\n$0";
    close LOCK;

    return MSG_ok("2260");
}

sub commit_aliases
# Commits the changes to the aliases file
# Side effects: renames $Altmp to $Aliasfile
# We keep a backup copy (aliases.old) in case newaliases barfs
{
  rename ($Altmp, $Aliasfile) or
      return MSG_error("4262");
  chmod (0644, $Aliasfile);
  unlink("$Lockdir/list-$$");
  return MSG_ok("2261");
}

sub unlock_aliases
# Unlocks the aliases file by wiping altmp
# Arguments: none
# Side effects: wipes altmp
{
    unlink($Altmp) || return MSG_error("4264", $Altmp);
    print STDERR "Removed $Altmp\n" if ($Debug);
    unlink("$Lockdir/list-$$");
    return MSG_ok("2262");
}

sub abort
# leave in a graceful manner
# Arguments: a string to print
# Side effects: unlocks anything we've locked
{
    unlock_aliases()  if (-e "$Lockdir/list-$$");
    return join('',@_);
}

sub uniq
# Removes replicates from a sorted list
# Arguments: input list
# Return value: uniq-ed list
{
    my ($prev, @list);
    while ($_ = shift) {
	 push @list, ($prev = $_) if ($prev ne $_);
    }
    return @list;
}

1;
