# $Id: Locale.pm,v 1.5 1999/10/06 15:13:27 duncan Exp $
#
# author: Kevin K.M. Chiu
#       Cobalt Networks
#
# Copyright 1999-2000, Cobalt Networks, Inc
# All rights reserved

package Cobalt::Locale;

use strict;

# class variables
my $defaultLocale='en_US';
my $Messages_dir='/etc/cobalt/messages';
my $Locale_file='/etc/cobalt/locale';

# constructor
sub new 
{
    my $proto=shift;
    my $class=ref( $proto ) || $proto;
    my $self={};
    bless( $self, $class );
    $self->init( @_ );
    return $self;
}

# initialization
sub init 
{
    my $self=shift;
    my $loc = $ENV{"LANGUAGE"};

    if (!$loc) {
	if (-f $Locale_file) {
	    $loc=`cat $Locale_file`;
	    chop($loc);
	} else {
	    $loc=$defaultLocale;
	}
    }
    %$self=(locale => $loc);
}

# instance method
sub getLocale 
{
    my $self=shift;
    return $self->{ locale };
}

# instance method
sub setLocale 
{
    my $self=shift;

    my ( $locale )=@_;
    $self->{ locale }=$locale;
}

# get all the supported locales by looking for locale properties files
# return a list of supported locale
# static method
sub getSupported 
{
    my $class=shift;
    my @locales;

    opendir(DIR, $Messages_dir);
    while (readdir(DIR)) {
	if ($_ =~ /^([^\.].+)\.prop$/) {
	    push @locales, $1;
	}
    }	
    closedir(DIR);

    return @locales;
}

# see if a locale given by a string is supported
# return 1 if supported, 0 otherwise
# static method
sub isSupported 
{
    my $class=shift;

    my ( $str )=@_;
#    if( grep /^$str$/, $class->getSupported() ) { return 1; }
    if( -e "$Messages_dir/$str.prop" ) { return 1; }
    else { return 0; }
}

# get a new Cobalt::Locale object given a locale string
# argument 0 is the locale string (e.g. en_US, ja)
# return a Cobalt::Locale object or
#   0 if the locale string is invalid or the locale is not supported
# static method
sub getByString 
{
    my $class=shift;

    my ( $str )=@_;

    # !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    # the more stuff this hash has, the more locale we support
    # !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    my %strToLocale=(
		     'cs' => 'cs_CZ',
		     'da' => 'da_DK',
		     'de' => 'de_DE',
		     'el' => 'el_GR',
		     'en' => 'en_US',
		     'es' => 'es_ES',
		     'et' => 'et_EE',
		     'eu' => 'eu_ES',
		     'fi' => 'fi_FI',
		     'fo' => 'fo_FO',
		     'fr' => 'fr_FR',
		     'ga' => 'ga_IE',
		     'hr' => 'hr_HR',
		     'hu' => 'hu_HU',
		     'in' => 'in_ID',
		     'is' => 'is_IS',
		     'it' => 'it_CH',
		     'iw' => 'iw_IL',
		     'ja' => 'ja_JP.SJIS',
		     'kl' => 'kl_GL',
		     'ko' => 'ko_KP',
		     'lt' => 'lt_LT',
		     'lv' => 'lv_LV',
		     'nl' => 'nl_NL',
		     'no' => 'no_NO',
		     'pl' => 'pl_PL',
		     'pt' => 'pt_PT',
		     'ro' => 'ro_RO',
		     'ru' => 'ru_RU',
		     'sk' => 'sk_SK',
		     'sl' => 'sl_SI',
		     'sr' => 'sr_YU',
		     'sv' => 'sv_SE',
		     'tr' => 'tr_TR',
		     'uk' => 'uk_UA',
		     'zh' => 'zh_CN',
		     );

    # standard says the country code should be uppercase
    # and preceeded with underscore,
    # but Microsoft IE uses lowercase and dash no matter what,
    # so let's fix it...
    if( length( $str )>3 ) {
	my $tail=substr( $str, 3 );
	$tail=~tr/a-z/A-Z/;
	$str=substr( $str, 0, 2 ).'_'.$tail;
    }

    if( $strToLocale{ $str } ) { $str=$strToLocale{ $str }; }

    my $locale=new Cobalt::Locale();
    $locale->setLocale( $str );

    return $locale;
}

# get a new Cobalt::Locale object given a HTTP_ACCEPT_LANGUAGE header
# only supported locale indicated by the presence of the locale properties file will be returned
# if no locales in the header are supported, defaultLocale is used
# argument 0 is the HTTP_ACCEPT_LANGUAGE header
# return a Cobalt::Locale object
# static method
sub getByHTTPHeader 
{
    my $class=shift;
    my ($header)=@_;

    # IE gives something like "en", "en,ja", "sq,ja;q=0.7,en-us;q=0.3" or "ja,en-us;q=0.5"
    # NN gives something like "en", "en, ja"
    # preferrence is from left to right
    # remove all the q stuff because IE already sorted the entries
    $header=~s/;q=[\d\.]+//g;
    my @languages=split /[, ]+/, $header;

    my $locale=0;
    my $language;
    foreach $language ( @languages ) {
	# use getByString to qualify locale string if necessary
	if( $class->isSupported( $class->getByString( $language )->getLocale() ) ) {
	    $locale=$class->getByString( $language );
	    last;
	}
    }

    if( $locale==0 ) {
	# use defaultLocale
	my $goodLocale=$defaultLocale;

	# if defaultLocale is not supported, but some others are
	my @locales=$class->getSupported();
	if( !$class->isSupported( $defaultLocale ) &&
	    $#locales>=0 ) {
	    # use any supported locale
	    $goodLocale=$locales[ 0 ];
	}

	$locale=$class->getByString( $goodLocale );
    }

    return $locale;
}

# get properties from the locale specific properties file
# argument 0 is the name of the property
sub getProperty
{
    my $self=shift;
    my $property=shift;
    my $locale=$$self{"locale"};

    open (PROP, "$Messages_dir/$locale.prop");
    while (<PROP>) {
        if (/^$property: (.+)$/) {
	    close(PROP);
	    return($1);
	}
    }
    close(PROP);
    return(0);
}

1;
