#!/usr/bin/perl -w

## ID: $Id: Meta.pm,v 1.40 1999/10/11 00:01:49 duncan Exp $ 
##
## +++++++
## Meta.pm
## +++++++
##
## Copyright (c) 1999-2000 Cobalt Networks, Inc.
##
## Authors:
##   Duncan Laurie <duncan@cobaltnet.com>
##   Harris Vaegan-Lloyd <harris@cobaltnet.com>
##
## simple object interface to PostgreSQL database
##
## the metadatabase module is a to objectify database access for simple
## structures herein deemed objects.  objects have a type, refered to by the
## "type" trait.  (SQL primary key)
##
## ++++++++++++++++
## GLOBAL FUNCTIONS
## ++++++++++++++++
##
## Cobalt::Meta::connect($dbname, $user, $port);
##
##   connects to the postgres database $dbname as $user on $port
##
## Cobalt::Meta::set_encoding($encoding);
##
##   sets the multibyte client encoding method, set based upon the value
##   in /etc/cobalt/locale when the database is first connected
##
## $encoding = Cobalt::Meta::get_encoding();
##
##   returns the current encoding method
##
## ++++++++++++++
## OBJECT METHODS
## ++++++++++++++
##
## objects must be created:
##
##	$obj = Cobalt::Meta->new( %arg );
## EX:$obj = Cobalt::Meta->new( "type"   => $type );
## EX:$obj = Cobalt::Meta->new( "type"   => $type,
##							  "name"   => $name,
##							  "modify" => $modify, ... );
## 
## they can be populated with values from the database using the retrieve
## function and providing a value for the "name" key and any additional conditions
##
##	$obj->retrieve("name", [ <$conj, $left, $op, $right>, ... ]);
##
## to get the value of a key use the get function:
##
##	$value = $obj->get($key);
##
## to insert a key=>value pair to the object use the put function:
##
##	$obj->put($key, $value);
##
## to save your object changes to the database use the save function:
##
##	$obj->save();
##
## delete objects (only local copy) with the DESTROY function:
##
##	$obj->DESTROY();
##
## delete objects from the database use the delete function:
##
##	$obj->delete();
##
## you can do more complex SQL queries:
## (this will return an array of objects)
##
## @meta=Cobalt::Meta::query("type"	  => $table,
##                           "distinct"   => 1,
##                           "range"      => [ low, high ],
##                           "where"      => [ <$left, $op, $right>,
##                                             <$conj, $left, $op, $right>, ... ],
##                           "sort"       => "name",
##                           "order"      => "dsc");
##
## to commit the object to reality (Actually create a vsite for example.)
##
##	$obj->commit();
##
## +++++++++++++
## XML FUNCTIONS
## +++++++++++++
##
## you may dump objects as XML.
##
##	$xml = $obj->as_xml();
##
## and create them from XML.
##
##	$obj = Cobalt::Meta->from_xml_elem( $xml );
##
## ++++++++++++++++
## META INHERITANCE
## ++++++++++++++++
##
## objects may have members, these only exist while the object is created
## and must be eplicitly populated for retrieved objects.
##
## add childer objects with
##
##	 $obj->add_member( $child_object );
##
## retrieve child objects with
##
##	 @children = $obj->populate_members();
##
## To populate the children of a retrieved object use
##
##	   $obj->populate_members();
##
## to see the members of an object (Not put them into the object.)
## use get_members. Args are the same as for query, but will only
## get you objects that are members and of the type specified.
##
##	 @members = $obj->get_members( ... );
##
## Any bad puns are welcome.
##
## for debugging purposes there is a display function that will simply
## output the key=>value pairs to STDERR:
##
##	$obj->display();
##
## though this is depreceated in favour of xml_dump which will handle hierachies
## and has a TLA so is inherently cooler ;).

##------------------------------------------------------------------------
package Cobalt::Meta;
##

use Pg qw(PGRES_EMPTY_QUERY
	  PGRES_COMMAND_OK
	  PGRES_CONNECTION_OK
	  PGRES_TUPLES_OK
	  PGRES_COPY_OUT
	  PGRES_COPY_IN
	  PGRES_BAD_RESPONSE
	  PGRES_NONFATAL_ERROR
	  PGRES_FATAL_ERROR);
use Pg;
use Cobalt::Util;
use Cobalt::Locale;
use XML::Parser;
use MIME::Base64;
use strict;

use vars qw($dbc $Meta_id %client_encoding $DEBUG);

$DEBUG = 0;

# Meta configuration variables
my $Meta_dbname = "cobalt";
my $Meta_dbuser = "admin";
my $Meta_dbport = 5583;
$Meta_id = "/etc/cobalt/.meta.id";

# list of encoding types, referenced by their locale
%client_encoding = ("en_US"      => "LATIN1",  # English/US
                    "ja_JP.SJIS" => "SJIS",    # Japanese SJIS
                    "ja_JP.EUC"  => "EUC_JP",  # Japanese EUC
                    "zh_CN"      => "EUC_CN",  # Chinese EUC
                    "ko_KP"      => "EUC_KR",  # Korean EUC
                    "unicode"    => "UNICODE",
                    "ascii"      => "SQL_ASCII");
#
## ---------------------------------------------------------------------
#

# start the client encoding with this value
my $localeObj = Cobalt::Locale->new();

# set the database handle
$dbc = &connect($Meta_dbname, $Meta_dbuser, $Meta_dbport);
&set_encoding($client_encoding{ $localeObj->getLocale() });

#
## ---------------------------------------------------------------------
#

#### Object deiniftion section.

# Maybe this should be put into a sperate file. But parsing that would
# be a little too much overhead and I'm worried about making the module 
# slower already.

# You can define commiters, children and files for an object.
# Commuters take the object and make it real.
# Children define which other objects are the children of it.
# Files define which files belong to an object when it's dumped
# to a cobalt-object file. 

# Ins and outs of how cobalt-object files handle the uids and gids
# are yet to be defined.

# These subroutines are used for commiting objects to "reality"
# the support a new type of object add a refrence to a function
# which takes an object of this type as the sole argument and
# makes your dreams come true.
#
# Note that the responsibility for adding children of the object lies
# with these functions so that for example if we were adding the site
# we would be able to the group name for the users depending on
# any mods we had to make to the site.
#
# In the future these functions should also handle modification as well as
# addition.
my %commiters = ("vsite" => \&Cobalt::Vsite::vsite_obj_add,
		 "users" => \&Cobalt::User::site_user_add);

# If a type can have children you should define a series of hashes
# for it. The first element is the type of child, the second is the qualify
# statment to find the children.
#
# Any strings of the form [self.\w+] will be replaced by the value of that field
# in the parent object.
my %member_find = ("vsite" => { "users" => ["vsite", "=", "[self.name]"],
				"bw"	=> ["name", "=", "[self.ipaddr]"] });

# Objects may have data associated with it that is not held in the database.
# for example the web content of a virtual site, the files and home directory
# of a user.
#
# Again, this is a hash or refrences to hashes where any [self.\w+]
# occurences will be turned into their value from the database.
#
# The key of a hash is it's identifier. So that the 'webdata' files of a
# virtual site may be in a completly differnt place in a future product or even
# in the situation where RaQ users may be moved to a qube and visa versa.

my %files = ( "vsite" => { 'webdata' => "/home/sites/[self.name]/web/",
						   'logs' => "/home/sites/[self.name]/logs/",
						   'certs' => "/home/sites/[self.name]/certs/" },
			  "users" => { 'data' => "/home/sites/[self.vsite]/users/[self.name]/",
			  			   'mail' => "/var/spool/mail/[self.name]" } );

############### create()
#
# syntax:  $obj = Cobalt::Meta->create("type" => $type);
#
#		  returns an empty object belonging to database $type
#
# syntax:  $obj = Cobalt::Meta->create("type" => $type,
#									  "key"  => $value, ...);
#
#		  returns an object with corresponding ("key" => $value) traits
#
#		  THIS FUNCTION IS DEPRECATED
#
sub create
{
    my $self = shift;
    my $obj=new($self, @_);
    return $obj;
}

############### new()
#
# syntax:  $obj = Cobalt::Meta->new("type" => $type);
#
#		  returns an empty object belonging to database $type
#
# syntax:  $obj = Cobalt::Meta->new("type" => $type,
#								   "key"  => $value, ...);
#
#		  returns an object with corresponding ("key" => $value) traits
#
sub new
{
    my ($self,%args) = @_;
    my ($key,$value,$obj,$seq);

    return unless(exists($args{"type"}));

    $obj = bless {
	"type"	=> $args{"type"}
    }, $self;

    while(($key,$value) = each(%args)) {
	$obj->{$key} = $value;
    }

    return $obj;
}

############### DESTROY()
#
# syntax:  $obj->DESTROY();
#
#		  object destructor
#
sub DESTROY
{
    my $self = shift;
    return;
}

########## connect()
#
#		  connects to the postgres database using passed arguments
#
#		  database name = $dbname
#		  database user = $pguser
#		  database port = $pgport
#
#
sub connect
{
    my ($dbname, $pguser, $pgport) = @_;

    my $p;
    if (-f $Meta_id) {
	$p = `cat $Meta_id`;
    }

    if ($DEBUG) {
	print STDERR "PostgreSQL User:	  $pguser\n";
	print STDERR "PostgreSQL Port:	  $pgport\n";
	print STDERR "PostgreSQL Database:  $dbname\n";
    }

    $dbc = Pg::connectdb("user=$pguser port=$pgport dbname=$dbname password=$p");

    unless ($dbc->status eq PGRES_CONNECTION_OK) {
	MSG_error("meta_invalid_db_connection");
    }

    if ($DEBUG) {
	open(TRACE, ">> /tmp/trace.out") || die "can not open trace: $!";
	$dbc->trace(*TRACE);
    }

    return $dbc;
}

sub setpw
{
    my $dbname = shift;
    my $port = shift;
    my $user = shift;
    my $pw = shift;

    my $hba = "/home/pgsql/pg_hba.conf";

    FILE_lock_w($hba, *HBA);
    print HBA "local all trust\n";
    FILE_unlock(*HBA);

    my $conn = Pg::connectdb("user=postgres port=$port dbname=$dbname");
    $conn->exec("ALTER USER $user WITH PASSWORD $pw");

    FILE_lock_w($hba, *HBA);
    print HBA "local all crypt\n";
    FILE_unlock(*HBA);

    FILE_lock_w($Meta_id, *PGC);
    print PGC "$pw";
    FILE_unlock(*PGC);
    chown 0, 0, $Meta_id;
    chmod 0600, $Meta_id;
}

############ getdb()
#
#			returns the database handle
#
sub getdb
{
    return $dbc;
}

############### set_encoding
#
# syntax:  Cobalt::Meta::set_encoding($client_encoding);
#
#		  sets the CLIENT_ENCODING variable 
#
sub set_encoding
{
    my $encoding = shift;
    $encoding = ($encoding) ? ($encoding) : "SQL_ASCII";

    unless ($dbc->status eq PGRES_CONNECTION_OK) {
	$dbc->reset;
    }

    my $res = $dbc->exec("SET CLIENT_ENCODING TO '$encoding'");
    $ENV{ "PGCLIENTENCODING" } = $encoding;
}

############### get_encoding
#
# syntax:  Cobalt::Meta::get_encoding();
#
#		  gets the CLIENT_ENCODING variable 
#
sub get_encoding
{
    my $encoding = $ENV{ "PGCLIENTENCODING" };

    return $encoding;
}

############### query()
#
# syntax:  $obj->query("type"	  => table,
#					  "distinct"  => 1,
#					  "range"	 => [ low, high ],
#					  "sort"	  => "name",
#					  "order"	 => "dsc",
#					  "where"	 => [ <left, op, right>,
#									   <conj, left, op, right>, ... ])
#
#		  update the <key, value> pairs in the object
#		  can pass it limiting arguments in condition argument
# syntax:  $obj->query
#
#		  query the database with a more complex SQL query than $obj->retrieve
#
sub query
{
    my %kv = @_;
    my ($query,$in,$w,$result);
    my $type;

    $type = $kv{"type"};
    return 0 unless(defined($type));

    $query="SELECT";
    if (exists($kv{"distinct"}) ) {
	$query.=" DISTINCT";
    }

    my $numkeys;
    if (exists($kv{"keys"})) {
	my @keys = @{ $kv{"keys"} };
	$numkeys = scalar(@keys);
	if ($numkeys) {
	    foreach (@keys) {
		$query .= " $_,";
	    }
	    chop $query;
	} else {
	    $query .= " *";
	}
    } else {
	$query .= " *";
    }

    if (exists($kv{"table"})) {
	$query .= " INTO TABLE " . $kv{"table"};
    }

    $query .= " FROM $type";

    #
    # handle conditions on SQL queries,
    # the $kv{where} key contains arrary of
    #
    # [ <left, op, right>, <conj, left, op, right>, ... ]
    #
    # <...> denotes required grouping
    #
    if (exists($kv{"where"})) {
	my ($conj,$left,$op,$right);
	my $count = 3;
	my @conditions = @{ $kv{"where"} };
	while (scalar(@conditions) >= $count) {
	    if ($count == 3) {
		$count = 4;
		$conj = "WHERE";
	    } else {
		$conj = shift(@conditions);
	    }
	    
	    $left = shift(@conditions);
	    $op = shift(@conditions);
	    $right = shift(@conditions);
	    
	    $query .= " $conj $left $op '$right'";
	}
    }

    # table sort order determined by kv{sort}
    if (exists($kv{"sort"})) {
	$query .= " ORDER BY";
	my $sort = $kv{"sort"};
	if ($sort =~ /^[a-zA-Z]/) {
	    $query .= " lower($sort) ";
	} else {
	    $query .= " $sort ";
	}
	$query .= $kv{"order"};
    }

    # select limit determined by kv{limit}
    if (exists($kv{"limit"})) {
	my @limit = @{ $kv{"limit"} };
	my $offset = pop(@limit);
	my $count = pop(@limit);
	$query .= " LIMIT $count OFFSET $offset";
    }

    print STDERR "QUERY: '$query'\n" if ($DEBUG);

    # select from SQL table
    unless ($dbc->status eq PGRES_CONNECTION_OK) {
	$dbc->reset;
    }

    $result = $dbc->exec($query);
    
    # if we can't find it in the database return
    if (($result->resultStatus ne PGRES_TUPLES_OK) ||
	($result->ntuples < 1)) {
	return;
    }

    my ($nfields) = ($result->nfields);
    my (@row,@rows,$tobj,$k);

    # if we only want one key in the result, return a flat array
    if ($numkeys == 1) {
	while (@row = $result->fetchrow()) {
	    push @rows, $row[0];
	}
    } else {
	while (@row = $result->fetchrow()) {
	    $tobj = Cobalt::Meta->create("type" => $type);
	    for($k=0; $k<$result->nfields; $k++) {
		$tobj->put($result->fname($k) => $row[$k]);
	    }
	    push @rows, $tobj;
	}
    }
    
    # otherwise return an array of meta objects
    return(@rows);
}

############### update()
#
# syntax:  $obj->update("type"   => table,
#					   "change" => [ <key, value>, ... ],
#					   "where"  => [ <left, op, right>,
#									 <conj, left, op, right>, ... ])
#
#		  update the <key, value> pairs in the object
#		  can pass it limiting arguments in condition argument
#
sub update
{
    my %opt = @_;

    my $type = $opt{"type"};
    my (@changes,@conditions);
    my ($change,$condition) = ('','');
    my ($conj,$left,$op,$right);
    my $args = 3;

    # the changes are a list of <key, value> pairs to change
    # format:  [ <key, value>, ... ]
    @changes = @{ $opt{"change"} };
    while (scalar(@changes) >= 2) {

	$left = shift(@changes);
	$right = shift(@changes);

	$change .= " $left = '$right',";
    }
    chop $change;

    # the conditions are a list of <(conj),left,op,right> groups
    # that limit the search for rows to update
    # format ex:  [ <left, op, right> ]
    # general ex: [ <left, op, right>, <conj, left, op, right>, ... ]
    @conditions = @{ $opt{"where"} };
    while (scalar(@conditions) >= $args) {
	if ($args == 3) {
	    $conj = '';
	    $args = 4;
	} else {
	    $conj = shift(@conditions);
	}

	$left = shift(@conditions);
	$op = shift(@conditions);
	$right = shift(@conditions);

	$condition .= " $conj $left $op '$right'";
    }

    my $query = "UPDATE $type SET $change WHERE $condition";
    print STDERR "QUERY:  $query\n" if ($DEBUG);

    # update SQL table
    unless ($dbc->status eq PGRES_CONNECTION_OK) {
	$dbc->reset;
    }

    my $result = $dbc->exec($query);
    
    # return the number of rows updated
    my $r = $result->cmdStatus;
    $r =~ s/UPDATE (.*)/$1/;
    return $r;
}

############### retrieve()
#
# syntax:  $obj->retrieve(name, [ <conj, left, op, right>, ... ]);
#
#		  build the object by connecting to the SQL database of object "type"
#
#		  can pass it more limiting arguments in
#		  [ <conj, left, op, right>, ... ] groups
#
sub retrieve
{
    my ($self,$name,@conditions) = @_;
    my ($result,$k,@row,$query);
    my $type = $self->gettype;

    my ($left,$op,$right,$conj,$condition);

    while (scalar(@conditions) >= 4) {
	$conj = shift(@conditions);
	$left = shift(@conditions);
	$op = shift(@conditions);
	$right = shift(@conditions);

	$condition .= " $conj $left $op '$right'";
    }

    $query = "SELECT DISTINCT * FROM $type WHERE name = '$name' $condition";

    print STDERR "QUERY:  $query\n" if ($DEBUG);

    # select from SQL table
    unless ($dbc->status eq PGRES_CONNECTION_OK) {
	$dbc->reset;
    }

    $result = $dbc->exec($query);
    
    # if we can't find it in the database return
    unless($result->resultStatus eq PGRES_TUPLES_OK && $result->ntuples) {
	return;
    }

    # now populate the object
    @row = $result->fetchrow();
    for ($k = 0; $k < $result->nfields; $k++) {
	$self->put($result->fname($k) => $row[$k]);
    }

    return $self;
}

############### get()
#
# syntax:  $value = $obj->get($key);
#
#                 returns the value of $self->{$key} if found,
#                 nothing if it can't be found
#
sub get
{
    my ($self,$key) = @_;
    if ($key) {
        if ($self->{$key}) {
	    return($self->{$key});
        } else {
	    return;
        }
    } else {
        return;
    }
}

############### getall()
#
# syntax:  @objects = $obj->getall;
#
#		  returns an array of all objects in the table
#		  accessable by name "name"
#
sub getall
{
    my ($self,$type) = @_;
    $type = $self->{type} unless (defined($type));
    my ($key,%h,@row,$result,$k,@all,$single);
    
    unless ($dbc->status eq PGRES_CONNECTION_OK) {
	$dbc->reset;
    }

    return unless($result = $dbc->exec("SELECT * FROM $type"));
    
    while(@row = $result->fetchrow) {
	$single = Cobalt::Meta->new("type" => $type);
	for ($k = 0; $k < $result->nfields; $k++) {
	    $single->put($result->fname($k) => $row[$k]);
	}
	push @all, $single->{"name"};
    }
    
    return @all;
}

############### getname()
#
# syntax:  $name = $obj->getname;
#
#		  returns the name of the object
#
sub getname
{
    my $self = shift;
    if ($self->{"name"}) {
	return($self->{"name"});
    } else {
	return MSG_error("meta_invalid_objectname");
    }
}

############### gettype()
#
# syntax:  $type = $obj->gettype;
#
#		  returns the name of the object
#
sub gettype
{
    my $self = shift;
    if ($self->{"type"}) {
	return($self->{"type"});
    } else {
	return MSG_error("meta_invalid_objecttype");
    }
}

############### put()
#
# syntax:  $obj->put("key" => "value", ...);
#
#		  insert object ("key",$value) pairs
#
sub put
{
    my ($self,%kv) = @_;
    foreach (keys %kv) {
	$self->{$_} = $kv{$_};
    }
    return $self;
}

############### delete()
#
# syntax:  $obj->delete();
#
#		  delete object from database
#		  based on objects name and type
#
sub delete
{
    my $self = shift;
    my ($result,$name,$type);
    
    $name = $self->getname;
    $type = $self->gettype;
    
    unless ($dbc->status eq PGRES_CONNECTION_OK) {
	$dbc->reset;
    }

    $result = $dbc->exec("DELETE FROM $type WHERE name = '$name'");
    
    return $self;
}

############### vacuum()
#
# syntax:  Cobalt::Meta::vacuum();
#
#          vacuums database to keep it fast and clean
#
sub vacuum
{
    unless ($dbc->status eq PGRES_CONNECTION_OK) {
	$dbc->reset;
    }

    $dbc->exec("VACUUM ANALYZE");

    return;
}

############### drop()
#
# syntax:  Cobalt::Meta::drop($table, ...);
#
#          drops tables from the sql database
#
sub drop
{
    my @tables = @_;

    unless ($dbc->status eq PGRES_CONNECTION_OK) {
	$dbc->reset;
    }

    foreach (@tables) {
	$dbc->exec("DROP TABLE $_");
    }

    return;
}

############### save()
#
# syntax:  $obj->save();
#
#		  interfacing to a postgresql database, write the object
#		  to database table specified by $self->{type}
#
#		  all information is dynamically generated from the object
#
sub save
{
    my $self = shift;
    my ($key,$val,$keys,$vals,$struct,$string,$result,$seq);
    my $type = $self->gettype;
    my $name = $self->getname;
    my $ret = 0;
    my $i = 1;

    $keys=""; $vals=""; $struct="";
    while (($key,$val)=each %$self) {
	next if ($key =~ /modify/);
	if ( $key eq 'members' ) {
	    # Woo! We have a members hash, let's be smart and save that as well.
	    foreach my $member ( @{$$self{'members'}} ) {
		$member->save();
	    }
	    next;
	}
	print STDERR "	table key = ", $key, " value = ", $val, "\n"  if ($DEBUG);
	if ($val =~ /^\d/) {
	    if ($val =~ /\d+\.\d+\.\d+\.\d+/) {
		$struct .= " $key inet,";
		$vals   .= " \'$val\',";
	    } elsif ($val =~ /\./) {
		$struct .= " $key float8,";
		$vals   .= " \'$val\',";
	    } else {
		$struct .= " $key int4,";
		$vals   .= " \'$val\',";
	    }
	} else {
	    $struct .= " $key text,";
	    $vals   .= " \'$val\',";
	}
	$keys .= " $key,";
    }
    $keys .= " modify";
    $vals .= " datetime('now'::abstime)";
    $struct .= " modify datetime";

    print STDERR "CREATE TABLE $type ($struct)\n" if ($DEBUG);
    unless ($dbc->status eq PGRES_CONNECTION_OK) {
	$dbc->reset;
    }

    $result = $dbc->exec("CREATE TABLE $type ( $struct )");

    print STDERR "created table, status = ", $result->cmdStatus, "\n" if ($DEBUG);

    # delete existing row
    $self->delete;

    print STDERR "INSERT INTO $type ($keys) VALUES ($vals)\n" if ($DEBUG);
    unless ($dbc->status eq PGRES_CONNECTION_OK) {
	$dbc->reset;
    }

    $result = $dbc->exec("INSERT INTO $type ($keys) VALUES ($vals)");

    unless($result->resultStatus eq PGRES_COMMAND_OK) {
	return MSG_error("meta_table_cant_insert");
    }
    
    if($DEBUG) {
	print STDERR "insert into table, last oid = ", $result->oidStatus, "\n";
	unless ($dbc->status eq PGRES_CONNECTION_OK) {
	    $dbc->reset;
	}

	$result = $dbc->exec("COPY $type TO STDOUT");

	printf(" copy table to STDOUT:\n");

	$ret = 0;
	while(-1 != $ret) {
	    $ret = $dbc->getline($string,256);
	    last if $string eq "\\.";
	    print STDERR "  ", $string, "\n";
	    $i++;
	}
    }

    $dbc->reset;

    return $self;
}

############### display()
#
# syntax:  $obj->display();
#
#		  output the object to STDOUT
#
#		  DEPRECATED:  use $obj->as_xml();
#
sub display
{
    my $self = shift;
    my $key;

    print STDERR "object:  BEGIN OUTPUT\n";
    foreach $key (keys %$self) {
	printf STDERR "object:	key = ".$key." \t val = ".$$self{$key}."\n";
    }
    print STDERR "object:  END OUTPUT\n";

    return $self;
}

############### as_xml()
#
# syntax:  $obj->as_xml();
#
#		  returns the object as a simple XML block.
#		   any child objects are nested inside it.
sub as_xml()
{
    my $self = shift;
    my $depth = shift;
    $depth ||= 0;
    my $tabs = "\t" x $depth;
    my $xml;
    my $key;
    my $member;
    
    if (! $$self{'type'} ) {
	return 0;
    }
    
    $xml =  "$tabs<$$self{'type'} name=\"$$self{'name'}\">\n";
    foreach $key (keys %$self) {
	if ( $key eq 'members' ) {
	    foreach $member ( @{$$self{'members'}} ) {
		$xml .= $member->as_xml( $depth + 1 );
	    }
	} elsif ( $key ne 'name' || $key ne 'type') {
	    $xml .= "$tabs\t<$key>$$self{$key}</$key>\n";
	}
    }
    $xml .= "$tabs</$$self{'type'}>\n";
    return $xml;
}

############## from_xml_elem()
#
# syntax: $obj = Cobalt::Meta=>from_xml_elem( $xml_element );
#
sub from_xml_elem()
{
    my $self = shift;
    my $xml = shift;
    
    my $parser = new XML::Parser(Handlers => {Start => \&xml_start,
					      End => \&xml_end,
					      Char => \&xml_char}
				 );
    
    # Pars calls the xml_* functins which will manipulate and call the global
    # XML_obj
    $parser->parse($xml);

    return get_xml_obj();
}

############## commit();
#
# syntax: $ret = $obj->commit();
#
# If the Object is of a supported type, commit it to disk and return
# a status string '\d{4} \w+'
sub commit()
{
    my $self = shift;
    if (defined $commiters{$self->gettype}) {
	return &{$commiters{$$self{'type'}}}($self);
    } else {
	return MSG_error("meta_cant_commit_obj_not_supported");
    }
}
########## populate_members()
#
# $obj->populate_members();
#
# Objects must have their members specifically requested or added.
sub populate_members()
{
    my $self = shift;
    foreach ( $self->get_members() ) {
	$self->add_member($_);
    }
}

######### get_members()
#
# @members = $obj->get_members( type=>'vsite', condition=>["ip", "=", "foo"])
#
# 
sub get_members
{
    my $self = shift;
    my %args = @_;
    my @condition;

    if ( !defined $args{"type"} ) {
	# Without a defined type we return all the members we have. Regardless
	# of race conditions, choice of text editor or window manager.
	# ( race, religion or colour. )
	my (@members,$sub_type,@cond);
	
	foreach $sub_type (keys %{$member_find{$self->gettype}}) {
	    
	    @cond = @{ $member_find{$self->gettype}{$sub_type} };
	    foreach (@cond) {
		# Replace the [self.value] lines.
		$_ =~ s/\[self\.(\w+)\]/$self->get($1)/ge;
	    }
	    
	    # Now we grab out everything as a query.
	    foreach (Cobalt::Meta::query("type"  => $sub_type,
					 "where" => \@cond)) {
		# If that object has sub members, we get them too..
		# Recursion is fun ;).
		$_->get_members();
		push @members, $_;
	    }
	}
	return @members;
    }
    
    # If we have a type arg, then we're lookin for specific members.
    # not just everything we can find.
    my @cond = @{ $member_find{$self->gettype}{$args{"type"}} };
    foreach (@cond) {
	# Replace the [self.value] lines.
	$_ =~ s/\[self\.(\w+)\]/$self->get($1)/ge;
    }

    if (defined $args{"where"}) {
	unshift @cond, "and";
	push @{$args{"where"}}, @cond;
    }

    # Now we just return whatever the query returns.
    return $self->query(%args);
}

############ get_parents()
#
# @parents = $obj->get_parents();
#
# Sometimes you've just got to know who you're daddy is.
sub get_parents
{
    my $self = shift;
    my ($parent,$qualifiers);
    my @parents;
    
    # This is a bit nasty, but to get it right we have to go through and find
    # out which objects can have objects of our type as children.
    while ( ($parent, $qualifiers) = each %member_find ) {
	if ( defined $$qualifiers{$self->gettype} ) {
	    my (@condition,$conj,$left,$op,$right);
	    my @cond = @{ $$qualifiers{$self->gettype} };
	    # This should be something like " vsite = '[self.name]' "
	    # we want to turn it into " name = '[self.vsite]'"
	    # Beleive it or not the line below does that.
	    my $arg = 3;
	    while (scalar(@cond) >= $arg) {
		if ($arg == 3) {
		    $conj = '';
		    $arg = 4;
		} else {
		    $conj = shift @cond;
		}
		my $left = shift @cond;
		my $op = shift @cond;
		my $right = shift @cond;
		
		if (defined($conj)) {
		    push @condition, $conj;
		}
		
		$right =~ /\[self\.(\w+)\]/;
		push @condition, $1;
		push @condition, $op;
		push @condition, $self->get($left);
	    }
#			   $qualify =~ s/([^\ ]+)\s+([^\ ]+)\s+\'\[self\.([^\ \]]+)\]\'/
#			   "$3 $2 '".$self->get($1)."'"/gex;
#			   print STDERR "PARENT_SEARCH " if $DEBUG; # query()'s debugging will pop
	    # the information about the actual query on for us.
	    # Now we get the objects we want with a new query ;).
	    @parents = Cobalt::Meta::query(type => $parent, "where" => \@condition);
	}
    }
    return @parents;
}

# Given a file handle to write to will dump out a tar file of all the data
# associated with an object.
sub data {
    my $self = shift;
    my $fh = shift;
    my $tarcmd;
    my $buffer;
    my ($directory, $identifier);

    # We need some way to know which uids/gids match which objects,
    # site users may own files in a web data area for example.
    print $fh "##### BEGIN COBALT DATA ID MAP #####\n";
	if ( $self->get("uid") ) {
	    print $fh 'uid:'.$self->get("uid").'='.$self->getname()."\n";
	}
    if ( $self->get("gid") ) {
	print $fh 'gid:'.$self->get("gid").'='.$self->getname()."\n";
    }

    foreach ( $self->get_members() ) {
	if ( $_->get("uid") ) {
	    print $fh 'uid:'.$_->get("uid").'='.$_->getname()."\n";
	}
	if ( $_->get("gid") ) {
	    print $fh 'gid:'.$_->get("gid").'='.$_->getname()."\n";
	}
    }
    print $fh "##### END COBALT DATA ID MAP #####\n";

	print $fh "##### BEGIN COBALT DATA FOR ";
    print $fh $self->gettype() . " " . $self->getname();
    print $fh " #####\n";

	# For the puposes of debugging for now this just exports
	# a lits of the files that *would* be in this data block
	$tarcmd = '/bin/tar cz ./';

    while ( ($identifier,$directory) = each %{$files{$self->gettype}} ) {
	print "##### BEGIN COBALT DATA BLOCK FOR $identifier #####\n";
	    $directory =~ s/\[self\.(\w+)\]/$self->get($1)/ge;

	# If we're dumping a direcctory.
	if ( $directory =~ /\/$/ ) {
	    if ( chdir $directory ) {
		open(DATA, "$tarcmd 2> /dev/null |") ||
		    return MSG_error("meta_cant_tar");
	    }
	} else {
	    # Else we're dumping a file
	    open(DATA, "<$directory");
	    # Don't worry if it's not there, we don't care.
	}

	while (read(DATA, $buffer, 57)) {
	    print $fh encode_base64($buffer);
	}

	print "##### END COBALT DATA BLOCK FOR $identifier #####\n";
	}
}

sub restore_data
{
    my $self = shift;
    my $fh = shift;

    my $line;
    my $location;

    my $tarcmd = '/bin/tar x';
    
    while (<$fh>) {
	if (/##### BEGIN COBALT DATA BLOCK FOR (.*) #####/o) {
	    $location = $1;
	    if ( $files{$self->gettype}{$location} =~ /\/$/ ) {
		#open(DATA, "|$tarcmd -C $location") ||
		#	return MSG_error("meta_cant_tar");
		print STDERR "Untarring to -C " . $files{$self->gettype}{$location} . "\n";
		open(DATA, "|/bin/tar tz 1>&2");
	    } elsif ( $files{$self->gettype}{$location} =~ /[^\/]$/) {
		#open(DATA, ">$location") ||
		#	warn MSG_error("meta_cant_write");
		print STDERR "Dumping file to $location.\n";
	    } else {
		# Eeek! No idea where to dump this.
		warn MSG_error("meta_unknown_data_identifier");
	    }
	    # bloody perl.
	    $line = <$fh>;
	    while( $line !~ /##### END COBALT DATA BLOCK .* #####/o ) {
		   print DATA decode_base64($line);
		   $line = <$fh>;
	       }
	}
    }
}

sub object
{
    my $self = shift;
    my $fh = shift;

    print "##### BEGIN COBALT OBJECT #####\n";

	print $fh $self->as_xml();

    $self->data( $fh );

    foreach ( $self->get('members') ) {
	print $fh $_->object();
    }
}

sub add_member
{
    my $self = shift;
    my $new_obj = shift;
    if ( ! defined $self->{'members'} ) {
	$$self{'members'} = [];
    }
    push @{$$self{'members'}}, $new_obj;
}

# Our xml parsing handlers and variables they work with live here in their
# own block so as to minimise name space polution.
my @swap;
my $xml_obj;
my $type;

sub xml_start {
    # First time we see a 'start' element, it's for what type of object
    # this is.
    my $parser = shift;
    my $section = shift;
    my %args = @_;
    # Only tags of type "<type name="name">" start of blocks.
    if ( $args{'name'} ) {
	if ( defined $$xml_obj{'type'} ) {
	    # We currently have an object being pricessed, dump it on the stack
	    # to be picjed up later.
	    #print STDERR "Pushing object onto stack\n";
	    push @swap, $xml_obj;
	}
	#print STDERR "Creating $section object with";
	#while (my($key) = each(%args)) {
	#	print ", $key = $args{$key}";
	#}
	#print "\n";
	$xml_obj = Cobalt::Meta->new(type => $section,
				     name => $args{'name'});
	return;
    }
}

sub xml_end {
    my $parser = shift;
    my $section = shift;
    if ( $section eq $$xml_obj{'type'} ) {
	# If we're ending a section for the type of object we're currentl parsing we're
	# done with this block.
	#print "Committing change to object of type ".$xml_obj->gettype()."\n";
	if ( scalar( @swap ) ) {
	    # O.K. We were parsing a sub block.
	    # Pop our parent block off the stack..
	    #print "Popping parent off the stack..\n";
	    my $swap = pop @swap;
	    # Add the current block as a member of the parent block.
	    #print "Adding child as a member of parent..\n";
	    $swap->add_member($xml_obj);
	    # And make the parent block the current so we can continue where we left off.
	    #print "Turning parent into current object..\n";
	    $xml_obj = $swap;
	}
	# If we were at the tope of the block we do nothing.
    }
}

sub xml_char {
    my $parser = shift;
    my $section = shift;
    # Skip white space.
    if ( $section =~ /^\s+$/) {
	return;
    }
    $xml_obj->put($parser->current_element(), $section);
    #print "Setting ".$parser->current_element()." to $section\n";
}

sub get_xml_obj {
    return $xml_obj;
}

1;

__END__


