#
# $Id: Monitor.pm,v 1.50 1999/10/02 16:44:26 duncan Exp $
#
# Routines for keeping track of the system's state
#
# Written by Brion Moss (brion@cobaltnet.com, brion@nda.com)
#
# Copyright 1998 Cobalt Networks http://www.cobaltnet.com
#

###### Overview of the Active Monitor system:
#
# Last updated 10/23/98 Brion Moss
#
# What does it do?
#
#   Active monitor watches the state of the system, notifies the admin
#   of problems, and where possible makes corrections for error conditions.
#
# What sort of notifications does it send?
#
#   1. Email (if enabled) 
#      Email is sent to the "sys" mailing list.  By default, this list
#      consists of just the "admin" user.  A future release may allow the
#      user to set the membership of this list.
# 
#    2. Flashing icon on the UI
#       There is a flashing 'Alarm' icon when admin connects to the server.
#       Clicking on the flashing 'Alarm' takes the admin. to the 'Maintenance'
#       page where the status of each of the services is indicated with an 
#       indicator light: 
#
#       Green  - OK 
#       Yellow - Warning 
#       Red    - Error 
#
#       Depending on the service and the product, there will be message
#       text associated with the light telling the admin what is going on.
#       In some cases, clicking on the lights takes you the detailed 
#       description of the warning/error/suggested corrective action.  
# 
# Where is all this information kept?
#
#       Active Monitor keeps all of its state information in a DB file
#       /var/cobalt/status.  It also (depending on the build) logs
#       everything into .log files under /var/cobalt; these files are
#       not used by the UI, and the sauce only uses them in that it
#       writes additional information to them.  However, they can be
#       useful to advanced users or to Cobalt Technical Support.
#       Extra output (debugging output or just captured program
#       output) goes into /var/cobalt/sauce.log.
#
#       The database format is discussed in more detail below.
#
# What is monitored?
#
#   1. System  
# 
#      a) Disk- Sends alarm if usage exceeds threshold.  This is
#        measured on a per-partition basis, and also checks on
#        disk usage for each user and group.  Admin will always
#        be notified; users and groups will also be notified if
#        they are going near their quota.
# 
#      b) Network - Sends alarm if collisions exceeds threshold
#        Statistics are kept for overall collision rate (since
#        the system was booted) and for the collision rate over
#        the last 15 minutes.
# 
#        Because of the way the kernel records collisions, the
#        collision rate is not a very useful number on these systems.
#        The problem is that retries are logged as seperate collisions,
#        so you can easily have >100% collision rates even under
#        moderate load. 
#
#      c) CPU - Sends alarm if CPU utilization exceeds threshold.
#        This is based on load average over the past 15 minutes.
#        For recommended action, we first look at swapping.  If
#        it's a matter of too much swapping, we warn about 
#        needing more memory; otherwise we warn that the CPU is
#        overloaded.
# 
#      d) Memory - Sends alarms if excessive swapping, or if
#        too much virtual memory is being used.  Swapping is
#        determined in terms of number of swaps per second over
#        a 15-minute interval.
#  
#    2. Services (answering requests, error log parsing to be implemented) 
#       Mail server, DNS server, etc.  The list of services checked
#       by Active Monitor is defined in the @Services array in Product.pm.
# 
#    3. Important system processes (just make sure they're running)
#       crond, portmap, etc.; the @Hidden_servers array has a list
#       of these.  These are generally not reported to the user, just
#       fixed quietly.
#  
#    4. Security (not implemented, no sign that it will be soon)
#       a) Excessive failed logins/authorization on Web, telnet, ftp
# 
# 
# DEFINITIONS:
# 
# Some terms used in the monitoring system:
# 
# - Service: a "service" is any function provided by the qube which we
#   can monitor.  Examples include "disk", "cpu", "ftp", and "lcd".
# 
# - Instance: an "instance" is a sub-category of a particular service.
#   For example, the "disk" service has an "instance" for each user to
#   describe the usage of each user.  The "mail" service has separate
#   instances for the SMTP server, the POP3 server, and the IMAP server.
# 
# - Threshold: Many services can be thought of in simple terms of
#   "works" and "broken", but others have certain statistics associated with
#   them that get monitored.  For example, the CPU is judged by it's
#   load average--if the load average is too high, an alarm is
#   triggered.  For these services, there is a "yellow alert" threshold,
#   and a "red alert" threshold.  The actions taken are identical, but
#   the different alert levels help the user to understand how bad
#   things are.
# 
# ---------------------------------------------------------------------
# 
# PROCESS:
# 
# The monitoring script (/usr/local/sbin/monitor) is run once every 15
# minutes out of cron.  There is a second script, /usr/local/sbin/custodiat,
# which runs as a standalone daemon and makes sure that crond stays
# running.
# 
# The flow of the monitor script is as follows:
# 
# 1) The script first checks to make sure that it isn't running during
#    bootup or shutdown.  In either case, services may have be
#    legitimately not responding, but the user shouldn't be confused by
#    being alerted about this.
# 
# 2) Next, the script checks to see if there's an existing monitor process.
#    It gets the PID from /etc/locks/monitor, and checks in /proc to see
#    if it's actually running.
# 
# 3) The lock directory is cleaned out.  Any files in /etc/locks which
#    have been around for a certain age are removed.
# 
# 4) Next, monitor creates a lock file so that there is no danger of
#    concurrent monitoring processes colliding.
# 
# 5) Finally, the testing begins.  Each service is examined, using the
#    Cobalt::Repair::repair_examine() function from Repair.pm.  
#    In early versions of the product, most disk checks were
#    only done once an hour, to reduce load, and a few (the du of user
#    and group areas) were only done once daily out of /etc/cron.daily.
#    Now the quota system is used for near-instantaneous response.
# 
# 6) Once the state of each service has been determined and entered in
#    the database, the script tries to repair any services which have
#    registered an alert condition.  The following steps are taken twice
#    for each problem service:
# 
#  a) The service is passed to the Cobalt::Repair::repair_fix() function, 
#     which (according to the service) will try to restart the service, etc.
# 
#  b) The service is Cobalt::Repair::repair_examine()ed again, to see if 
#     Cobalt::Repair::repair_fix() was successful. 
# 
#  c) If this is the second time through this loop, the administrator
#     (and whoever else needs be told) is notified of the problem
#     using the Cobalt::Repair::repair_notify() function.  Exception: disk
#     checks pass on to the notify stage immediately, as there is no way
#     to "repair" a full disk without intervention.
# 
#     Notification of any given problem should be sent exactly once
#     per day.  If the problem occurs, gets fixed, and re-occurs, 
#     a repeat notification will be sent.
# 
# 7) The system alert page in the UI is rebuilt with the new state of
#    each of the services.
# 
#---------------------------------------------------------------------
# NOTES:
# 
# o Process cache: in order to reduce the overhead of the initial pass
#   through the monitoring script, a cache of what processes are running
#   is created, and all of the process monitoring code refers to this in
#   the initial Cobalt::Repair::repair_examine() calls.  In the repair 
#   stage (step 5, above), the cache is removed and all checks to see if 
#   a process is running are made against a live call to the ps command.
# 
# o Recommendations: The messages generated by the alert system are all
#   kept in a separate module, Messages.pm.  The function msg_get_message is
#   used to get the recommended action for any given service.
# 
# o Retries: Cobalt::Repair::repair_fix() will try a number of times
#   (currently 7) to fix any problem services.  After that, it will give 
#   up and wait 24 hours before clearing its retry counter and starting 
#   the process again. (the numbers 7 tries and 24 hours are coded into 
#   Product_*.pm)
#
# ---------------------------------------------------------------------
# 
# SERVICES
#   Note that this list is getting added to constantly, so this list
#   may not be fully up-to-date.  Also, not all services or instances
#   are supported on all products.
# 
#   Service   Instance   Description
#   -------   --------   -----------
#   adm       adm        Administrative (port 81) web server (httpd)
#   atalk     nbplkup    Appletalk file services
#   cpu       load       CPU load (load average) over the past 15 minutes
#   crond     crond      cron daemon
#   custo     custo      custodiat script (cron watcher)
#   disk      /          usage of the system disk (/)
#   disk      /var       usage of the spool/log disk (/var)
#   disk      /home      usage of the user disk (/home)
#   disk      users      state of the worst user
#   disk      *,0        per-user disk usage (e.g. "disk","admin,0")
#   disk      groups     state of the worst group
#   disk      *,1        per-group disk usage
#   disk      *,mail     per-user mail spool usage 
#   disk      <dir>      Usage in bytes of the directory (user top or group top)
#   dhcpd     dhcpd      DHCP server
#   dns       dns        DNS port (53)
#   dns       named      DNS server process
#   ftp       ftp        ftp port (run from inetd)
#   inetd     inetd      inetd process
#   lcd       lcdsleep   lcdsleep process
#   mail      SMTP       SMTP server (sendmail)
#   mail      POP3       POP3 server (run from inetd)
#   mail      IMAP       IMAP server (run from inetd)
#   mem       swapping   Number of swaps per second
#   mem       vmem       Amount of virtual memory used
#   modem     diald      the diald process
#   net       collisions Collision rate (%) since boot
#   net       col15      Collision rate (%) for past 15 minutes
#   nfs       rpc.nfsd   nfsd process
#   nfs       rpc.mountd mountd process
#   nfs       showmount  showmount test results
#   ntp       ntpd       ntpd process
#   portmap   portmap    portmap process
#   smb       smb        connections to the smb port
#   smb       smbd       smbd process
#   smb       nmbd       nmbd process
#   snmp      snmp       SNMP process
#   ssl       ssl        SSL daemon (port)
#   telnet    telnet     telnet daemon
#   www       www        main web server port
# 
# lcd, inetd, crond, portmap, and custo are not reported in the UI.  All
# of the rest are.
#
# Due to problems in the collision reporting with the Cobalt ethernet
# drivers, the "net" service is not currently reported on.
#
#########
#
# At present, the sauce for Active Monitor consists of:
#
#  Monitor.pm -- database routines and functions to check services.
#                Used by Repair.pm and the monitor script, and by 
#                any CGIs that need to read information from the DB.
#  Repair.pm  -- wrapper routines called by the monitor script
#                to examine, repair, or complain
#  monitor    -- the script that runs every 15 minutes to check everything
#  custodiat  -- the script that runs as a daemon to keep cron running
# 
# ---------------------------------------------------------------------
# Database format:
#
#    server|instance => condition|time condition began|time of last check|
#                        notified|number of repair attempts|reserved|data
#    (data is a ,-seperated list of additional status information)
#    (reserved is for future expansion)
#
# There is a special PRIME instance that represents the service as a whole.
# For example, the mail service has smtp at condition red, pop at yellow,
# and imap at green, then mail|PRIME is at red.  The instance which PRIME
# represents (i.e., the worst instance) is named in PRIME's $data field.
# Generally, when no instance is specified in a query, the values for the 
# PRIME instance are the ones returned.
#
# There is also a LIST instance that is a comma-seperated list of all the 
# instances within the service.  For example, mail|LIST would be 
# "smtp,pop,imap".
#
# The condition will be:
#     -1     No state information -- we just booted and can't say
#      0     Nothing going on -- service is deactivated (or idle, see below)
#      1     Green  -- things are just fine
#      2     Yellow -- warning state, we should start to worry
#      3     Red    -- error condition, things are broken
# The only time the difference between -1 and 0 is significant is for
# reporting the state of the network interface, where we need to be
# able to report "we are idle" as distinct from "we have not checked yet".
# Since we don't currenly examine the network interface, really, we
# could theoretically just get rid of this distinction.
#
# The "time condition began" should reflect any state change; any time
# the condition field changes.  This will be expressed in seconds since
# the epoch.  This field gets reset to zero (or just ignored) if we
# have not checked a service since the system rebooted.
#
# The "time of last check" is set when we examine a service.  If a service
# is inactive or unsupported we don't check it, so this won't get
# updated.  This is also in seconds since the epoch.
#
# Notified is a boolean (0 or 1) of wether we've sent an email notification
# that this service is having a problem.  This field gets reset to 0
# when the condition is not yellow or red (e.g., if it's green or unchecked).
#
# Number of attempts is a counter of how many times we've tried to do
# something about a problem service.  We use this to figure out when
# to give up; once we've tried several times it is not considered useful
# to keep annoying the service by trying again.  This gets reset
# when the condition goes green.  It also gets reset if the condition
# is over a day old (actually, after $Stat_lim_expire seconds).
#
# I stuck in a "reserved" field so that if we add something later
# we don't get stuck with massive modifications to the monitoring
# system...again.
#
# The data field has different things for different servers.  It is a
# comma-seperated list of additional information we may want (or need)
# to store.  Here is what you'll find:
#
#     service        data contents
#     -------        -------------
#     filesystem     %capacity, freespace
#     userspace      %used,used,max
#     groupspace     %used,used,max
#     dirspace       size
#     mailspace      size
#     cpu|load       15-minute average
#     net|col15      tx,cols,(from prev. test period),%collisions in past 15
#     net|collisions %collisions overall
#     mem|vmem       %used
#     mem|swapping   swap count from prev. test period, swaps/second
#
# (PRIME stores the name of the instance it represents in its data field)
#
# There are some special services (which are only applicable within
# Cobalt::Monitor::mon_query()):
#
#    memberspace is the state of the group and all its members
#    sitespace is the state of all the sites (all the memberspaces)
#    diskspace is the state of groupspace, userspace, and filesystem

package Cobalt::Monitor;
use vars qw(@ISA @EXPORT @EXPORT_OK);
require Exporter;
require SelfLoader;
@ISA    = qw(Exporter SelfLoader);
@EXPORT = qw(mon_query mon_get_state mon_set_state mon_list_instances
	     mon_diskwatch mon_quotawatch mon_loadwatch mon_collwatch
	     mon_memwatch mon_procwatch mon_swapwatch mon_spoolwatch
	     mon_areastat mon_cronwatch mon_portwatch mon_nfstest
	     mon_atalktest mon_nslookuptest mon_telnettest mon_modem);

use DB_File;
use Fcntl qw(O_CREAT O_RDWR O_RDONLY);
use integer;   # we don't need floating point
use Socket qw(PF_INET SOCK_STREAM inet_aton sockaddr_in);

require Cobalt::Language;
import Cobalt::Language qw(:DEFAULT);
require Cobalt::List;
import Cobalt::List qw(list_get_mem);
require Cobalt::Network;
import Cobalt::Network qw(net_get_hostname);
require Cobalt::Quota;
import Cobalt::Quota qw(repquota);
require Cobalt::Vsite;
import Cobalt::Vsite qw(vsite_list vsite_get_users);
use Cobalt::Util;
use Cobalt::Product;

use DB_File;
use Fcntl qw(O_CREAT O_RDWR O_RDONLY);
use integer;   # we don't need floating point
use Socket qw(PF_INET SOCK_STREAM inet_aton sockaddr_in);
use Locale::gettext;
Locale::gettext::textdomain("special-sauce");

use subs qw(PF_INET SOCK_STREAM);

1;

__DATA__

#######################################################################
#
# Database functions

sub mon_query
# Get the state of a service for display in the UI
# Arguments: service or (service,instance)
#   Given just the service return the most severe instance within
#   the service (e.g., any red, then any yellow, then any green)
# Return value: condition (red/yellow/green), message text
# Example: ($state,$text) = Cobalt::Monitor::mon_query("mem");
{
  my ($service,$instance) = @_;
  my ($condition,$cond_time,$update_time,$notified,$attempts,$reserved,$data);

  my %STATUS;
  my $fit_to_be_untied = 0;
  if (-e "$DB_dir_root/$service.state") {
    # Do it the text file way
    if (open(US, "$DB_dir_root/$service.state")) {
      while (<US>) {
	chomp;
	my @parts = split(/\|/o, $_, 2);
	$STATUS{"$service|$parts[0]"} = $parts[1];
      }
      close(US);
    }

    # Special case services that include other services
    if (($service eq "memberspace") ||
	($service eq "sitespace") ||
	($service eq "diskspace")) {
      if (open(US, "$DB_dir_root/userspace.state")) {
	while (<US>) {
	  chomp;
	  my @parts = split(/\|/o, $_, 2);
	  $STATUS{"userspace|$parts[0]"} = $parts[1];
	}
	close(US);
      }
      if (open(US, "$DB_dir_root/groupspace.state")) {
	while (<US>) {
	  chomp;
	  my @parts = split(/\|/o, $_, 2);
	  $STATUS{"groupspace|$parts[0]"} = $parts[1];
	}
	close(US);
      }
    }
    if ($service eq "diskspace") {
      if (open(US, "$DB_dir_root/filesystem.state")) {
	while (<US>) {
	  chomp;
	  my @parts = split(/\|/o, $_, 2);
	  $STATUS{"filesystem|$parts[0]"} = $parts[1];
	}
	close(US);
      }
    }
  }
  else {
    # Do it with da database...
    unless(tie(%STATUS, 'DB_File', $Status_db, O_RDONLY(),0644)) {
      return (0,MSG_error("monitor_cant_read_status_db", $Status_db));
    }
    $fit_to_be_untied = 1;
  }

  # If we aren't given an explicit instance, take the most severe
  # as stored in the PRIME instance
  unless ($instance) {
    my @ins = split(/\|/o,$STATUS{"$service|PRIME"});
    $instance = $ins[6];  # PRIME stores the instance name in the data field
  }

  # Yey.  Special cases.  This is how we handle nested alert lights.
  if ($service eq "memberspace") {
    # This is the condition of the given group and all its members
    # first set the condition to that of the group
    ($condition,$cond_time,$update_time,$notified,$attempts,$reserved,$data) =
      split(/\|/o,$STATUS{"groupspace|$instance"});
    $service = "groupspace";
    my $user;
    # override with the most severe condition among the users
    # (if more severe than the group condition)
    if ($Features{'vmachine'}) {
      foreach $user (vsite_get_users($instance)) {
	my ($u_condition,$u_cond_time,$u_update_time,$u_notified,$u_attempts,
	    $u_reserved,$u_data) = split(/\|/o,$STATUS{"userspace|$user"});
	if ($u_condition > $condition) {
	  ($condition,$cond_time,$update_time,$notified,$attempts,
	   $reserved,$data) = 
	     ($u_condition,$u_cond_time,$u_update_time,$u_notified,$u_attempts,
	      $u_reserved,$u_data);
	  ($service,$instance) = ("userspace",$user);
	}
	last if ($condition == $Status{"red"});  # doesn't get any worse
      }
    } else {  # just a normal group of users
      foreach $user (Cobalt::List::list_get_mem("", "", "$instance")) {
	my ($u_condition,$u_cond_time,$u_update_time,$u_notified,$u_attempts,
	    $u_reserved,$u_data) = split(/\|/o,$STATUS{"userspace|$user"});
	if ($u_condition > $condition) {
	  ($condition,$cond_time,$update_time,$notified,$attempts,
	   $reserved,$data) = 
	     ($u_condition,$u_cond_time,$u_update_time,$u_notified,$u_attempts,
	      $u_reserved,$u_data);
	  ($service,$instance) = ("userspace",$user);
	}
	last if ($condition == $Status{"red"});  # doesn't get any worse
      }
    }
  }
  elsif ($service eq "sitespace") {
    my $vsite;
    my $worst = 0;
    # Check each virtual site...and also the main site, of course.
    foreach $vsite (Cobalt::Vsite::vsite_list(),[1,1,$MainGroup,'']) {
      my $sitegroup = @$vsite[2];
      # first check the group itself
      my ($g_condition) = split(/\|/o,$STATUS{"groupspace|$sitegroup"});
      if ($g_condition > $worst) {
	($service,$instance) = ("groupspace",$sitegroup);
	$worst = $g_condition;
      }
      last if ($worst == $Status{"red"});  # doesn't get any worse

      # then check each user
      my $user;
      foreach $user (vsite_get_users($instance)) {
	my ($u_condition) = split(/\|/o,$STATUS{"userspace|$user"});
	if ($u_condition > $worst) {
	  ($service,$instance) = ("userspace",$user);
	  $worst = $u_condition;
	}
	last if ($worst == $Status{"red"});  # doesn't get any worse
      }
    }
    # need some defaults...
    $service ||= "groupspace"; 
    $instance ||= $MainGroup;
    ($condition,$cond_time,$update_time,$notified,$attempts,$reserved,$data) =
      split(/\|/o,$STATUS{"$service|$instance"});
  }
  elsif ($service eq "diskspace") {
    # Pick the worst of filesystem, groupspace, and userspace
    my ($groupcond,$usercond,$fscond);
    $fscond = (split(/\|/o,$STATUS{"filesystem|PRIME"}))[0];
    $groupcond = (split(/\|/o,$STATUS{"groupspace|PRIME"}))[0];
    $usercond = (split(/\|/o,$STATUS{"userspace|PRIME"}))[0];
    if (($fscond >= $usercond) && ($fscond >= $groupcond)) {
      $service = "filesystem";
    }
    elsif (($groupcond >= $usercond) && ($groupcond > $fscond)) {
      $service = "groupspace";
    }
    else {
      $service = "userspace";
    }
    
    # Okay, now that we've figured out what service is the worst,
    # figure out which instance is the worst so we can have an
    # intellegable message (not that it gets used, but it makes
    # me feel better)
    ($condition,$cond_time,$update_time,$notified,$attempts,$reserved,$data) =
      split(/\|/o,$STATUS{"$service|PRIME"});
    $instance = $data;  # PRIME stores the worst offender in its data field
    ($condition,$cond_time,$update_time,$notified,$attempts,$reserved,$data) =
      split(/\|/o,$STATUS{"$service|$instance"});
  }
  else {
    ($condition,$cond_time,$update_time,$notified,$attempts,$reserved,$data) =
      split(/\|/o,$STATUS{"$service|$instance"});
  }
  untie %STATUS if ($fit_to_be_untied);
  $condition ||= 0;

  # If we've just booted and haven't re-run monitor yet, we consider
  # ourselves to have no state information
  $condition = -1
    if ((stat($Issuefile))[10] > (stat($Status_db))[10]);

  my $message = Cobalt::Monitor::mon_build_alert($service,$instance,
						 $condition,$cond_time,
						 $update_time,$data);
  return ($condition,$message);
}

sub mon_get_state
# Get the state of a service
# Arguments: service or (service,instance)
#   Given just the service return the most severe instance within
#   the service (e.g., any red, then any yellow, then any green)
# Return value:  ($condition,$cond_time,$update_time,$notified,$attempts,
#                $reserved,$data)
# Example: @info = Cobalt::Monitor::mon_get_state("mem");
{
  my ($service,$instance) = @_;
  $instance ||= "PRIME";

  my ($condition,$cond_time,$update_time,$notified,$attempts,$reserved,$data);
  # The .state file is the textfile database which stores the state
  # for this service.  Check if we're using that instead of the Berkeley
  # DB file
  if (-e "$DB_dir_root/$service.state") {
    # Do it the text file way
    if (open(US, "$DB_dir_root/$service.state")) {
      "$instance|" =~ /^$instance\|/;  # pre-compile the pattern
      while (<US>) {
	next unless (//);

	# If we get here, we have a match
	chomp;
	($instance,$condition,$cond_time,$update_time,$notified,$attempts,$reserved,$data) = split(/\|/o);
	last;
      }
      close(US);
    }

    # If we've just booted and haven't re-run monitor yet, we consider
    # ourselves to have no state information
    $condition = -1
      if ((stat($Issuefile))[10] > (stat("$DB_dir_root/$service.state"))[10]);

  }
  else {
    # Do it with da database...
    my %STATUS;
    unless(tie(%STATUS, 'DB_File', $Status_db, O_RDONLY(),0644)) {
      return MSG_error("monitor_cant_read_status_db", $Status_db);
    }
    ($condition,$cond_time,$update_time,$notified,$attempts,$reserved,$data) =
      split(/\|/o,$STATUS{"$service|$instance"});
    untie %STATUS;

    # If we've just booted and haven't re-run monitor yet, we consider
    # ourselves to have no state information
    $condition = -1
      if ((stat($Issuefile))[10] > (stat($Status_db))[10]);
  }

  return -1 unless ($update_time);  # no state at all

  return ($condition,$cond_time,$update_time,$notified,$attempts,
	  $reserved,$data);
}

sub mon_list_instances
# Get the list of instances associated with a service
# Arguments: service name
# Return value: list of instances, sorted by severity
#    most severe will be at the top of the list
# Example: @disks = Cobalt::Monitor::mon_list_instances("filesystems");
{
  my ($service) = @_;

  my %STATUS;
  my $fit_to_be_untied = 0;
  my @instances;
  if (-e "$DB_dir_root/$service.state") {
    # Do it the text file way
    if (open(US, "$DB_dir_root/$service.state")) {
      while (<US>) {
	chomp;
	my @parts = split(/\|/o, $_, 2);
	$STATUS{"$service|$parts[0]"} = $parts[1];
	push(@instances,$parts[0]);
      }
      close(US);
    }
  }
  else {
    # Let's do it databasefully
    unless(tie(%STATUS, 'DB_File', $Status_db, O_RDONLY(),0644)) {
      return MSG_error("monitor_cant_read_status_db", $Status_db);
    }
    $fit_to_be_untied = 1;
    @instances = split(/\,/o,$STATUS{"$service|LIST"});
  }
  # sort the instances by condition
  @instances = sort {
    (split(/\|/o,$STATUS{"$service|$b"}))[0] <=>
      (split(/\|/o,$STATUS{"$service|$a"}))[0]
  } @instances;
  untie %STATUS if ($fit_to_be_untied);
  return @instances;
}

sub mon_set_state
# Set the state of a service
# There are two special instances: "PRIME" is the main instance for
#   this server.  "LIST" is the list of all instances for this server.
# Arguments: service, instance, condition, notified, attempts, reserved, data
#   An argument of "-" in any field indicates that the previous value
#   should be preserved.
#   If the service is userspace or groupspace and the instance begins
#   with *, then do the quota check on all the users/groups
# Return value: status string of the form MSG_error("3105")
{
  my ($service,$instance,$condition,$notified,$attempts,
      $reserved,$data) = @_;
  $instance ||= "PRIME";

  # Open and peek at our database
  my %STATUS;
  my %instances;  # all the instances there are
  if (($Features{'mon_textdb_only'}) || (-e "$DB_dir_root/$service.state")) {
    # Do it the text file way
    if (open(US, "$DB_dir_root/$service.state")) {
      while (<US>) {
	chomp;
	my @parts = split(/\|/o, $_, 2);
	$STATUS{"$service|$parts[0]"} = $parts[1];
	$instances{$parts[0]} = 1;
      }
      close(US);
    }
  }
  else {
    # Do it with da database...
    unless(tie(%STATUS, 'DB_File', $Status_db, O_RDWR()|O_CREAT(),0644)) {
      return MSG_error("monitor_cant_read_status_db", $Status_db);
    }
    %instances = map {$_ =~ s/^.+\|//o; $_ , 1} grep(/$service\|/,keys %STATUS);
  }
  delete $instances{'PRIME'}; # PRIME is special
  delete $instances{'LIST'};  # LIST is also special

  # Open the log file just once...
  open(LOG,">>$MonDBDir/$service.log") ||
    return MSG_error("monitor_cant_write_log", $MonDBDir);

  # Now we get into special cases...
  # We do the quota checks right here for increased performance.
  # This way we read the database exactly once and write the database
  # exactly once.
  # We also build our list of all the instances associated with this
  # service at this time, while we're here anyway.
  # For users and groups %instance is a hash of name => ID
  # However, sometimes we really do want to call this function for
  # just one user--e.g., to set the "notified" flag.  In that case,
  # don't do this.  How do we know?  Real usernames don't begin with
  # an asterisk (*).
  my @instances;  # the instances we're checking now
  if    (($service eq "userspace") && ($instance =~ /^\*/o)) {
    # Do the magic bulk user check
    setpwent();
    my ($name,$uid);
    while (($name,$uid) = (getpwent())[0,2]) {
      next if ($name eq $Adminuser);
      next if ($name =~ /^[A-Z]/o);  # capitalized names are special
      push (@instances,$name) unless ($System_users{$name});
      $instances{$name} = $uid;
    }
    endpwent();
  }
  elsif (($service eq "groupspace")  && ($instance =~ /^\*/o)) { 
    # Do the magic bulk group check
    setgrent();
    my ($name,$gid);
    while (($name,$gid) = (getgrent())[0,2]) {
      push (@instances,$name) unless ($System_groups{$name});
      $instances{$name} = $gid;
    }
    endgrent();
  }
  else { 
    # Normal update of a service's state information
    @instances = ($instance); 
  }

  # Make sure this instance is on the list of instances. 
  $instances{$instance} ||= 1;

  # We don't want to keep state between reboots; it looks weird
  # Find out how long we've been up (outside the loop)
  open(UPTIME,"/proc/uptime") || die "Cannot read /proc/uptime: $!\n";
  my $uptime = (split('\s',<UPTIME>))[0];
  close UPTIME;

  # Figure out my name for the logs
  my $svname = $Servers{$service};
  $svname ||= $service;

  my ($update_time,$cond_time,$inst);
 INST:
  foreach $inst (@instances) {

    if (($service =~ /^(user|group)space$/o) && ($instance =~ /^\*/o)) {
      # magic bulk quota check; here we do the work 
      # of retrieving the values for this actual user.
      # code copied from mon_quotawatch
      my ($qtype);
      ($service =~ /^group/o) ? ($qtype = $GrpQuota) : ($qtype = $UsrQuota);
      my ($used,$soft) = Cobalt::Quota::repquota($inst,$qtype,
				  $instances{$inst},$Htusers_dir);
      my ($severity,$state);
      next INST if (!$soft);
      
      my $share = int(($used / $soft) * 1000) / 10;
      if    ($share >= $QuotRed) { $state = "red";    }
      elsif ($share >= $QuotYel) { $state = "yellow"; }
      else                       { $state = "green";  }
      
      $condition = $Status{$state};
      $notified = "-";
      $attempts = "-";
      $reserved = "-";
      $data = "$share,$used,$soft";
    }

    # o_ for old, p_ for prime
    my ($o_condition,$o_cond_time,$o_update_time,$o_notified,$o_attempts,
	$o_reserved,$o_data) = split(/\|/o,$STATUS{"$service|$inst"});
    $update_time = $^T;

    # - means the prevous value should be preserved
    if ($notified eq "-")  { $notified = $o_notified; }
    if ($attempts eq "-")  { $attempts = $o_attempts; }
    if ($reserved eq "-")  { $reserved = $o_reserved; }
    if ($condition eq "-") { $condition = $o_condition; }
    if ($data eq "-")      { $data = $o_data; }

    if ($condition == $o_condition) {     # condition is the same
      if ($o_cond_time > $^T) {
	# The clock has been set back, or time is moving backwards again
	$cond_time = $^T;
      } elsif (($^T - $uptime) < $o_cond_time) {
	# we've been up all this time
        if (-e "$Subsys_dirs{$service}") {
          my $subsys_time = (stat("$Subsys_dirs{$service}"))[10];
          if ($subsys_time > $o_cond_time) {
            # The subsys file is newer, the service must have been
            # stopped & restarted or something
            $cond_time = $subsys_time;
          }
          else {
            # No change to the subsys file
            $cond_time = $o_cond_time;
          }
        } else {
          # No subsys file (may not be appropriate for this service)
          $cond_time = $o_cond_time;
        }
      } else {
	# we've rebooted
	$cond_time = $^T;
      }
    }
    else {                                # condition has changed
      $cond_time = $^T;
    }
    
    # if all is well, reset the auto-repair counters
    # Note that these counters mostly apply to the PRIME instance; 
    # the others don't, in general, use them because regardless of
    # how many things are wrong with the CPU, we only want to be
    # told once, "The CPU is overloaded".  The exception is for
    # quota problems--each user and group going over quota gets
    # an individual notification.
    if ($condition <= $Status{"green"}) {
      $notified = 0;
      $attempts = 0;
    }

    # sanitation defaults
    $condition ||= 0;
    $attempts ||= 0;
    $notified ||= 0;
    $reserved ||= 0;
    $data ||= 0;

    $STATUS{"$service|$inst"} = join('|',($condition,$cond_time,
					      $update_time,$notified,
					      $attempts,$reserved,$data));

    if ($#instances < 50) { 
      # This is a luxury we can only afford if we aren't in a huge loop
      my $message = Cobalt::Monitor::mon_build_alert($service,$inst,$condition,
				 $cond_time,$update_time,$data);
      print LOG "<FONT COLOR=\"$Status{$condition}\">$svname, ",
      Cobalt::Language::msg_localtime($update_time), 
      ": $message</FONT><!-- $inst ($data)-->\n";
    }
  }
  if ($#instances >= 50) { 
    # we need to record *something* in the log...
    print LOG "<FONT COLOR=\"$Status{$condition}\">$svname, ",
    Cobalt::Language::msg_localtime($update_time), ": examination complete at ",
    scalar(localtime(time())), "</FONT>\n";
  }
  close LOG;

  # Set the PRIME instance to the worst offender, and clean out
  # stale entries in the database
  # Could possibly get a slight performance boost by doing this within the
  # foreach loop, but then we wouldn't clean out old entries...and this
  # still only cycles us through the user list twice total
  my $instp;
  my ($p_condition,$p_cond_time,$p_update_time,$p_notified,
      $p_attempts,$p_reserved,$p_data) = 
	split(/\|/o,$STATUS{"$service|PRIME"});
  my ($w_condition,$w_cond_time) = (0,$^T);
  foreach $instp (keys %instances) {
    if(!$STATUS{"$service|$instp"}) {
      delete $instances{$instp};          # Take it off the list
      delete $STATUS{"$service|$instp"};  # Clear the entry
    }
    my ($l_condition,$l_cond_time,$l_update_time,$l_notified,$l_attempts,
	$l_reserved,$l_data) = split(/\|/o,$STATUS{"$service|$instp"});
    next unless ($l_condition);  # no state yet
    # If this hasn't been checked in over an hour, then it is probably
    # not a valid service anymore (or it's just stale information)
    if (($^T - $l_update_time) > 4000) {
      delete $instances{$instp};          # Take it off the list
      delete $STATUS{"$service|$instp"};  # Clear the entry
    }
    # otherwise, it's a candidate for Worst Offender
    elsif ($w_condition < $l_condition) {
      $w_condition = $l_condition;
      $w_cond_time = $l_cond_time;
      $p_data = $instp; # PRIME stores the instance of the worst in $data
    }
    last if ($w_condition == $Status{"red"});  # doesn't get any worse
  }

  # if all is well, reset the auto-repair counters
  # (see also above comments)
  if ($w_condition <= $Status{"green"}) {
    $p_attempts = 0;
    $p_notified = 0;
  }

  # Store the primary information
  $STATUS{"$service|PRIME"} = join('|',($w_condition,$w_cond_time,
					$^T,$p_notified,$p_attempts,
					$p_reserved,$p_data))
				   unless ($instance eq "PRIME");

  # Now, we also store our list of instances
  unless (($service eq "userspace") || ($service eq "groupspace")) {
    $STATUS{"$service|LIST"} = join(',',sort keys(%instances));
  } elsif ($instance =~ /^\*/o) {  # only on the bulk check
    # for users and groups, %instances includes *all* users and groups;
    # @instances is just the ones we check quotas for.
    # %instances contains the rest so we can cache the information
    $STATUS{"$service|LIST"} = join(',',sort @instances);
  }

  # Save this information in the database
  if (($Features{'mon_textdb_only'}) || (-e "$DB_dir_root/$service.state")) {
    sleep 10 if (-e "$DB_dir_root/$service.state.lock");
    if (-e "$DB_dir_root/$service.state.lock") {
      return MSG_error("monitor_cant_lock_statefile", $DB_dir_root);
    }
    open(STATE,">$DB_dir_root/$service.state.lock") ||
      return MSG_error("monitor_cant_write_statefile", $DB_dir_root);
    my $key;
    foreach $key (keys %STATUS) {
      $key =~ s/^[^\|]+\|//o;  # strip off the "$service|" part of the key
      print STATE join('|',$key, $STATUS{"$service|$key"}, "\n");
    }
    close STATE;
    unless (rename("$DB_dir_root/$service.state.lock",
		   "$DB_dir_root/$service.state")) {
      unlink("$DB_dir_root/$service.state.lock");
      return MSG_error("monitor_cant_save_statefile", $DB_dir_root);
    }
  }
  else {
    untie %STATUS;
  }
  return MSG_ok("monitor_ok_set_state", $service);
}


#######################################################################
#
# Monitoring/examination functions

sub mon_areastat
# Get statistics for how much disk space is being used under a
#   given directory (like du -s); gives size in bytes
# Arguments: the name of the top-level directory
# Return value: none
# Side effects: adds entries to the database
{
  my ($dir) = @_;
  my $size =  (Cobalt::Monitor::du_recursively($dir) + (stat($dir))[7]);
  Cobalt::Monitor::mon_set_state("dirspace",$dir,$Status{"green"},"-","-","-",$size);
}

sub du_recursively
# Recursively gets the usage in a directory tree
# Arguments: the directory to start at
# Return value: the usage below the given directory
{
    my ($top) = @_;
    my (@files,$filename,$file,$usage);
    $usage = 0;

    if ($top =~ /^([^;\|]+)$/o)  {$top = $1; }  # untaint ourselves

    print STDERR "Reading usage in $top\n" if ($Debug);
    if ((-d $top) && (! -l $top))
    {
	# It's a directory, recurse
	opendir(DIR,"$top") 
	    || return MSG_error("monitor_cant_read_dir", $top);
	@files = grep(!/^\.\.?$/o,readdir(DIR));
	closedir DIR;
	if (@files)
	{
	    foreach $filename (@files)
	    {
		my $file = "$top/$filename";
		print STDERR "Looking at $top/$filename ($file\n" if ($Debug);
		if ((-d $file) && (! -l $file))
		{
		    $usage += (stat(_))[7];
		    $usage += Cobalt::Monitor::du_recursively($file);
		}
		else
		{
		    $usage += (stat(_))[7];
		}
	    }
	}
    }
    else
    {
	# it's just a file...
	$usage += (stat(_))[7];
    }
    print STDERR "Usage for $top is $usage\n" if ($Debug);
    return $usage;
}


sub mon_diskwatch
# File systems: check if a specific disk is full
# Arguments: a filesystem to look at, 
#   yellow threshold (%), and red threshold (%)
# Return value: a description of any problems found, as per the tests given
# Side effects: adds entries to the database
{
    my ($filesys,$yelp,$redp) = @_;
    my ($cap,$header,$freespace);
    
    # Why the -e?  Because I'm paranoid.
    unless ((-e $filesys) && (open (DF, "$Df_cmd $filesys|")))
    {
	print STDERR "Cannot run $Df_cmd $filesys: $!\n";
	return -1;
    }

    $header = <DF>; # toss the header line

    while (<DF>)
    {
        ($cap,$freespace) = (split(/\s+/o))[4,3];
	chop $cap; # remove the % sign
    }
    close DF;
    if ($cap >= $redp)
    {
      Cobalt::Monitor::mon_set_state("filesystem",$filesys,$Status{"red"},
		    "-","-","-","$cap,$freespace");
    }
    elsif ($cap >= $yelp)
    {
      Cobalt::Monitor::mon_set_state("filesystem",$filesys,$Status{"yellow"},
		    "-","-","-","$cap,$freespace");
    }
    else
    {
      Cobalt::Monitor::mon_set_state("filesystem",$filesys,$Status{"green"},
		    "-","-","-","$cap,$freespace");
    }
}

sub mon_quotawatch
# monitor the quota for a user or group (or site)
# Arguments: name, type, yellow threshold (%), red threshold (%)
#   type = $UsrQuota or $GrpQuota
# Side effects: updates the database
# Return value: none
{
    my ($name,$type,$yelp,$redp) = @_;

    if ($name eq "*USERS") {
      Cobalt::Monitor::mon_set_state("userspace","*USERS");
      return;
    }
    elsif ($name eq "*GROUPS") {
      Cobalt::Monitor::mon_set_state("groupspace","*GROUPS");
      return;
    }

    # The rest of this code never gets used...we moved all
    # the quota checking logic directly into Cobalt::Monitor::mon_set_state()
    my ($used,$soft) = Cobalt::Quota::repquota($name,$type);
    my ($t,$severity,$state);
    return unless ($soft); # No quota.  No point.
    ($type) ? ($t = "group") : ($t = "user");

    my $share = int(($used / $soft) * 1000) / 10;
    if ($share >= $redp)    { $state = "red";    }
    elsif ($share >= $yelp) { $state = "yellow"; }
    else                    { $state = "green";  }

    Cobalt::Monitor::mon_set_state("${t}space",$name,$Status{$state},
		  "-","-","-","$share,$used,$soft");
    return($Status{$state},$used);

}

sub mon_loadwatch
# Log when the CPU load is too great
# Arguments: yellow threshold, red threshold
# Side effects: updates the database
# Return value: none
{
    my ($yelt,$redt) = @_;
    my ($five,$fifteen);
    open(LOAD,"/proc/loadavg") ||
	return MSG_error("monitor_cant_read_load", $!);
    my $line = <LOAD>;
    $line =~ /^(\d+\.\d+)\s*(\d+\.\d+)\s*(\d+\.\d+)/o;
    close LOAD;
    #$five = $2;
    $fifteen = $3;

    if ($fifteen >= $redt)
    {
      Cobalt::Monitor::mon_set_state("cpu","load",$Status{"red"},"-","-","-","$fifteen");
    }
    elsif ($fifteen >= $yelt)
    {
      Cobalt::Monitor::mon_set_state("cpu","load",$Status{"yellow"},"-","-","-","$fifteen");
    }
    else
    {
      Cobalt::Monitor::mon_set_state("cpu","load",$Status{"green"},"-","-","-","$fifteen");
    }
}

sub mon_collwatch
# Keep an eye on the network interface and report excessive collisions
# Monitors the number of collisions since system boot, and the number
#   over just the last 15 minutes
# Arguments: yellow threshold (%), red threshold (%), optional network
#   interface (defaults to primary)
# Note: this function has little or no value.  It should never be used.
{
    my ($yelp,$redp,$netif) = @_;
    my ($tx,$colls);
    $netif ||= $Netif;
    # there seems to be some inconsistency about the format of this
    # file.  Let's hope that it's okay.
    open(NS,"/proc/net/dev") || die "Cannot read /proc/net/dev: $!\n";
    while (<NS>)
    {
        if (/^\s*$netif:\s*([\d\s]+)/)
        {
            ($tx,$colls) = (split(/\s+/o,$1))[5,9];
	    last;
        }
    }
    close NS;

    # First do the current collision rate (since the last monitor run)
    my ($condition,$cond_time,$update_time,$notified,$attempts,$reserved,$data) = 
      Cobalt::Monitor::mon_get_state("net","col15");
    my ($prevtx,$prevcols) = split(/,/o,$data);

    # If we rebooted within the last 15 minutes, our state will
    # be from the previous boot.  Ignore it and reset our counter.
    open(UPTIME,"/proc/uptime") || die "Cannot read /proc/uptime: $!\n";
    my $uptime = (split('\s',<UPTIME>))[0];
    close UPTIME;
    if (($tx < $prevtx) || ($uptime < 1000))
    {
	$prevtx = 0;
	$prevcols = 0;
    }

    my ($colpcent,$col15);  # overall, past 15 mins

    # What if there's been no activity at all?
    if (($tx - $prevtx) == 0)
    {
	# No activity == no collisions
	$col15 = 0;
	$colpcent = 0;
    }
    else
    {
	# There's been activity
	$col15 = int((($colls - $prevcols) / ($tx - $prevtx)) * 100);
	$colpcent = int(($colls / $tx) * 100);
    }

    # record the past-15-minutes state
    if (($tx - $prevtx) < $MinTx)
    {
      Cobalt::Monitor::mon_set_state("net","col15",0,"-","-","-","$tx,$colls,$col15");
    }
    elsif ($col15 >= $redp)
    {
      Cobalt::Monitor::mon_set_state("net","col15",$Status{"red"},"-","-","-",
		    "$tx,$colls,$col15");
    }
    elsif ($col15 >= $yelp)
    {
      Cobalt::Monitor::mon_set_state("net","col15",$Status{"yellow"},"-","-","-",
		    "$tx,$colls,$col15");
    }
    else
    {
      Cobalt::Monitor::mon_set_state("net","col15",$Status{"green"},"-","-","-",
		    "$tx,$colls,$col15");
    }

    # record the overall-since-boot state
    if ($tx < $MinTx)
    {
      Cobalt::Monitor::mon_set_state("net","collisions",0,"-","-","-",$colpcent);
    }
    elsif ($colpcent >= $redp)
    {
      Cobalt::Monitor::mon_set_state("net","collisions",$Status{"red"},"-","-","-",$colpcent);
    }
    elsif ($colpcent >= $yelp)
    {
      Cobalt::Monitor::mon_set_state("net","collisions",$Status{"yellow"},"-","-","-",$colpcent);
    }
    else
    {
      Cobalt::Monitor::mon_set_state("net","collisions",$Status{"green"},"-","-","-",$colpcent);
    }
}

sub mon_memwatch
# Keep an eye on the memory state and report excessive swapping
# Arguments: yellow threshold, red threshold
#   thresholds are in percent of virtual memory used
# Side effects: updates the database
# Return value: none
{
    my ($yelp,$redp) = @_;
    my ($ptot,$pfree,$stot,$sfree,$vtot,$vused,$usepcent);
    # there seems to be some inconsistency about the format of this
    # file.  Let's hope that it's okay.
    open(MEM,"/proc/meminfo") || die "Cannot read /proc/meminfo: $!\n";
    while (<MEM>)
    {
	if (/^MemTotal:\s+(\d+)/o)
	{
	    $ptot = $1;
	}
	elsif (/^MemFree:\s+(\d+)/o)
	{
	    $pfree = $1;
	}
	elsif (/^SwapTotal:\s+(\d+)/o)
	{
	    $stot = $1;
	}
	elsif (/^SwapFree:\s+(\d+)/o)
	{
	    $sfree = $1;
	}
	elsif (/^Buffers:\s+(\d+)/o)
	{
	  # Linux uses whatever excess physical memory isn't actually
	  # in use for buffer space.  For our purposes, this shouldn't
	  # be counted as memory actually in use, so we count it as
	  # free memory.
	  $pfree += $1;
	}
    }
    close MEM;
    $vtot = $ptot + $stot;
    $vused = $vtot - ($pfree + $sfree);
    $usepcent = int(($vused / $vtot) * 1000) / 10;

    if ($usepcent >= $redp)
    {
      Cobalt::Monitor::mon_set_state("mem","vmem",$Status{"red"},"-","-","-",$usepcent);
    }
    elsif ($usepcent >= $yelp)
    {
      Cobalt::Monitor::mon_set_state("mem","vmem",$Status{"yellow"},"-","-","-",$usepcent);
    }
    else
    {
      Cobalt::Monitor::mon_set_state("mem","vmem",$Status{"green"},"-","-","-",$usepcent);
    }
}

sub mon_swapwatch
# Keep an eye on the memory state and report excessive swapping
# Arguments: yellow threshold, red threshold
#   thresholds are in pages per second (in or out)
# Side effects: updates the database
# Return value: none
{
    my ($yelp,$redp) = @_;
    my ($swapped,$swapness);
    open(MEM,"/proc/stat") || die "Cannot read /proc/stat: $!\n";
    while (<MEM>)
    {
	if (/^swap\s+(\d+)\s+(\d+)/o)
	{
	    $swapped = $1 + $2;
	}
    }
    close MEM;
    
    my ($condition,$cond_time,$update_time,$notified,$attempts,
	$reserved,$data) = Cobalt::Monitor::mon_get_state("mem","swapping");
    my ($hadswapped,$swaps_per_second) = split(/,/o,$data);

    # If we rebooted within the last 15 minutes, our state will
    # be from the previous boot.  Ignore it and reset our counter.
    open(UPTIME,"/proc/uptime") || die "Cannot read /proc/uptime: $!\n";
    my $uptime = (split('\s',<UPTIME>))[0];
    close UPTIME;
    $hadswapped = 0 if ($uptime < 1200); # 15 min + margin

    unless ($update_time && $hadswapped)
    {
	# no previous state
	Cobalt::Monitor::mon_set_state("mem","swapping",0,"-","-","-",$swapped);
	return;
    }

    my $elapsed = time() - $update_time;  $elapsed ||= 1;
    my $elapsmin = int($elapsed / 60);
    $swapness = ($swapped - $hadswapped) / $elapsed;  # swaps per second
    $swapness = int($swapness * 100) / 100;   # get rid of excessive digits
    $swapness = 0 if ($swapness <= 0);

    if ($swapness >= $redp)
    {
      Cobalt::Monitor::mon_set_state("mem","swapping",$Status{"red"},"-","-","-",
		    "$swapped,$swapness");
    }
    elsif ($swapness >= $yelp)
    {
      Cobalt::Monitor::mon_set_state("mem","swapping",$Status{"yellow"},"-","-","-",
		    "$swapped,$swapness");
    }
    else
    {
      Cobalt::Monitor::mon_set_state("mem","swapping",$Status{"green"},"-","-","-",
		    "$swapped,$swapness");
    }
}

sub mon_procwatch
# Check to see if a process is running
# Arguments: process name (the command) and the name of the server
#   (for use in the database)
# Side effects: updates the database
# Return value: none
{
    my ($process,$service) = @_;
    # Use the cached process list instead of running ps over and over
    if ((-e "$ProcCache") && ((time() - (stat(_))[10]) < 300))
    {
	unless (open (PS,"$ProcCache"))
	{
	    return MSG_error("monitor_cant_read_pslist", $!);
	}
    }
    else
    {
	unless (open (PS,"$Ps_cmd axw|"))
	{
	    print STDERR "Cannot run $Ps_cmd: $!\n" if ($Debug);
	    return MSG_error("monitor_cant_run_ps", $Ps_cmd);
	}	
	<PS>;  # chop the top line (which is just header)
    }

    "$process" =~ /$process/;  # minor speed increase by pre-compiling
    while (<PS>)
    {
	next unless (//);
	chomp;
	/^\s*\d+\s+\S+\s+\S+\s+\S+\s+(.+)/o;
	my ($cmd) = ($1);
	next unless ($cmd);
	if ($cmd =~ /$process/)
	{
	    close PS;
	    print STDERR "Found $process in $cmd\n" if ($Debug);
	    Cobalt::Monitor::mon_set_state($service,$process,$Status{"green"},"-","-","-","-");
	    return;
	}
    }
    close PS;

    # got down here without seeing it...
    Cobalt::Monitor::mon_set_state($service,$process,$Status{"red"},"-","-","-","-");
}

sub mon_spoolwatch
# monitor the size of a user's mail spool
# Arguments: name
# Side effects: updates the database
# Return value: none
{
    my ($name) = @_;
    # Get the size of the user's mail spool, in kilobytes
    my ($size) = int((stat("$Mailspool/$name"))[7] / 1024);
    if ($size >= $MailSpoolMax)
    {
      Cobalt::Monitor::mon_set_state("mailspace",$name,$Status{"red"},"-","-","-",$size);
    }
    else
    {
      Cobalt::Monitor::mon_set_state("mailspace",$name,$Status{"green"},"-","-","-",$size);
    }
}


#############################################################################
#
# Connect to various TCP ports to verify that network services are running
#

sub mon_portwatch
# Try to connect to a network port on a machine and verify that it
#            responds properly.
# Arguments:
#            the name of the service (text description)
#            the host (name or IP address) on which the service is running
#            the port (name or number) on which the service is running
#            the instance of the service (defaults to service name if null)
# Return value: none of note
# Side effects: updates the database to record the last time the state
#            of the service changed
{
    my ($service,$remote,$port,$inst) = @_;
    $inst ||= $service;  # default value
    my ($proto,$iaddr,$paddr);

    # Set our variables for the connection
    $proto = getprotobyname('tcp');
    if ($port =~ /\D/o) { $port = getservbyname($port, 'tcp') }
    unless ($port)
    {
	return MSG_error("monitor_invalid_port");
    }
    unless ($iaddr = inet_aton($remote))
    {
	return MSG_error("monitor_cant_unknown_host", $remote);
    }
    $paddr = sockaddr_in($port, $iaddr);

    # Make sure we don't hang forever...
    alarm 10;
    $SIG{'ALRM'} = sub { 
      print LOG "Failed to send to $service $inst: $!\n";
      Cobalt::Monitor::mon_set_state($service,$inst,$Status{"red"},"-","-","-","-");
      close LOG;
      return;
    };

    # Try to bind to a socket and connect to the service
    unless (socket(S, PF_INET(), SOCK_STREAM(), $proto))
    {
      print STDERR "Can't create socket: $!\n" if ($Debug);
      Cobalt::Monitor::mon_set_state($service,$inst,$Status{"red"},"-","-","-","-");
      return 0;
    }
    unless (connect(S, $paddr))
    {
      print STDERR "Can't connect to $port on $remote: $!\n" if ($Debug);
      Cobalt::Monitor::mon_set_state($service,$inst,$Status{"red"},"-","-","-","-");
      return 0;
    } elsif ($service eq 'adm') # Fix complaining Apache 
    {
      print S "GET /.cobalt/sysManage/index.html HTTP/1.0\n\n";
    }

    # test beyond the fact that something answered
    if ($Features{'mon_port_testing'})
    {
      my $bad_response = 0;
      open(LOG,">>$Sauce_log") || open(LOG,">>$Console");
      $| = 1; select(S); $| = 1; select(STDOUT);
      if ($service eq "mail") {
	my $line = <S> || "";
	unless ($line) {
	  $bad_response = 1;
	  print LOG scalar(localtime($^T)), " Null response from $inst";
	  goto BAD;
	}
	if ($inst eq "pop3") {
	  print S "quit\n" || print LOG "Can't send quit to $inst server: $!\n";
	  $line = <S> || "";
	  unless ($line =~ /^\+OK/o) {
	    $bad_response = 1;
	    print LOG scalar(localtime($^T)), " Bad response from $inst: $line";
	    goto BAD;
	  }
	}
	elsif ($inst eq "imap") {
	  unless ($line =~ /OK/o) { 	  # check the banner
	    $bad_response = 1;
	    print LOG scalar(localtime($^T)), " Bad response from $inst: $line";
	    goto BAD;
	  }
	  print S "cobalt LOGIN Active_Monitor\n";
	  $line = <S> || "";
	  unless ($line =~ /^cobalt/o) {  # now check the response
	    $bad_response = 1;
	    print LOG scalar(localtime($^T)), " Bad response from $inst: $line";
	    goto BAD;
	  }
	  print S "cobalt LOGOUT\n";
	}
	elsif ($inst eq "smtp") {
	  print S "ETRN\n";
	  $line = <S> || "";
	  unless ($line =~ /^\d/o) {  # check the response
	    $bad_response = 1;
	    print LOG scalar(localtime($^T)), " Bad response from $inst: $line";
	    goto BAD;
	  }
	  print S "QUIT\n";
	}
      }
      elsif ($service eq "ftp") {
	my $line = <S> || "";
	unless ($line) {
	  $bad_response = 1;
	  print LOG scalar(localtime($^T)), " Null response from $service";
	  goto BAD;
	}
	print S "user active_monitor\n";
	$line = <S> || "";
	unless ($line =~ /^\d/o) {
	  $bad_response = 1;
	  print LOG scalar(localtime($^T)), " Bad response from $service: $line";
	  goto BAD;
	}
	print S "quit\n" || print LOG "Can't send quit to $inst server: $!\n";
      }
      elsif ($service eq "telnet") {
	# Leaves messy log entries...how do we cleanly send a close?
	print S "quit\r\n";
      }

      # abort the communication and jump here when things break
BAD:
      close S;
      close LOG;
      alarm 0;
      $SIG{'ALRM'} = 'DEFAULT';

      if ($bad_response) {
	Cobalt::Monitor::mon_set_state($service,$inst,$Status{"red"},"-","-","-","-");
	return;
      }
    }

    # Make sure the alarm is off
    alarm 0;
    $SIG{'ALRM'} = 'DEFAULT';
    Cobalt::Monitor::mon_set_state($service,$inst,$Status{"green"},"-","-","-","-");
    return;
}

sub mon_nfstest
# Check to see if the server is talking NFS
# Arguments: none
# return value: none
# Side effects: updates the database
{
    my $try =`$Showmount -e 2>&1`;
    unless ($?)
    {
      Cobalt::Monitor::mon_set_state("nfs","showmount",$Status{"green"},"-","-","-","-");
    }
    else
    {
      Cobalt::Monitor::mon_set_state("nfs","showmount",$Status{"red"},"-","-","-","-");
    }
}

sub mon_atalktest
# Check to see if the server is talking appletalk
# Arguments: none
# return value: none
# Side effects: updates the database
{
    my ($host) = Cobalt::Network::net_get_hostname();
    $host =~ s/\..+$//o; # basename w/o domain
    my ($afp,$neta,$ws);
    open(NL,"$Nbplkup|") ||
	return MSG_error("monitor_cant_run_nbplkup", $Nbplkup);
    while (<NL>)
    {
	if (/$host:AFPServer/)
	{
	    $afp = 1;
	}
	elsif (/$host:netatalk/)
	{
	    $neta = 1;
	}
	elsif (/$host:Workstation/)
	{
	    $ws = 1;
	}
    }
    close NL;
    unless ($afp && $neta && $ws)
    {
      Cobalt::Monitor::mon_set_state("atalk","nbplkup",$Status{"red"},"-","-","-","-");
    }
    else
    {
      Cobalt::Monitor::mon_set_state("atalk","nbplkup",$Status{"green"},"-","-","-","-");
    }
}

sub mon_nslookuptest
# Check to see if the server is talking appletalk
# Arguments: none
# return value: none
# Side effects: updates the database
{
    my ($host) = Cobalt::Network::net_get_hostname();

    open(NSL,"$Nslookup $host 2>&1 |") ||
	return MSG_error("monitor_cant_run_nslookup", $Nslookup);
    while (<NSL>)
    {
	if (/Name:\s+$host/)
	{
	  #Cobalt::Monitor::mon_set_state("dns","nslookup",$Status{"green"},"-","-","-","-");
          mon_set_state("mail","nslookup",$Status{"green"},"-","-","-","-");

	}
	elsif (/^\*\*\*/o)
	{
	  #Cobalt::Monitor::mon_set_state("dns","nslookup",$Status{"red"},"-","-","-","-");
+         mon_set_state("mail","nslookup",$Status{"red"},"-","-","-","-");
	}
    }
    close NSL;
}

sub mon_cronwatch
# Check to see that cron is updated its logfile.
# Cron tends to die easily if the clock gets messed with.
# Arguments: none
# Side effects: checks the timestamp on /var/log/cron against the current time
#   updates the monitor database
{
  if ((stat($Cron_log))[9] < (time() - 360)) {
    Cobalt::Monitor::mon_set_state("crond","log",$Status{"red"},"-","-","-","-");
  } else {
    Cobalt::Monitor::mon_set_state("crond","log",$Status{"green"},"-","-","-","-");
  }
}

sub mon_build_alert
# Build the message
# Arguments: ($server,$instance,$severity,$cond_time,$update_time,$data)
#   $server:     the name of the server (or service); e.g., "mem",
#                "ftp", or "smb".
#   $instance:   a particular subsystem or instance of the server;
#                for example, the "userspace" server has an instance for
#                each user; the "mail" server has instances for pop,
#                imap, and smtp; etc.
#   $severity:   A number indicating the state of this instance 
#                1 = green, 2 = yellow, 3 = red
#   $cond_time:  The time at which this severity was entered
#   $update_time: The time when this particular report was generated
#   $data:       Additional information, defined per service and instance
# Return value: a nicely stated message string
{
    my ($server,$instance,$severity,$cond_time,$update_time,$data) = @_;
    my $return;
    my @data = split(/,/o,$data);
    $instance = lc($instance);  # lowercase everything for consistency

    $severity = -1 unless (defined($severity) && $severity);

    my $time = Cobalt::Language::msg_localtime($cond_time);
    my $servername = Locale::gettext::dgettext("special-sauce", "servers_$server");

    if ($server eq "cpu") { # CPU load
      if ($severity == $Status{"green"}) {
	$return = Locale::gettext::dgettext("special-sauce", "mon_state_cpu_green");
	$return = eval "\"$return\"";
	return $return;
      } elsif ($severity <= 0) {
	$return = Locale::gettext::dgettext("special-sauce", "mon_state_noneyet");
	$return = eval "\"$return\"";
	return $return;
      } elsif ($severity == $Status{"yellow"}) { 
	$return = Locale::gettext::dgettext("special-sauce", "mon_state_cpu_yellow");
	$return = eval "\"$return\"";
	return $return;
      } else {
	$return = Locale::gettext::dgettext("special-sauce", "mon_state_cpu_red");
	$return = eval "\"$return\"";
	return $return;
      }
    }
    elsif ($server eq "filesystem") {
      if ($severity <= 0)
      {
	  return Locale::gettext::dgettext("special-sauce", "mon_state_noneyet");
      }
      # $data[0] is the percent of capacity used
      $instance =~ s/\//slash/o;
      my $diskname = Locale::gettext::dgettext("special-sauce", "disks_$instance");
      $return = Locale::gettext::dgettext("special-sauce", "mon_state_disk");
      $return = eval "\"$return\"";
      return $return;
    }
    elsif (($server eq "userspace")  ||
	   ($server eq "groupspace") ||
	   ($server eq "mailspace")) {
      $return = Locale::gettext::dgettext("special-sauce", "mon_state_$server");
      $return = eval "\"$return\"";
      return $return;      
    }
    elsif ($server eq "mail") {
      $servername = $instance;
      if ($severity == $Status{"green"}) {
	$return = Locale::gettext::dgettext("special-sauce", "mon_state_green");
	$return = eval "\"$return\"";
	return $return;
      } elsif ($severity <= 0) {
	$return = Locale::gettext::dgettext("special-sauce", "mon_state_noinfo");
	$return = eval "\"$return\"";
	return $return;
      } else {
	$return = Locale::gettext::dgettext("special-sauce", "mon_state_red");
	$return = eval "\"$return\"";
	return $return;
      }
    }
    elsif ($server eq "mem") {
        if ($severity <= 0)
	{
	    return Locale::gettext::dgettext("special-sauce", "mon_state_noneyet");
	}
	# how much swapping is going on?
	if    ($instance eq "swapping")  
	{
	  if ($severity == $Status{"green"}) {
	    $return = Locale::gettext::dgettext("special-sauce", "mon_state_mem_swaps_green");
	    $return = eval "\"$return\"";
	    return $return;
	  } elsif ($severity <= 0) {
	    $return = Locale::gettext::dgettext("special-sauce", "mon_state_mem_noinfo");
	    $return = eval "\"$return\"";
	    return $return;
	  } elsif ($severity == $Status{"yellow"}) { 
	    $return = Locale::gettext::dgettext("special-sauce", "mon_state_mem_swaps_yellow");
	    $return = eval "\"$return\"";
	    return $return;
	  } else {
	    $return = Locale::gettext::dgettext("special-sauce", "mon_state_mem_swaps_red");
	    $return = eval "\"$return\"";
	    return $return;
	  }
	}
	elsif ($instance eq "vmem")  # virtual memory use
	{
	  # data[0] is the percentage of virtual memory used
	  $return = Locale::gettext::dgettext("special-sauce", "mon_state_mem_vmem");
	  $return = eval "\"$return\"";
	  return $return;
	}
    }
    elsif ($server eq "net") {
      # The network collision rate is a special case.  Also, it is broken,
      # has always been broken, will always be broken, and should be
      # removed from the product.  Who listens to me, though?
      my ($severity) = Cobalt::Monitor::mon_get_state($server,$instance);

	if ($severity < 0)
	{
	  return Locale::gettext::dgettext("special-sauce", "mon_state_noneyet");
	}

	if    ($instance eq "collisions") { # collisions since boot
	  # data is the collision rate since boot
	  if ($severity == 0) {
	    return Locale::gettext::dgettext("special-sauce", "mon_state_net_idle");
	  }
	  else {
	    # how long have we been up?
	    open(UPTIME,"/proc/uptime") 
	      || die "Cannot read /proc/uptime: $!\n";
	    my $uptime = (split('\s',<UPTIME>))[0];  
	    close UPTIME;
	    
	    my $since = $uptime - ($^T - $update_time);
	    my $days = int($since / 86400);
	    if ($days) {
	      ($since -= ($days * 86400));
	      $days = "$days days ";
	    }
	    else {
	      $days = "";
	    }
	    my $hours = int($since / 3600);
	    if ($hours) {
	      ($since -= ($hours * 3600));
	      $hours = "$hours hours ";
	    }
	    else {
	      $hours = "";
	    }
	    my $mins = int($since / 60);
	    ($mins) ? ($mins = "$mins minutes") : ($mins = "");
	    
	    if ($severity == $Status{"green"}) {
	      $return = Locale::gettext::dgettext("special-sauce", "mon_state_net15_green");
	      $return = eval "\"$return\"";
	      return $return;
	    } elsif ($severity == $Status{"yellow"}) {
	      $return = Locale::gettext::dgettext("special-sauce", "mon_state_net15_yellow");
	      $return = eval "\"$return\"";
	      return $return;
	    } else {
	      $return = Locale::gettext::dgettext("special-sauce", "mon_state_net15_red");
	      $return = eval "\"$return\"";
	      return $return;
	    }
	  }
	}
    }
    elsif ($instance eq "col15") { # collisions over past 15 minutes
      # The collision rate over the past 15 minutes is stored
      # in $data3
      if (! $severity){
	return Locale::gettext::dgettext("special-sauce", "mon_state_net15_idle");
      }
      elsif ($severity == $Status{"green"}) {
	$return = Locale::gettext::dgettext("special-sauce", "mon_state_net15_green");
	$return = eval "\"$return\"";
	return $return;
      } elsif ($severity == $Status{"yellow"}) {
	$return = Locale::gettext::dgettext("special-sauce", "mon_state_net15_yellow");
	$return = eval "\"$return\"";
	return $return;
      } else {
	$return = Locale::gettext::dgettext("special-sauce", "mon_state_net15_red");
	$return = eval "\"$return\"";
	return $return;
      }
    }
    else { # most services work fine within this model
      if ($severity == $Status{"green"}) {
	$return = Locale::gettext::dgettext("special-sauce", "mon_state_green");
	$return = eval "\"$return\"";
	return $return;
      } elsif ($severity <= 0) {
	$return = Locale::gettext::dgettext("special-sauce", "mon_state_noinfo");
	$return = eval "\"$return\"";
	return $return;
      } else {
	$return = Locale::gettext::dgettext("special-sauce", "mon_state_red");
	$return = eval "\"$return\"";
	return $return;
      }
    }
}

sub mon_telnettest
# Test the state of the telnet server
# We use /usr/bin/telnet because getting perl to talk the telnet
# protocol is painful.
# Arguments: the name of the host to test
# Return value: none
# Side effects: makes entries in the monitor database
{
  my ($host) = @_;

  # Check that the inetd process is running
  mon_procwatch("inetd","telnet");

  # Check if the binary is there
  if (! -x $Telnet_bin) {
    Cobalt::Monitor::mon_set_state("telnet","binary",$Status{"red"},"-","-","-","-");
  } else {
    Cobalt::Monitor::mon_set_state("telnet","binary",$Status{"green"},"-","-","-","-");
  }

  # ...and that's pretty much good enough.
  return;

  #use IPC::Open2;
  #use IO::Handle;

  open(LOG,">>$MonDBDir/telnet.log") || open(LOG,">>$Console");

  # Make sure we don't hang forever trying to connect...
  alarm 10;
  $SIG{'ALRM'} = sub { 
    alarm 0;
    $SIG{'ALRM'} = 'DEFAULT';
    Cobalt::Monitor::mon_set_state("telnet","telnet",
				   $Status{"red"},"-","-","-","-");
    print LOG "Failed to send to telnet server: $!\n";
    close LOG;
    return;
  };

  # Open the session
  # open2 allows us to open a session for both reading and writing;
  # the $telnet_rdhandle handle is the read-only filehandle and 
  # $telnet_wrhandle is the write-only filehandle.  
  my $telnet_rdhandle = new IO::Handle;
  my $telnet_wrhandle = new IO::Handle;
  my $pid = open2($telnet_rdhandle,$telnet_wrhandle,"$Telnet_cmd -r -8 $host 2>&1");
  unless ($pid) {
    # the telnet command didn't work for us
    alarm 0;
    $SIG{'ALRM'} = 'DEFAULT';
    Cobalt::Monitor::mon_set_state("telnet","telnet",
				   $Status{"red"},"-","-","-","-");
    print LOG "Could not open telnet session: $!\n";
    close LOG;
    return;
  }

  # Talk to the telnet server
  while (<$telnet_rdhandle>) {
    print STDERR "telnet> $_" if ($Debug && /\S/o);
    print $telnet_wrhandle "\n";  # seems to need this...
    if (/^(login|password):/io) {
      # looks like we got a prompt...
      alarm 0;
      $SIG{'ALRM'} = 'DEFAULT';
      Cobalt::Monitor::mon_set_state("telnet","telnet",
				     $Status{"green"},"-","-","-","-");
      print $telnet_wrhandle "\n~.\n";  # close the session
      $telnet_rdhandle->close;
      $telnet_wrhandle->close;
      close LOG;
      return;
    }
  }

  # If we got here, something is wrong...
  alarm 0;
  $SIG{'ALRM'} = 'DEFAULT';
  Cobalt::Monitor::mon_set_state("telnet","telnet",$Status{"red"},"-","-","-","-");
  print $telnet_wrhandle "\n~.\n";  # close the session
  $telnet_rdhandle->close;
  $telnet_wrhandle->close;
  print LOG "No login prompt from telnet server.\n";
  close LOG;
  return;
}

sub mon_modem
# Check the state of the modem
# Arguments: none
# Side effects: none, hopefully
# Return value: none
{
  # Check if diald is running
  # let's go on the theory that diald is responsible for making sure
  # the modem is okay...so as long as diald is up, our job is done.
  Cobalt::Monitor::mon_procwatch("diald","modem");
}

1;
