#
# $Id: Nat.pm,v 1.33 1999/10/02 16:44:26 duncan Exp $
#
# Written by Brion Moss (brion@cobaltnet.com, brion@nda.com)
#
# Copyright 1998 Cobalt Networks http://www.cobaltnet.com
#
# Network Address Translation and IP Filtering

package Cobalt::Nat;
require Exporter;
require SelfLoader;
@ISA    = qw(Exporter SelfLoader);
@EXPORT = qw(nat_set_rules nat_get_default nat_get_rules
	     nat_enable nat_get_on nat_check_user
	     nat_enable_masq nat_disable_masq nat_get_masq);

require Cobalt::Network;
import Cobalt::Network qw(:DEFAULT);
require Cobalt::Netutil;
import Cobalt::Netutil qw(netutil_test_network_in_network);
use Cobalt::Util;
use Cobalt::Product;

#use Cobalt::Paths;
#use strict;

1;
__DATA__

sub nat_enable
# Enable or disable the filtering rules both now and at boot
# Arguments: 0 = disable, 1 = enable filtering, 2 = enable
#   masquerading, 3 = enable both filtering and masquerading
# Side effects: runs the ip filter script, sets up the boot-time symlink,
#   modifies $MODE in the script
# Return value: status string
# Note: it's a bad idea to have the rules go into effect immediately while
#   the user is editing them.  If they do their edits in the wrong order,
#   things could get really weird and connections could die.
{
  return MSG_error("nat_cant_unsupported") unless ($Features{'nat'});
  my ($filter_mode) = @_;

  # Set the $MODE variable in the filter script
  return MSG_error("nat_cant_noscript") if (! -e $Filter_script);
  return MSG_error("nat_cant_locked") if (-e "$Lockdir/ipfilters");
  open(LOCK,">$Lockdir/ipfilters") || return MSG_error("nat_cant_lockscript", $!);
  open(NAT,$Filter_script) || return MSG_error("nat_cant_read_filter_config", $!);
  while (<NAT>) {
    if (/^MODE=(\d+)/o) {
      print LOCK "MODE=$filter_mode\n";
    } else {
      print LOCK;
    }
  }
  close NAT;
  close LOCK;
  rename("$Lockdir/ipfilters",$Filter_script)
    || return MSG_error("nat_cant_writescript", $!);
  chmod(0755,$Filter_script);

  # Set the symlink.  The symlink should be there regardless,
  # because we turn things on and off with the $MODE variable
  # at the top of the ipfilters script.
  # We know the filter script exists, because we already exited
  # above if it doesn't.
  if( ! -l $Filter_link ) {
    symlink($Filter_script,$Filter_link)
      || return MSG_error("nat_cant_bootscript", $!);
  }

  # Now put those rules (or lack of rules) into effect
  system("$Filter_script >>$Sauce_log 2>&1")
    && return MSG_error("nat_cant_start", $!);

  # Give a return code so we know what just happened...
  if ($filter_mode) {
	return MSG_ok("nat_ok_enabled");
  } else {
	return MSG_ok("nat_ok_disabled");
  }
}

sub nat_get_on
# Checks to see if filtering rules are in effect
# Return value: 0 if filtering & masquerading are off, 
#               1 if filtering is on & masquerading is off,
#               2 if filtering is off and masquerading is on,
#               3 if filtering & masquerading are on.
# Arguments: none
# Side effects: none
{
  return 0 if (! -e $Filter_script);
  open(NAT,$Filter_script) 
    || return MSG_error("nat_cant_read_filter_config", $!);
  while (<NAT>) {
    if (/^MODE=(\d+)/o) {
      close NAT;
      return $1;
    }
  }
  close NAT;
  return 0;
}

sub nat_set_rules
# Save the ruleset to the filter script file
# Arguments: default policy, list of rules
#   each rule is a list consisting of:
#   source ip, source mask, source port, dest ip, dest mask, dest port, 
#              proto, policy, weight
#
#   Policy: "accept" or "deny".
#   Proto: "tcp", "udp", "icmp", or "all".  You must specify a protocol if
#         you specify a port.  You cannot specify a port when "all" is the
#         protocol.
#   Port: a port number or a range of ports seperated by a colon
#          (100:150 would be ports 100 through 150)
#         or "all" for all ports.
#   Netmask: should be in decimal notation (24 instead of 255.255.255.0)
#   Weight: an arbitrary decimal number used in sorting.
#
#   Why do you have to specify a default policy?  Because that's how
#   ipfwadm works.  Don't blame me.
#
#   The rules get sorted automatically by user-specified weight.
#
# Side effects: builds the script which sets the filtering and NAT rules,
#   and sets IP forwarding on in the kernel if we're doing any masquerading.
#   The script is actually built with three clauses--one for if filtering
#   is on, one for if masquerading is on, and one for if filtering and
#   masquerading are on.  A variable at the top is used to switch between
#   them.
# Return value: status string of the form MSG_error("4811")
# Example: Cobalt::Nat::nat_set_rules("deny",
#        [
#         ("207.9.25.0",16,"all","0.0.0.0",0,"all","all","accept",10),
#         ("0.0.0.0",0,"all","207.9.25.1",24,25,"tcp","accept", 5),
#        ]);
#
#    In this example, all outgoing traffic from the 207.9.25.0 network
#    is allowed; incoming traffic to the mail port on 207.9.25.1 is allowed;
#    and all else is denied by default.
#
# NOTE: There is no way to do name-based filtering.
# NOTE 2: This should not be used as a substitute for a real firewall package.
{
  return MSG_error("nat_cant_unsupported") unless ($Features{'nat'});
  my ($defpol,@rulelist) = @_;

  my ($localip,$localmask) = Cobalt::Network::net_get_ipaddr();
  my $filter_mode = Cobalt::Nat::nat_get_on();

  my $we_do_forwarding = 0; # should IP forwarding be on in the kernel?

  my ($r,@rules,@rules_nomasq,@masq_rules,$rulenum);
  # Build the list of addresses and interfaces for this system
  my @all_addresses;          # All the IP addresses this system answers to
  # Doing hashes of arrays doesn't work, so we do this...yeesh.
  my %masqed_interfaces_addr; # The interfaces that get masqueraded
  my %masqed_interfaces_mask; # The netmasks of interfaces that get masqueraded
  my $netif;
  # On a system that supports virtual interfaces, this should be
  # net_get_vinterfaces() or something.
  foreach $netif (Cobalt::Network::net_get_interfaces()) {
    my ($n_addr,$n_mask) = Cobalt::Network::net_get_ipaddr($netif);

    # some interface may not be setup with an IP address
    if (!$n_addr || $n_addr eq '0.0.0.0') { next; }

    push(@all_addresses,$n_addr);
    if (Cobalt::Nat::nat_get_masq($netif)) {
      $masqed_interfaces_addr{$netif} = $n_addr;
      $masqed_interfaces_mask{$netif} = $n_mask;
      push(@masq_rules,
	   "$Filter_cmd -F -a accept -m -S $n_addr/$n_mask # masq $netif");
      # Turn on IP forwarding if masquerading is on
      $we_do_forwarding = 1;
    }
  }

  # Handle each rule in turn
  foreach $r (@rulelist) {
    my @rule = @$r;
    $rulenum++;  # keep track, for error messages
    print STDERR "Rule $rulenum: ", join(' ',@rule), "\n" if ($Debug);
    my ($src_ip,$src_mask,$src_port,$dst_ip,$dst_mask,$dst_port,$proto,
	$policy,$weight) = @rule;

    # Set defaults, fix the arguments
    # Arguments will default to "everyone/everything" if not specified.
    $proto = "all" if ((!$proto) || ($proto =~ /^a/o));
    $src_ip =~ s/(\.|^)0+(\d)/$1$2/go;  # get rid of excess leading zeros
    $src_ip = "0.0.0.0" unless ($src_ip);
    $src_mask = 0 unless ($src_mask);
    $src_mask = Cobalt::Network::net_convert_netmask($src_mask) 
      if ($src_mask =~ /\./o); # decimalize dotted quad netmasks
    $src_port = "" if ((!$src_port) || ($src_port =~ /^a/o));
    $src_port =~ s/-/:/o;  # make dash-specified ranges fit syntax
    $dst_ip =~ s/(\.|^)0+(\d)/$1$2/go;  # get rid of excess leading zeros
    $dst_ip = "0.0.0.0" unless ($dst_ip);
    $dst_mask = 0 unless ($dst_mask);
    $dst_mask = Cobalt::Network::net_convert_netmask($dst_mask) 
      if ($dst_mask =~ /\./o); # decimalize dotted quad netmasks
    $dst_port = "" if ((!$dst_port) || ($dst_port =~ /^a/o));
    $dst_port =~ s/-/:/o;  # make dash-specified ranges fit syntax
    $policy = "accept" if ($policy =~ /^a/oi); # deal with synonyms
    $defpol ||= "accept";
    $weight ||= 50;

    print STDERR "Adding rule $src_ip/$src_mask:$src_port => $dst_ip/$dst_mask:$dst_port [$policy] # $weight\n" if ($Debug);
    
    # Check our arguments, thoroughly
    if ($src_ip !~ /^(\d+\.\d+\.\d+\.\d+)$/o)
    {
      return MSG_error("nat_invalid_src_addr", $src_ip);
    }
    elsif (($src_mask !~ /^\d+$/o) || ($src_mask < 0) || ($src_mask > 32))
    {
      return MSG_error("nat_invalid_src_mask", $rulenum);
    }
    elsif ($src_port !~ /^[\d\:]*$/o)
    {
      return MSG_error("nat_invalid_src_port", $rulenum);
    }
    elsif ($dst_ip !~ /^(\d+\.\d+\.\d+\.\d+)$/o)
    {
      return MSG_error("nat_invalid_dst_addr", $dst_ip);
    }
    elsif (($dst_mask !~ /^\d+$/o) || ($dst_mask < 0) || ($dst_mask > 32))
    {
      return MSG_error("nat_invalid_dst_mask", $rulenum);
    }
    elsif ($dst_port !~ /^[\d\:]*$/o)
    {
      return MSG_error("nat_invalid_dst_port", $rulenum);
    }
    elsif ($proto !~ /^(tcp|udp|icmp|all)$/oi)
    {
      return MSG_error("nat_invalid_proto", $proto);
    }
    elsif ($policy !~ /^(accept|deny)$/o)
    {
      return MSG_error("nat_invalid_policy", $policy);
    }
    elsif ($weight !~ /^\d+$/o)
    {
      return MSG_error("nat_invalid_weight");
    }  
    elsif (($proto eq "all") && ($src_port || $dst_port))
    {
      return MSG_error("nat_invalid_port_no_proto");
    }

    # Figure out wether this rule should apply to incoming traffic
    # or outgoing traffic (ipfwadm makes a distinction).
    my ($dirflag,$forward);
    for (@all_addresses) {
      if (($dst_mask == 32) && ($dst_ip eq $_)) {
	# aimed specifically at this system (dest == my IP/32)
	# Use the -V flag to establish the interface via which
	# this packet is coming (otherwise the rule will apply
	# to all interfaces, which would suck).
	$dirflag = "-I -V $_";
	$forward = 0;
	last;
      }
      elsif (($src_mask == 32) && ($src_ip eq $_)) {
	# coming specifically from this system (source == my IP/32)
	$dirflag = "-O -V $_";
	$forward = 0;
	last;
      }
      elsif (Cobalt::Network::net_network_ismember($_,$dst_ip,$dst_mask)) {
	# this rule involves this system itself (dest == my IP/!32)
	$dirflag = "-I -V $_";
	$forward = 1;
      }
      elsif (Cobalt::Network::net_network_ismember($_,$src_ip,$src_mask)) {
	# this rule involves this system itself (source == my IP/!32)
	$dirflag = "-O -V $_";
	$forward = 1;
      }
    }

    # If we haven't matched anything, then the rule doesn't involve
    # the qube itself...so this must be just a forwarding rule
    # (no -I or -O)
    $forward = 1 unless ($dirflag);

    # Check if this rule is for an address on a masqueraded network--
    # if it is, the forwarding rules need the -m argument
    # Only accepted packets get masqueraded--denied packets just get
    # dropped.
    my ($netif,$masq);
    foreach $netif (keys %masqed_interfaces_addr) {
      my $n_addr = $masqed_interfaces_addr{$netif};
      my $n_mask = $masqed_interfaces_mask{$netif};
      if (($policy eq "accept") &&
	  ((Cobalt::Netutil::netutil_test_network_in_network($src_ip,$src_mask,
							     $n_addr,$n_mask)
	    ||
	    Cobalt::Netutil::netutil_test_network_in_network($dst_ip,$dst_mask,
							     $n_addr,$n_mask)))
	  )
      {
	# Either the source or the destination address is on an interface
	# that is being masqueraded
	$masq = "-m";
      }
    }

    # If a rule has multiple commands added for its sake (for example,
    # if this is a input and forwarding rule), then when
    # we read them in later it's hard to figure out that they should
    # all be the same rule.  So, we set things up so that any rule that
    # is grouped with an earlier rule is indented a few spaces, and when
    # we read the file in we just ignore or merge any such indented rules.
    # Note that we use space, not tab.
    my $indent = "";

    if ($dirflag) {  # is it an Input or Output rule?
      push(@rules,"$Filter_cmd $dirflag -a $policy -P $proto -S $src_ip/$src_mask $src_port -D $dst_ip/$dst_mask $dst_port # $weight");
      push(@rules_nomasq,"$Filter_cmd $dirflag -a $policy -P $proto -S $src_ip/$src_mask $src_port -D $dst_ip/$dst_mask $dst_port # $weight");
      $indent = "   ";
    }
    if ($forward) {  # is it a forwarding rule?
      push(@rules,"$indent$Filter_cmd -F -a $policy $masq -P $proto -S $src_ip/$src_mask $src_port -D $dst_ip/$dst_mask $dst_port # $weight");
      push(@rules_nomasq,"$indent$Filter_cmd -F -a $policy -P $proto -S $src_ip/$src_mask $src_port -D $dst_ip/$dst_mask $dst_port # $weight");
    }
  }

  @rules = sort {
    # sort by weight
    my $a_weight = $1 if ($a =~ /\#\s*(\d+)/o);
    $a_weight ||= 50;
    my $b_weight = $1 if ($b =~ /\#\s*(\d+)/o);
    $b_weight ||= 50;
    return ($a_weight <=> $b_weight);
  } @rules;

  @rules_nomasq = sort {
    # sort by weight
    my $a_weight = $1 if ($a =~ /\#\s*(\d+)/o);
    $a_weight ||= 50;
    my $b_weight = $1 if ($b =~ /\#\s*(\d+)/o);
    $b_weight ||= 50;
    return ($a_weight <=> $b_weight);
  } @rules_nomasq;

  # Now build the script
  return MSG_error("nat_cant_locked") if (-e "$Lockdir/ipfilters");
  open(NAT,">$Lockdir/ipfilters") || return MSG_error("nat_cant_lockscript", $!);
  print NAT "#!/bin/sh\n\n";
  print NAT "# This script sets the IP filtering rules\n";
  print NAT "# This file is edited automatically by the Cobalt admin\n";
  print NAT "# interface.  Any changes made here may be overwritten.\n\n";
  $defpol = "accept" if ($defpol eq "masq");
  print NAT "# 0 = off; 1 = filter only; 2 = masq only; 3 = masq+filter\n";
  print NAT "MODE=$filter_mode\n\n";

  print NAT "# Flush all existing state\n";
  print NAT "$Filter_cmd -F -f\n";
  print NAT "$Filter_cmd -I -f\n";
  print NAT "$Filter_cmd -O -f\n";

  # If filtering and masquerading are off, then just set the default
  # policy to accept; otherwise, set it to the actual default policy.
  # This check is actually done in the shell script, not here.
  print NAT "\nif [ \$MODE = 0 ]; then\n";
  print NAT "\t# Filters are off; set the default policy to accept\n"; 
  print NAT "\t$Filter_cmd -F -p accept\n";
  print NAT "\t$Filter_cmd -I -p accept\n";
  print NAT "\t$Filter_cmd -O -p accept\n";
  print NAT "\texit\n";
  print NAT "else\n";
  print NAT "\t# Set the default policy\n"; 
  print NAT "\t$Filter_cmd -F -p $defpol\n";
  print NAT "\t$Filter_cmd -I -p $defpol\n";
  print NAT "\t$Filter_cmd -O -p $defpol\n";
  print NAT "fi\n";

  print NAT "\n# The rules will be slightly different depending on\n";
  print NAT "# if masquerading is on or not.  Switch between them.\n";
  print NAT "case \"\$MODE\" in\n";
  print NAT "  1)\n";
  print NAT "\t# Filters only\n";
  print NAT join("\n\t","\t# Begin rules", @rules_nomasq, "# End rules\n");
  print NAT "  ;;\n";

  print NAT "  3)\n";
  print NAT "\t# Masquerade and filters\n";
  print NAT join("\n\t","\t# Begin rules", @rules, "# End rules\n");
  print NAT "  ;;\n";
  print NAT "esac\n";  # end case

  print NAT "\n# Add the general maquerading rules, if masquerading is on\n";
  print NAT "if [ \$MODE -gt 1 ]; then\n";
  print NAT join("\n\t",
		 "\t# Begin masquerading rules", 
		 @masq_rules, 
		 "# End masquerading rules\n",
		);
  print NAT "\t/bin/true\n" if ($#masq_rules == -1);
  print NAT "fi\n";
  print NAT "\n";
  close NAT;
  rename("$Lockdir/ipfilters",$Filter_script)
    || return MSG_error("nat_cant_writescript", $!);
  chmod(0755,$Filter_script);

  # Make sure forwarding is turned on
  Cobalt::Network::net_set_forwarding(1) if ($we_do_forwarding);

  return MSG_ok("nat_ok_setrules");
}

sub nat_get_rules
# Read in the rules from the file
# We don't look at what's loaded in the kernel; we look at the file
# Arguments: none
# Side effects: none
# Return value: a list of rules, which is empty if there are none.
#    Each rule is a list of source IP, source netmask, source port, 
#      dest IP, dest netmask, dest port, proto, policy, weight
{
  return if (! -e $Filter_script);
  my @rules;
  open(NAT,$Filter_script) 
    || return MSG_error("nat_cant_read_filter_config", $!);
  while (<NAT>) {
    # Skip forward to the begin marker
    last if (/^\s*\# Begin rules/o);
  }
  while (<NAT>) {
    last if (/^\s*\# End rules/o);
    if (/^\t?$Filter_cmd -(\w)(\s-[VW]\s\S+)? -a (\w+)\s+(-m\s+)?-P (\w+) -S ([\d\.]+)\/([\d\.]+) ([\d:]+)? -D ([\d\.]+)\/([\d\.]+)\s*([\d:]+)?(\s*\#\s*(\d+)\s*(\w*))?/o) {
      my ($type,$policy,$proto,$src_ip,$src_mask,$src_port,$dst_ip,$dst_mask,$dst_port,$weight,$flags) = 
	($1,$3,$5,$6,$7,$8,$9,$10,$11,$13,$14);
      $src_port ||= "all";
      $dst_port ||= "all";
      $weight ||= 50;
      push(@rules,[($src_ip,$src_mask,$src_port,$dst_ip,$dst_mask,$dst_port,
			$proto,$policy,$weight)]);
      print STDERR "Found rule $src_ip/$src_mask:$src_port => $dst_ip/$dst_mask:$dst_port proto=$proto [$policy] # $weight\n" if ($Debug);
      }
  }
  close NAT;
  # Now sort the rules by weight (although they should already be in the
  # file that way)
  @rules = sort { return ($$b[8] <=> $$a[8]); } @rules;
  return @rules;
}

sub nat_get_default
# Find out what the default policy is
# Arguments: none
# Return value: "accept", "deny", "masq", or "none" if no rules are set.
{
  return "none" if (! -e $Filter_script);
  my $policy;
  open(NAT,$Filter_script) 
    || return MSG_error("nat_cant_read_filter_config", $!);
  while (<NAT>) {
    if (/^\t?$Filter_cmd -[FOI] -p (\w+)/o) {
      $policy = $1;
    } elsif (/\# 65536 defmasq/o) {
      $policy = "masq";
      last;
    }
  }
  close NAT;
  return $policy;
}

sub nat_check_user
# Check that the user is still allowed to connect using this ruleset
# Arguments: user's IP, default policy, ruleset
#   Each rule should be a list of:
#   source ip, source netmask, source port, dest ip, dest netmask, dest port, 
#   protocol, policy
# Side effects: none
# Return value: status string
{
  my ($user_ip,$defpol,@rules) = @_;
  my $r;
  foreach $r (@rules) {
    my @rule = @$r;
    my ($src_ip,$src_mask,$src_port,$dst_ip,$dst_mask,$dst_port,$prot,$pol) = @rule;
    if (($pol eq "deny") && (Cobalt::Network::net_network_ismember($user_ip,$src_ip,$src_mask)))
      {
	return MSG_error("nat_cant_address_blocked", $pol);
      }
    elsif (($pol eq "accept") && (Cobalt::Network::net_network_ismember($user_ip,$src_ip,$src_mask)))
      {
	return MSG_ok("nat_ok_explicitly_allowed");
      }
  }
  if ($defpol eq "accept")
    {
      return MSG_ok("nat_ok_defaultly_allowed");
    }
  else
    {
      return MSG_error("nat_cant_defaultly_blocked");
    }
}

sub nat_enable_masq
# Enable IP masquerading for an interface.
# In theory this function should handle virtual interfaces just fine.
# Arguments: interface name or IP and netmask
#   We accept both forms because there isn't really a network interface
#   associated with the modem (at least, not one the sauce understands).
# Return value: message of the form MSG_ok("2671")
# Side effects: Adds a rule to the IP filtering rules to turn on masquerading
#   for this interface
# Examples: nat_enable_masq("eth0");
#            nat_enable_masq("192.168.45.0","255.255.255.0");
#            nat_enable_masq("192.168.45.0","24");
{
  return MSG_error("nat_cant_unsupported") unless ($Features{'nat'});

  # Handle arguments
  my ($n_addr,$n_mask);
  if ($#_) {
    # Arguments given were IP and netmask
    ($n_addr,$n_mask) = @_;
    $n_mask = Cobalt::Network::net_convert_netmask($n_mask) 
      if ($n_mask =~ /\./o); # decimalize dotted quad netmasks
  } else {
    # Argument was the interface name
    my ($netif) = @_;
    ($n_addr,$n_mask) = Cobalt::Network::net_get_ipaddr($netif);
  }

  return MSG_error("nat_cant_locked") if (-e "$Lockdir/ipfilters");
  open(NATOUT,">$Lockdir/ipfilters")
    || return MSG_error("nat_cant_lockscript", $!);
  open(NATIN,"$Filter_script") || return MSG_error("nat_cant_readscript", $!);
  my ($in_masq_section,$masq_set);
  # Scan through the ip filters rule for the masquerading section.
  # If this interface is already being masqueraded in the masquerading
  # section, then we don't need to do anything.
  # Otherwise, add it to the end.
  while (<NATIN>) {
    if (/^\t?\# Begin masquerading rules/o) {
      $in_masq_section = 1;
    }
    elsif (/^\t?\# End masquerading rules/o) {
      $in_masq_section = 0;
      print NATOUT "\t$Filter_cmd -F -a accept -m -S $n_addr/$n_mask # masq\n";
    }
    elsif ($in_masq_section && (/-S $n_addr\/$n_mask \# masq/)) {
      # Oh look, it's already here.  Never mind, let's just clean up and go.
      close NATOUT;
      close NATIN;
      unlink("$Lockdir/ipfilters");
      return MSG_ok("nat_ok_masqed_already");
    }
    print NATOUT;
  }
  close NATOUT;
  close NATIN;
  rename("$Lockdir/ipfilters",$Filter_script)
    || return MSG_error("nat_cant_writescript", $!);
  # The chmod will be done by nat_set_rules

  # Now rebuild the rules to make sure all the -m's are in the
  # right place.
  return Cobalt::Nat::nat_set_rules(Cobalt::Nat::nat_get_default(),
				    Cobalt::Nat::nat_get_rules());
}

sub nat_disable_masq
# Disable IP masquerading for an interface
# Arguments: interface name or (IP, netmask)
#   We accept both forms because there isn't really a network interface
#   associated with the modem (at least, not one the sauce understands).
# Return value: message of the form MSG_error("3124")
# Side effects: Removes rules from the IP filtering rules to masquerading on
#   this interface
# Example: nat_enable_masq("eth0");
{
  return MSG_error("nat_cant_unsupported") unless ($Features{'nat'});

  # Handle arguments
  my ($n_addr,$n_mask);
  if ($#_) {
    # Arguments given were IP and netmask
    ($n_addr,$n_mask) = @_;
    $n_mask = Cobalt::Network::net_convert_netmask($n_mask) 
      if ($n_mask =~ /\./o); # decimalize dotted quad netmasks
  } else {
    # Argument was the interface name
    my ($netif) = @_;
    ($n_addr,$n_mask) = Cobalt::Network::net_get_ipaddr($netif);
  }

  return MSG_error("nat_cant_locked") if (-e "$Lockdir/ipfilters");
  open(NATOUT,">$Lockdir/ipfilters")
    || return MSG_error("nat_cant_lockscript", $!);
  open(NATIN,"$Filter_script") || return MSG_error("nat_cant_readscript", $!);
  my ($in_masq_section,$masq_set);
  # Scan through the ip filters rules for the masquerading rule
  while (<NATIN>) {
    unless (/-S $n_addr\/$n_mask \# masq/) {
      print NATOUT;
    }
  }
  close NATOUT;
  close NATIN;
  rename("$Lockdir/ipfilters",$Filter_script)
    || return MSG_error("nat_cant_writescript", $!);
  # The chmod will be done by nat_set_rules

  # Now rebuild the rules to make sure all the -m's are in the
  # right place.
  return Cobalt::Nat::nat_set_rules(Cobalt::Nat::nat_get_default(),
				    Cobalt::Nat::nat_get_rules());
}

sub nat_get_masq
# Is an interface being masqueraded?
# Arguments: interface name, or IP and netmask
# Return value: 1 if it is, 0 if it isn't
# Example: nat_get_masq("eth0");
{
  return 0 if (! -e $Filter_script);  # there's no masquerading at all...

  # Handle arguments
  my ($n_addr,$n_mask);
  if ($#_) {
    # Arguments given were IP and netmask
    ($n_addr,$n_mask) = @_;
    $n_mask = Cobalt::Network::net_convert_netmask($n_mask) 
      if ($n_mask =~ /\./o); # decimalize dotted quad netmasks
  } else {
    # Argument was the interface name
    my ($netif) = @_;
    ($n_addr,$n_mask) = Cobalt::Network::net_get_ipaddr($netif);
  }

  open(NAT,"$Filter_script") || return 0;
  # Scan through the ip filters rules for the masquerading rule
  while (<NAT>) {
    if (/-S $n_addr\/$n_mask \# masq/) {
      close NAT;
      return 1;
    }
  }
  close NAT;
  return 0;
}

1;
