#
# $Id: Samba.pm,v 1.20 1999/10/02 16:44:27 duncan Exp $
#
# Samba manipulation routines
#
# John D. Blair <jdblair@cobaltnet.com>
# Copyright 1998 Cobalt Networks, Inc., http://www.cobaltnet.com
#
# set_smb_server_on and set_smb_server_off remain in Services.pm

package Cobalt::Samba;
use vars qw(@ISA @EXPORT @EXPORT_OK);
require Exporter;
require SelfLoader;
@ISA    = qw(Exporter SelfLoader);
@EXPORT	= qw(get_master_browsers
	     get_node_status
	     get_workgroups
	     get_smb_services
	     get_smb_service_parms
	     set_smb_service_parms
	     samba_del_service
	     restart_samba
	     scan_for_workgroups
	     scan_for_workgroups_bg
	     is_wins_server
	     samba_service_exists);
use strict;
use Cobalt::Product;
use Cobalt::Util;

1;
__DATA__

sub get_master_browsers
# Returns: a list of the IP addresses of master browsers on the local
#          net segment
# Arguments: none
# Side effects: none
{
  my @ip;
  my $tmpip;
  open(SCAN, "$Smb_nmblookup -M -|");
  while(<SCAN>) {
    next if (/^\D/o);
    ($tmpip) = split;
    push(@ip, $tmpip);
  }
  close(SCAN);
  return @ip;
}


sub get_node_status
# Gets the node status of a SMB server
# Returns: this long list:
# (netbios hostname, workgroup name, username, is browser, is potential
# browser, is local master browser, is domain master browser and/or PDC,
# is logon server)
# failure indicates that the machine is offline or is a win95 host
# Arguments: ip address of the host to scan
# Side effects: if a win95 query is performed nmbd must not be running
{
  my ($ip, $win95) = @_;
  my ($hname, $workgroup, $user, $ismb, $ispmb, $islmb, $isdmb, $islogon);
  my ($name, $type, $group, $win95hack, $status);
  
  $isdmb = 0;
  $ismb = 0;
  $ispmb = 0;
  $islmb = 0;
  $islogon = 0;
  
  if ($win95) {
    $win95hack = ' -r';
  }

  open(STATUS, "$Smb_nmblookup$win95hack -S -U $ip '*'|");
#  print "$Smb_nmblookup $win95hack-S -U $ip '*'|\n";
  foreach $status (<STATUS>) {

    if (/No status response/o) {
      print STDERR "no status from $ip\n";
      return 0;
    }

    if ($status =~ /^\t(.{15,15}) <(..)> . (.{7,7})/o) {
      $name = $1;
      $type = $2;
      $group = ($3 eq '<GROUP>');
    }
    
    # workgroup/domain name
    if (($type eq '00') &&
	($group)) {
      $workgroup = $name;
      $workgroup =~ s/^\s*(.*?)\s*$/$1/;
    }
    
    # hostname
    if (($type eq '00') &&
	(! $group) &&
	(! $hname)) {
      $hname = $name;
      $hname =~ s/^\s*(.*?)\s*$/$1/;
    }
    
    # domain master browser / PDC
    if (($type eq '1b') && (! $group)) {
      $isdmb = 1;
    }
    
    # master browser
    if (($type eq '01') && (! $group)) {
      $ismb = 1;
    }
    
    # local master browser
    if (($type eq '1d') && (! $group)) {
      $islmb = 1;
    }
    
    # potential master browser
    if (($type eq '1e') && ($group)) {
      $ispmb = 1;
    }
    
    # domain logon server
    if (($type eq '1c') && ($group)) {
      $islogon = 1;
    }
  }
  close(STATUS);
  return ($hname, $workgroup, $user, $ismb, $ispmb, $islmb, $isdmb, $islogon);
}


sub get_workgroups 
# Arguments: netbios name, IP address
# Returns: browse list or undefined (failure)
# Side effects: none
{
  my ($netbios, $ip) = @_;
  my ($temp, @list);
  my $inwg = 0;

  open(CLIENT, "$Smbclient -L $netbios -I $ip -U ''|") or 
    return MSG_error("smb_cant_exec_smbclient", $Smbclient);
  
  while (<CLIENT>) {
    if (/^\tWorkgroup\s+Master/o) {
      $inwg = 1;
    }
    
    elsif (/---------/o) {
      # do nothing
    }

    elsif ($inwg &&
	   (/^\t(.{16,16})/o)) {
      $temp = $1;
      # trim whitespace
      $temp =~ s/^\s*(.*?)\s*$/$1/;
      push @list, $temp;
    }
  }
  close(CLIENT);
  return @list;
}    
				     

sub get_smb_services 
# Returns a list of the services in smb.conf
# Arguments: none
# Side effects: none
{
  my (@services);
  
  open(SMB, "$Smbconf") or 
    return MSG_error("smb_cant_read_smbconf", $Smbconf);
  while (<SMB>) {
    # ignore comments
    if (! /^.*[;\#]/o) {
      # watch for service header
      if (/\[([^\]]+)\]/o) {
	push @services, $1;
      }
    }
  }
  close(SMB);
  
  return @services;
}

sub samba_service_exists
# return 1 if service exists, 0 if it does not
# Arguments: $service
#
# this may lead to potentially inefficient code (multiple readings of the
# smb.conf file when a single read could have been done) but using this
# makes my code more readable to me.
{
  my $service = shift;
  my $found = 0;
  
  open(SMB, "$Smbconf") or 
    return MSG_error("smb_cant_read_smbconf", $Smbconf);
  while (<SMB>) {
    # ignore comments
    if (! /^.*[;\#]/o) {
      # watch for service header
      if (/\[([^\]]+)\]/o) {
	$found = 1 if ($1 eq $service);
      }
    }
  }
  close(SMB);
  
  return $found;
}

sub get_smb_service_parms 
{
  my ($name, $hash) = @_;
  my ($parm, $value);
  my $inserv = 0;
  my $found = 0;

  open(SMB, "$Smbconf") or 
    return MSG_error("smb_cant_read_smbconf", $Smbconf);
  while (<SMB>) {
    # ignore comments
    next if (/^.*[;\#]/o);

    # watch for service header
    if (/\[([^\]]+)\]/o) {
      $inserv = ($1 =~ /$name/i);
      $found = 1 if ($inserv);
    }
    # load setting
    if (($inserv) &&
	(/^([\w ]+)=(.*)$/o)) {
      $parm = $1;
      $value = $2;
      $parm =~ s/^\s*(.*?)\s*$/$1/;
      $parm =~ tr/[A-Z]/[a-z]/;
      $value =~ s/^\s*(.*?)\s*$/$1/;	
      $$hash{$parm} = $value;
    }
  }
  close(SMB);
  
  if ($found) {
    return 0;
  } else {
    return MSG_error("samba_cant_nosuchservice");
  }  
}


sub set_smb_service_parms
{
  my ($name, $newname, $hash) = @_;
  my ($parm, $value);
  my $inserv = 0;

  return MSG_error("smb_cant_smbconf_locked") if (-e "$Lockdir/smb.conf");
  open(SMBLOCK,">$Lockdir/smb.conf") 
    or return MSG_error("smb_cant_write_lock");
  open(SMB, "$Smbconf") 
    or return MSG_error("smb_cant_read_smbconf", $Smbconf);
  while (<SMB>) {
    # ignore comments
    if (/^.*[;\#]/o) {
      print SMBLOCK $_;
      next;
    }

    # watch for service header
    if (/\[([^\]]+)\]/o) {
      if ($1 =~ /$name/i) {
	# write parameters
	if (defined($newname)) {
	  print SMBLOCK "[$newname]\n";
	} else {
	  print SMBLOCK;
	}
	  
	foreach $parm (sort (keys %$hash)) {
	  print SMBLOCK "   $parm = $$hash{$parm}\n";
	}
	$inserv = 1;
	next;
      } else {
	$inserv = 0;
      }
    }
    
    # write setting
    if (! $inserv) {
      print SMBLOCK $_;
    }
  }
  close(SMB);
  close(SMBLOCK);
  rename("$Lockdir/smb.conf", $Smbconf);

  return 1;
}

sub samba_del_service 
{
  my $name = shift;
  my $inserv = 0;

  return MSG_error("smb_cant_smbconf_locked") if (-e "$Lockdir/smb.conf");
  open(SMBLOCK,">$Lockdir/smb.conf") 
    or return MSG_error("smb_cant_write_lock");
  open(SMB, "$Smbconf") 
    or return MSG_error("smb_cant_read_smbconf", $Smbconf);
  while (<SMB>) {
    # ignore comments
    if (/^.*[;\#]/o) {
      print SMBLOCK;
      next;
    }

    # watch for service header
    if (/\[([^\]]+)\]/o) {
      if ($1 =~ /$name/i) {
	# do nada
	$inserv = 1;
	next;
      } else {
	$inserv = 0;
      }
    }
    
    # write setting
    if (! $inserv) {
      print SMBLOCK;
    }
  }
  close(SMB);
  close(SMBLOCK);
  rename("$Lockdir/smb.conf", $Smbconf);

  return 0;
}

sub restart_samba
# Start a child process to restart the samba server
{
    my $pid;
    if (!defined($pid = fork)) {
	die "cannot fork: $!";
    } elsif ($pid) {
	# i'm the parent
	print STDERR "forking a child to restart daemons\n" if ($Debug);
    }
    else
    {
	# otherwise I'm the child
	open(STDERR,">>$Sauce_log");
	open(STDOUT,">>$Sauce_log");
	# restart samba
	if (-f "$Subsys/smb")
	{
	    system("$Smbscript stop >>$Sauce_log 2>&1 ");
	    system("$Smbscript start >>$Sauce_log 2>&1 ");
	}
	system ("$Lcdshowscript start >>$Sauce_log 2>&1")
	    if (-x $Lcdshowscript);
	exit;
    }

    # done!
    return 1;
}

sub scan_for_workgroups
# scan_for_workgroups
# Scans the network for workgroups.
{
  my ($hname, $workgroup, $user, $ismb, $ispmb, $islmb, $isdmb, $islogon, %wglist, $ip, $wg);

  foreach $ip (Cobalt::Samba::get_master_browsers()) {
    # we have to try twice because of the Win95 node status bug
    # use win95 workaround first
    ($hname, $workgroup, $user, $ismb, $ispmb, $islmb, $isdmb, $islogon) = Cobalt::Samba::get_node_status($ip, 1);
    if ($hname eq '') {
      # use normal node status request
      ($hname, $workgroup, $user, $ismb, $ispmb, $islmb, $isdmb, $islogon) = Cobalt::Samba::get_node_status($ip, 0);
    }
    
    # get workgroups from master browser
    foreach $wg (Cobalt::Samba::get_workgroups($hname, $ip)) {
      $wglist{$wg} = 1;
    }
  }
  
  return keys (%wglist);
  
  # done!
  return 1;
}

sub scan_for_workgroups_bg
# scan_for_workgroups_bg
# Forks a process to scan the network for a list of workgroups.
{
  my ($pid, @wglist, %smbconf);

  if (!defined($pid = fork)) {
    die "cannot fork: $!";
  } elsif ($pid) {
    # i'm the parent
    print STDERR "forking a child to scan for workgroups\n" if ($Debug);
  }
  else
    {
      # otherwise I'm the child
      if (-e $Smb_scanlock) {
	# another process is already performing the scan
	exit;
      }
      
      # there's a race condition between the above command and the next,
      # but I'm over it-- its unlikely to cause us trouble.
      
      # set the scan lock
      open(LOCK, ">$Smb_scanlock"); 
      print LOCK "$$\$0";
      close(LOCK);
      # open the workgroup list file
      open(LIST, ">$Smb_wglist");
      
      # add locally configured workgroup to the file, if necessary
      my $found = 0;
      Cobalt::Samba::get_smb_service_parms('global', \%smbconf);
      @wglist = Cobalt::Samba::scan_for_workgroups();
      foreach (@wglist) {
	if (/$smbconf{'workgroup'}/io) {
	  $found = 1;
	}
      }
      if (! $found) {
	push @wglist, $smbconf{'workgroup'};
      }

      print LIST join("\n", sort(@wglist));;

      close(LIST);
      # remove scan lock
      unlink($Smb_scanlock);
      
      exit;
    }
  
  # done!
  return 1;
}

sub is_wins_server
{
  my ($ip) = @_;
  my $queryip;

  my @status = Cobalt::Samba::get_node_status($ip, 0);
  
  open (QUERY, "$Smb_nmblookup -U $ip -R $status[1]#00|")
    || return 0;  # Can't run nmblookup for some reason...
  foreach (<QUERY>) {
    ($queryip) = split(' ', $_);
    next if ($queryip eq 'Sending');
    return 0 if ($queryip eq 'name_query');
    return 1 if ($queryip eq '255.255.255.255');
  }
  close(QUERY);

  # if we're here there's been a parsing failure
  return 0;
}

1;
