#!/usr/bin/perl

# useful functions... nuff said!
#
# written by Duncan Laurie <duncan@cobaltnet.com>
#

package Cobalt::Util;

use POSIX;
use Locale::gettext;

require Exporter;
require SelfLoader;
@ISA = qw(Exporter SelfLoader);
@EXPORT = qw(FILE_lock_r FILE_lock_w FILE_lock_rw
	     FILE_unlock FILE_unlock_rw
	     MSG_error MSG_abort MSG_ok MSG_get MSG_get_ui);

use vars qw($DEBUG);
$DEBUG=0;

1;

__DATA__

# Cobalt::Util::MSG_get_ui
#
# generate an message for the UI
# using predetermined locale
#
# arguments are substituted for $arg in the message string
#
sub MSG_get_ui
{
    my $msgkey = shift;
    my @vars = @_;
    my $msg = Locale::gettext::dgettext("ui", $msgkey);
    foreach (@vars) {
	$msg =~ s/\$[^\\\ \.\"\']+/$_/;
    }
    return $msg;
}

# Cobalt::Util::MSG_get
#
# generate an message
# using predetermined locale
#
# arguments are substituted for $arg in the message string
#
sub MSG_get
{
    my $msgkey = shift;
    my @vars = @_;
    my $msg = Locale::gettext::dgettext("special-sauce", $msgkey);
    foreach (@vars) {
	$msg =~ s/\$[^\\\ \.\"\']+/$_/;
    }
    return $msg;
}

# Cobalt::Util::MSG_error
#
# generate an error message prefixed with "4999" from gettext
# using predetermined locale
#
# arguments are substituted for $arg in the message string
#
sub MSG_error
{
    my $msgkey = shift;
    my @vars = @_;
    my $msg = Locale::gettext::dgettext("special-sauce", $msgkey);
    foreach (@vars) {
	$msg =~ s/\$[^\\\ \.\"\']+/$_/;
    }
    $msg = '4999 '.$msg;
    return $msg;
}

# Cobalt::Util::MSG_abort
#
# generate an error message prefixed with "3999" from gettext
# using predetermined locale
#
# arguments are substituted for $arg in the message string
#
sub MSG_abort
{
    my $msgkey = shift;
    my @vars = @_;
    my $msg = Locale::gettext::gettext("special-sauce", $msgkey);
    foreach (@vars) {
	$msg =~ s/\$[^\\\ \.\"\']+/$_/;
    }
    $msg = '3999 '.$msg;
    return $msg;
}

# Cobalt::Util::MSG_ok
#
# generate a message from the provided key using pre-defined locale
# prefix this message with "2999" to show that it is a good message
#
# arguments are substituted for $arg in the message string
#
sub MSG_ok
{
    my $msgkey = shift;
    my @vars = @_;
    my $msg = Locale::gettext::dgettext("special-sauce", $msgkey);
    foreach (@vars) {
	$msg =~ s/\$[^\\\ \.\"\']+/$_/;
    }
    $msg = '2999 '.$msg;
    return $msg;
}

# Cobalt::Util::FILE_lock_rw
#
# inputs:  filein = name of file to lock (& full path)
# return:  two filehandles, the first for the input and the second for output
#
# creates a lockfile in $Lockdir (/etc/cobalt/locks)
# for writing, while opening the $file for reading
# 
sub FILE_lock_rw
{
    my $filein = shift;
    local *FHIN = shift;
    local *FHOUT = shift;
    my $LOCK_EX = 2;
    my $Lockdir = "/etc/locks";

    # extract filename from fully-qualified filename
    my @fary = split("/", $filein);
    my $ret = pop(@fary);
    my $fileout = "$Lockdir/".pop(@fary)."_$ret";

    # we want to have exclusive locking on the input file
    print STDERR "FILE_lock_rw:  opening $filein... " if ($DEBUG);
    unless(open(FHIN, "< $filein")) {
	print STDERR "error!\n" if ($DEBUG);
	return;
    }
    print STDERR "done\n" if ($DEBUG);

    print STDERR "FILE_lock_rw:  locking $filein... " if ($DEBUG);
    unless(flock(FHIN,$LOCK_EX)) {
	print STDERR "error!\n" if ($DEBUG);
	return;
    }
    print STDERR "done\n" if ($DEBUG);

    print STDERR "FILE_lock_rw:  opening $fileout... " if ($DEBUG);
    unless(open(FHOUT, "> $fileout")) {
	print STDERR "error!\n" if ($DEBUG);
	return;
    }
    print STDERR "done\n" if ($DEBUG);

    print STDERR "FILE_lock_rw:  locking $fileout... " if ($DEBUG);
    unless(flock(FHOUT,$LOCK_EX)) {
	print STDERR "error!\n" if ($DEBUG);
	return;
    }
    print STDERR "done\n" if ($DEBUG);

    return(1);
}

#
# Cobalt::Util::FILE_lock_w
#
# inputs:  file = name of file to lock (& full path)
# return:  filehandle
#
# opens the $file for reading
#
sub FILE_lock_w
{
    my $file = shift;
    local *FH = shift;
    my $LOCK_EX = 2;

    # open the file
    print STDERR "FILE_lock_w:  opening $file... " if ($DEBUG);
    unless(open(FH, "> $file")) {
	print STDERR "error!\n" if ($DEBUG);
	return;
    }
    print STDERR "done\n" if ($DEBUG);

    # lock the file
    print STDERR "FILE_lock_w:  locking $file... " if ($DEBUG);
    unless(flock(FH,$LOCK_EX)) {
	print STDERR "error!\n" if ($DEBUG);
	return;
    }
    print STDERR "done\n" if ($DEBUG);

    return(1);
}

#
# Cobalt::Util::FILE_lock_r
#
# inputs:  file = name of file to lock (& full path)
# return:  filehandle
#
# opens the $file for reading
#
sub FILE_lock_r
{
    my $file = shift;
    local *FH = shift;
    my $LOCK_EX = 2;

    # open the file
    print STDERR "FILE_lock_r:  opening $file... " if ($DEBUG);
    unless(open(*FH,"< $file")) {
	print STDERR "error!\n" if ($DEBUG);
	return;
    }
    print STDERR "done\n" if ($DEBUG);

    # lock the file
    print STDERR "FILE_lock_r:  locking $file... " if ($DEBUG);
    unless(flock(FH,$LOCK_EX)) {
	print STDERR "error!\n" if ($DEBUG);
	return;
    }
    print STDERR "done\n" if ($DEBUG);

    return(1);
}

#
# Cobalt::Util::FILE_unlock
#
# inputs:  array of filehandles to unlock and close
# return:  1 if success
#
sub FILE_unlock
{
    my (@fhs) = @_;
    my ($fh);
    my $LOCK_UN = 8;

    foreach $fh (@fhs) {
	print STDERR "FILE_unlock:  unlocking filehandle >$fh<... " if ($DEBUG);
	flock($fh, $LOCK_UN);
	close($fh);
	print STDERR "done\n" if ($DEBUG);
    }

    return(1);
}

#
# Cobalt::Util::FILE_unlock_rw
#
# inputs:  $filein = full path to file
#          $fhin = input filehandle returned by FILE_lock_rw
#          $fhout = output filehandle returned by FILE_lock_rw
# return:  1 if succes
#
sub FILE_unlock_rw
{
    my ($filein,$fhin,$fhout) = @_;
    my $LOCK_UN = 8;
    my $Lockdir = "/etc/locks";

    # extract filename from fully-qualified filename
    my @fary = split("/", $filein);
    my $ret = pop(@fary);
    my $fileout = "$Lockdir/".pop(@fary)."_$ret";

    print STDERR "FILE_unlock_rw:  unlocking $fileout... " if ($DEBUG);
    flock($fhout, $LOCK_UN);
    close($fhout);
    print STDERR "done\n" if ($DEBUG);

    print STDERR "FILE_unlock_rw:  moving $fileout to $filein... " if ($DEBUG);
    rename($fileout,$filein);
    print STDERR "done\n" if ($DEBUG);

    print STDERR "FILE_unlock_rw:  unlocking $filein... " if ($DEBUG);
    flock($fhin, $LOCK_UN);
    close($fhin);
    print STDERR "done\n" if ($DEBUG);

    return(1);
}

1;

