#!/usr/bin/perl

# perl component of swatch
#
# written by Duncan Laurie <duncan@cobaltnet.com>
#

package Cobalt::swatch;

use Cobalt::Meta;
use Cobalt::Util;
use Cobalt::Vsite;
use Cobalt::Quota;
use Cobalt::DiskUtils;
#use Cobalt::Product;

require Exporter;
require SelfLoader;

@ISA = qw(Exporter SelfLoader);
@EXPORT = qw(swatch_get
	     swatch_get_modify
	     swatch_get_defcon
	     swatch_get_defcon3_all
	     swatch_get_defcon_disk
	     swatch_get_defcon_sites
	     swatch_get_defcon_email
	     swatch_get_defcon_quota_user
	     swatch_get_defcon_quota_vsite
	     swatch_get_defcon_quota_member
	     swatch_set_service_on
	     swatch_set_service_off);

1;

__DATA__

# swatch_get
# 
# get a component monitored by swatch
#
# returns a meta object
#
sub swatch_get
{
    my $type = shift;
    my $name = shift;
    
    $name = $type unless ($name);
    
    $obj = Cobalt::Meta->create("type" => $type);
    $obj->retrieve($name);
    
    return $obj;
}

# swatch_get_modify
# 
# get a datetime value of a component
#
# returns a datetime string
#
sub swatch_get_modify
{
    my $special = shift;
    my $name = shift;
    
    $name = $special unless (defined($name));
    
    $obj = Cobalt::Meta->create("type" => "defcon");
    $obj->retrieve($name, ["and", "special", "=", $special]);

    my $modify = $obj->get("modify");

    return $modify;
}

# swatch_get_defcon
# 
# get a defcon value of a component monitored by swatch
#
# inputs: name
#         special (aka type)
#
# returns a defcon value (0-3)
#
sub swatch_get_defcon
{
    my $special = shift;
    my $name = shift;

    $name = $special unless (defined($name));
    
    $obj = Cobalt::Meta->create("type" => "defcon");
    $obj->retrieve($name, ["and", "special", "=", $special]);

    my $defcon = $obj->get("defcon");
    $defcon = $defcon ? $defcon : 0;
    
    return $defcon;
}

# swatch_get_defcon3_all
#
# returns number of monitored objects that have defcon 3
#
sub swatch_get_defcon3_all
{
    my @s = Cobalt::Meta::query("type"  => "defcon",
				"keys"  => ["name", "defcon"],
				"where" => ["defcon", "=", "3"]);
    return (scalar(@s));
}

# swatch_get_defcon_email
# 
# get a defcon value of the email subsystem
# tests services
#   snmp, imap, pop3
#
# returns a defcon value (0-3)
#
sub swatch_get_defcon_email
{
    my ($d,$defcon);

    my $defcon = swatch_get_defcon("service", "smtp");
    $d = swatch_get_defcon("service", "pop3");
    $defcon = ($d > $defcon) ? $d : $defcon;
    $d = swatch_get_defcon("service", "imap");
    $defcon = ($d > $defcon) ? $d : $defcon;

    return $defcon;
}

# swatch_get_defcon_sites
# 
# get a defcon value for the sites
#
# returns a defcon value (0-3)
#
sub swatch_get_defcon_sites
{
    my $users = swatch_get_defcon("quota", "users");
    my $vsite = swatch_get_defcon("quota", "vsite");
    my $defcon = ($users > $vsite) ? $users : $vsite;

    return $defcon;
}

# swatch_get_defcon_disk
# 
# get a defcon value for the disks
#
# returns a defcon value (0-3)
#
sub swatch_get_defcon_disk
{
    my @s = Cobalt::Meta::query("type"  => "defcon",
				"keys"  => ["name", "defcon"],
				"where" => ["special", "=", "quota"]);
    my $defcon = 0;
    my $d;

    foreach (@s) {
	$d = $_->get("defcon");
	$defcon = ($d > $defcon) ? $d : $defcon;
    }

    return $defcon;
}

sub swatch_get_defcon_quota_user
{
    my $name = shift;

    my $defcon = swatch_get_defcon("usrquota", $name);
    $defcon = $defcon ? $defcon : 1;

    return ($defcon);
}

sub swatch_get_defcon_quota_vsite
{
    my $site = shift;

    my $defcon = swatch_get_defcon("grpquota", $site);
    $defcon = $defcon ? $defcon : 1;

    return $defcon;
}

# get defcon value for site members
sub swatch_get_defcon_quota_member
{
    my $group = shift;
    my @users = vsite_get_users($group);

    my $defcon = swatch_get_defcon_quota_vsite($group);

    my ($user,$udef);
    foreach $user (@users) {
	$udef = swatch_get_defcon_quota_user($user);
	$defcon = ($defcon > $udef) ? $defcon : $udef;
    }

    return $defcon;
}

sub swatch_set_service_on
{
    my $service = shift;

    my $conf = "/etc/cobalt/swatch/services/$service";

    FILE_lock_rw($conf, *IN, *OUT) ||
	return MSG_error("swatch_cant_set_service_on");

    while (<IN>) {
	if (/^state/) {
	    print OUT "state = on\n";
	} else {
	    print OUT;
	}
    }

    FILE_unlock_rw($conf, *IN, *OUT);
    return MSG_ok("swatch_set_service_on");
}

sub swatch_set_service_off
{
    my $service = shift;

    my $conf = "/etc/cobalt/swatch/services/$service";

    FILE_lock_rw($conf, *IN, *OUT) ||
	return MSG_error("swatch_cant_set_service_off");

    while (<IN>) {
	if (/^state/) {
	    print OUT "state = off\n";
	} else {
	    print OUT;
	}
    }

    FILE_unlock_rw($conf, *IN, *OUT);
    return MSG_ok("swatch_set_service_off");
}

1;
